/*
 * _mem.cpp - MemoryMgr layer.
 */

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#include "precomp.h"

#ifndef OS_PALM

#include "palmlayer.h"

class Block
{
    public:
	static Block*	FromPtr( MemPtr p );
	static Block*	FromHandle( MemHandle h );

    public:
			Block( Boolean fMovable );

	MemHandle	GetMemHandle() const { return MemHandle( this ); }
	MemPtr		GetMemPtr() const { return ( UInt8* )m_pv + sizeof( MemPtr ); }

	UInt32		GetSize() const { return m_cbSize; }
	UInt16		GetLockCount() const { return m_cLocks; }

	Err		Alloc( UInt32 newSize );
	Err		Resize( UInt32 newSize );
	void		Free();

	MemPtr		Lock();
	Err		Unlock();
	Err		ResetLock();

    private:
	MemPtr		m_pv;

	UInt32		m_cbSize;		// actual size of block
	UInt16		m_cLocks;		// locks, just for sanity
};

Block::Block( Boolean fMovable )
{
	m_pv = 0;
	m_cbSize = 0;
	m_cLocks = fMovable ? 0 : memPtrLockCount;
}

Block*
Block::FromPtr( MemPtr p )
{
	void** ppv;
	Block* pBlock;

	Assert( p );

	ppv = ( void** )MemPtr( p );
	ppv--;

	pBlock = ( Block* )*ppv;

	Assert( pBlock->GetMemPtr() == p );

	return pBlock;
}

Block*
Block::FromHandle( MemHandle h )
{
	Assert( h );
	return ( Block* )h;
}

Err
Block::Alloc( UInt32 newSize )
{
	MemPtr pv;

	Assert( newSize );
	Assert( !m_pv );

	pv = MemPtr( MemAlloc( newSize + sizeof( MemPtr ) ) );
	if( !pv )
	    return memErrNotEnoughSpace;

	*( MemPtr* )pv = this;

	m_pv = pv;
	m_cbSize = newSize;
	return 0;
}

Err
Block::Resize( UInt32 const newSize )
{
	if( newSize != m_cbSize )
	{
	    MemPtr pv;

	    Assert( newSize );

	    if( m_cLocks )
		return memErrChunkLocked;

	    pv = MemPtr( MemReAlloc( m_pv, newSize + sizeof( MemPtr ) ) );
	    if( !pv )
		return memErrNotEnoughSpace;

	    *( MemPtr* )pv = this;

	    m_pv = pv;
	    m_cbSize = newSize;
	}

	return 0;
}

void
Block::Free()
{
	MemFree( m_pv );
	m_pv = 0;
	m_cbSize = 0;
	m_cLocks = 0;
}

MemPtr
Block::Lock()
{
	Assert( m_cLocks < memPtrLockCount );

	if( m_cLocks < memPtrLockCount )
	    m_cLocks++;

	return GetMemPtr();
}

Err
Block::Unlock()
{
	if( m_cLocks < memPtrLockCount )
	{
	    if( m_cLocks )
		m_cLocks--;
	    else
		PanicSz( "under locked" );
	}

	return 0;
}

Err
Block::ResetLock()
{
	if( m_cLocks != memPtrLockCount )
	    m_cLocks = 0;
	return 0;
}

/*
 * Memory Manager Interface.
 */

MemPtr
MemChunkNew( UInt16 heapID, UInt32 size, UInt16 attr )
{
	SUPPRESS_UNUSED_VAR( heapID );
	SUPPRESS_UNUSED_VAR( attr );

	return MemPtrNew( size );
}

Err
MemChunkFree( MemPtr chunkDataP )
{
	Err err;
	MemHandle h;

	h = MemPtrRecoverHandle( chunkDataP );
	if( !h )
	    return memErrInvalidParam;

	err = MemHandleUnlock( h );
	if( err )
	    return err;

	return MemHandleFree( h );
}

MemPtr
MemPtrNew( UInt32 size )
{
	Block* const pBlock = new Block( false );

	if( !pBlock )
	    return 0;

	if( pBlock->Alloc( size ) )
	{
	    delete pBlock;
	    return 0;
	}

	return pBlock->GetMemPtr();
}

Err
MemPtrUnlock( MemPtr p )
{
	MemHandle h;

	Assert( p );

	h = MemPtrRecoverHandle( p );
	if( !h )
	    return memErrInvalidParam;

	return MemHandleUnlock( h );
}

MemHandle
MemHandleNew( UInt32 size )
{
	Block* const pBlock = new Block( true );

	if( !pBlock )
	    return 0;

	if( pBlock->Resize( size ) )
	{
	    delete pBlock;
	    return 0;
	}

	return pBlock->GetMemHandle();
}

Err
MemHandleFree( MemHandle h )
{
	if( !h )
	    return memErrInvalidParam;

	Block* const pBlock = Block::FromHandle( h );

	// Strange not to check the lock count, but go look at the OS source.

	pBlock->Free();
	delete pBlock;

	return 0;
}

MemHandle
MemPtrRecoverHandle( MemPtr p )
{
	return Block::FromPtr( p )->GetMemHandle();
}

UInt32
MemPtrSize( MemPtr p )
{
	return Block::FromPtr( p )->GetSize();
}

UInt32
MemHandleSize( MemHandle h )
{
	return Block::FromHandle( h )->GetSize();
}

UInt16
MemHandleLockCount( MemHandle h )
{
	return Block::FromHandle( h )->GetLockCount();
}

LocalID
MemHandleToLocalID( MemHandle h )
{
	return LocalID( h );
}

Err
MemHandleResize( MemHandle h, UInt32 newSize )
{
	return Block::FromHandle( h )->Resize( newSize );
}

MemPtr
MemHandleLock( MemHandle h )
{
	return Block::FromHandle( h )->Lock();
}

Err
MemHandleUnlock( MemHandle h )
{
	return Block::FromHandle( h )->Unlock();
}

Err
MemHandleResetLock( MemHandle h )
{
	return Block::FromHandle( h )->ResetLock();
}

MemPtr
MemLocalIDToGlobal( LocalID local, UInt16 cardNo )
{
	SUPPRESS_UNUSED_VAR( cardNo );

	return MemPtr( local );
}

MemPtr
MemLocalIDToLockedPtr( LocalID local, UInt16 cardNo )
{
	SUPPRESS_UNUSED_VAR( cardNo );

	return Block::FromHandle( MemHandle( local ) )->Lock();
}

#if 0
MemPtr
MemEmergencyPtrFromHandle( MemHandle h )
{
	return Block::FromHandle( h )->GetMemPtr();
}
#endif

#endif // !OS_PALM

