/*
 * _pdb.cpp - Database class to abstract access to a Palm .PDB database file.
 */

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#include "precomp.h"

#ifndef OS_PALM

#include "palmlayer.h"
#include "str.h"
#include "sys.h"

#include <ctype.h>
#include <time.h>

/*
 * PdbHeader.
 */

PdbHeader::PdbHeader()
{
	memset( this, 0, sizeof( *this ) );
}

UInt32
PdbHeader::GetNextUid()
{
	m_ulSeed++;

	if( m_ulSeed & 0xff000000 )
	    m_ulSeed = ( dmRecordIDReservedRange + 1 ) << 12;

	return m_ulSeed;
}

UInt32
PdbHeader::GetAppInfoSize( RecordEntry const* pRecordZero, UInt32 cbFileSize ) const
{
	Assert( FImplies( NumRecords(), pRecordZero ) );
	Assert( FImplies( !NumRecords(), !pRecordZero ) );

	if( GetOfsSortInfo() )
	    return GetOfsSortInfo() - GetOfsAppInfo();

	if( NumRecords() )
	    return pRecordZero->GetOfsData() - GetOfsAppInfo();

	return cbFileSize - GetOfsAppInfo();
}

/*
 * RawRecordEntry.
 */

UInt32
RawRecordEntry::GetUid() const
{
	UInt32 uid;

#ifdef OS_MAC
	uid = ( m_uid[2] << 16 ) + ( m_uid[1] << 8 ) + ( m_uid[0] );
#else
	uid = ( m_uid[0] << 16 ) + ( m_uid[1] << 8 ) + ( m_uid[2] );
#endif

	return uid;
}

void
RawRecordEntry::SetUid( UInt32 uid )
{
#ifdef OS_MAC
	m_uid[2] = UInt8( uid >> 16 );
	m_uid[1] = UInt8( uid >> 8 );
	m_uid[0] = UInt8( uid >> 0 );
#else
	m_uid[0] = UInt8( uid >> 16 );
	m_uid[1] = UInt8( uid >> 8 );
	m_uid[2] = UInt8( uid >> 0 );
#endif
}

/*
 * RecordEntry.
 */

UInt32
RecordEntry::GetSize() const
{
	return MemHandleSize( m_hData );
}

/*
 * RecordList.
 */

RecordList::RecordList()
{
	memset( this, 0, sizeof( *this ) );
}

RecordList::~RecordList()
{
	Reset();
}

void
RecordList::FreeArray( RecordEntry*** ppprg, UInt16* pc )
{
	if( *ppprg )
	{
	    while( ( *pc )-- )
		delete ( *ppprg )[*pc];

	    MemFree( *ppprg );
	}

	*ppprg = 0;
	*pc = 0;
}

Err
RecordList::Insert( unsigned ii, MemHandle h )
{
	// No fancy data structures to make this fast; this is freeware and I
	// don't have the time.

	if( ii > GetCount() )
	    return dmErrIndexOutOfRange;

	if( GetCount() + 1 >= LOWORD( noRecordSelected ) )
	    return dmErrIndexOutOfRange;

	if( m_c >= m_cAlloc )
	{
	    RecordEntry** pprg;
	    size_t const cAlloc = Min( ( m_c + 1 ) * 2, 65534 ); // Geometric growth.

	    Assert( cAlloc < 65535 );

	    pprg = ( RecordEntry** )MemReAlloc( m_pprg, cAlloc * sizeof( m_pprg[0] ) _MEM_NOSIZECHECK );
	    if( !pprg )
		return dmErrMemError;

	    m_pprg = pprg;
	    m_cAlloc = UInt16( cAlloc );
	}

	RecordEntry* const p = new RecordEntry;
	if( !p )
	    return dmErrMemError;

	p->SetAttr( 0 );
	p->SetUid( m_pdb->GetNextUid() );
	p->m_hData = h;

	memmove( &m_pprg[ii + 1], &m_pprg[ii], ( m_c - ii ) * sizeof( m_pprg[0] ) );

	m_pprg[ii] = p;
	m_c++;

	return 0;
}

Err
RecordList::Remove( unsigned ii, MemHandle* phOld )
{
	Assert( ii < GetCount() );

	if( ii >= GetCount() )
	    return dmErrIndexOutOfRange;

	RecordEntry* const pEntry = GetEntry( ii );
	Assert( pEntry );

	*phOld = pEntry->m_hData;

	pEntry->m_hData = 0;
	delete pEntry;

	m_c--;

	memmove( &m_pprg[ii], &m_pprg[ii + 1], ( m_c - ii ) * sizeof( m_pprg[0] ) );

	return 0;
}

Err
RecordList::Move( unsigned from, unsigned to )
{
	RecordEntry* p = GetEntry( from );

	if( !p )
	{
	    Panic1( "'from' out of range (%u)", from );
	    return dmErrIndexOutOfRange;
	}

	if( int( to ) < 0 || to > GetCount() )
	{
	    Panic1( "'to' out of range (%u)", to );
	    return dmErrIndexOutOfRange;
	}

	// Move other record pointers.

	if( to < from )
	{
	    memmove( m_pprg + to + 1, m_pprg + to, ( from - to ) * sizeof( *m_pprg ) );
	}
	else if( from < to )
	{
	    memmove( m_pprg + from, m_pprg + from + 1, ( to - from ) * sizeof( *m_pprg ) );
	    to--;
	}

	// Put record pointer in new location.

	m_pprg[to] = p;

	return 0;
}

Err
RecordList::FindByID( UInt32 uniqueID, unsigned* indexP )
{
	Assert( indexP );

	RecordEntry const** ppreWalk = const_cast<RecordEntry const**>( m_pprg );
	RecordEntry const* const* ppreEnd = m_pprg + GetCount();

	while( ppreWalk < ppreEnd )
	{
	    if( ( *ppreWalk )->GetUid() == uniqueID )
	    {
		*indexP = ( ppreWalk - m_pprg );
		return 0;
	    }

	    ppreWalk++;
	}

	return dmErrUniqueIDNotFound;
}

static Int32 s_other;
static DmComparF* s_compar;
static Database* s_pdbSort;

static int
cmpQsort( void const* pv1, void const* pv2 )
{
	RecordEntry const* const p1 = *( ( RecordEntry const** ) pv1 );
	RecordEntry const* const p2 = *( ( RecordEntry const** ) pv2 );
	MemPtr const pA = MemHandleLock( p1->m_hData );
	MemPtr const pB = MemHandleLock( p2->m_hData );

	int const cmp = s_compar( pA, pB, s_other );

	MemHandleUnlock( p1->m_hData );
	MemHandleUnlock( p2->m_hData );

	return cmp;
}

Err
RecordList::QuickSort( DmComparF* compar, Int32 other )
{
	if( m_pprg )
	{
	    s_other = other;
	    s_compar = compar;
	    s_pdbSort = m_pdb;

	    qsort( m_pprg, m_c, sizeof( m_pprg[0] ), cmpQsort );
	}

	return 0;
}

int
RecordList::ReadEntries( FileIO* pFile, UInt16 cRecords, IProgress* const pProgress )
{
	int		nRet = -1;
	SysError	e = SE_OK;
	UInt16		ii;
	RecordEntry**	pprg = 0;
	RecordEntry*	p;

	Assert( !m_pprg );
	Assert( !m_c );
	Assert( !m_cAlloc );

	if( cRecords )
	{
	    pprg = ( RecordEntry** )MemAlloc( sizeof( *pprg ) * cRecords _MEM_NOSIZECHECK );

	    if( !pprg )
		goto LError;

	    memset( pprg, 0, sizeof( *pprg ) * cRecords );

	    UInt16 iUpdate = 100;
	    for( ii = 0; ii < cRecords; ii++ )
	    {
		p = new RecordEntry;

		if( !p )
		    goto LError;

		pprg[ii] = p;

		if( !pFile->Read( &p->m_raw, sizeof( p->m_raw ), e ) )
		    goto LError;

		if( pProgress && ii == iUpdate )
		{
		    pProgress->Update( 0 );
		    iUpdate += 100;
		}
	    }
	}

	m_pprg = pprg;			// transfer ownership of array
	pprg = 0;
	m_c = cRecords;
	m_cAlloc = cRecords;

	nRet = 0;			// success

LError:
	FreeArray( &pprg, &cRecords );
	return nRet;
}

int
RecordList::ReadData( FileIO* pFile, UInt32 cbFile, IProgress* const pProgress )
{
	int		nRet = -1;
	SysError	e = SE_OK;
	UInt16		ii;

	Assert( m_pdb );

	for( ii = 0; ii < m_c; ii++ )
	{
	    Boolean		fSuccess = false;
	    MemHandle		h = 0;
	    MemPtr		pv = 0;
	    RecordEntry*	pCur = GetEntry( ii );
	    RecordEntry*	pNext = GetEntry( ii + 1 );
	    unsigned		cb = ( pNext ? pNext->GetOfsData() : cbFile ) - pCur->GetOfsData();

	    Assert( cb < 65000 );

	    if( pProgress )
		pProgress->Update( ii + 1 );

	    if( cb )
	    {
		h = m_pdb->DmNewHandle( cb );
		if( h )
		{
		    pv = MemHandleLock( h );
		    if( pv )
		    {
			if( pFile->Read( pv, cb, e ) )
			{
			    pCur->m_hData = h;
			    fSuccess = true;
			}
		    }
		}
	    }
	    else
	    {
		pCur->m_hData = 0;
		fSuccess = true;
	    }

	    if( pv )
		MemHandleUnlock( h );

	    if( !fSuccess )
	    {
		if( h )
		    MemHandleFree( h );
		goto LError;
	    }
	}

	nRet = 0;			// Success.

LError:
	return nRet;
}

bool
RecordList::WriteEntries( FileIO* pFile, SysError& e, IProgress* const pProgress )
{
	UInt16 iUpdate = 100;
	for( UInt16 ii = 0; ii < m_c; ii++ )
	{
	    RawRecordEntry*	praw = &m_pprg[ii]->m_raw;

	    if( !pFile->Write( praw, sizeof( *praw ), e ) )
		return false;

	    if( pProgress && ii == iUpdate )
	    {
		pProgress->Update( 0 );
		iUpdate += 100;
	    }
	}

	return true;
}

bool
RecordList::WriteData( FileIO* pFile, SysError& e, IProgress* const pProgress )
{
	for( UInt16 ii = 0; ii < m_c; ii++ )
	{
	    MemPtr		pv;
	    UInt32		cb;
	    RecordEntry*	pCur = GetEntry( ii );

	    if( pProgress )
		pProgress->Update( ii + 1 );

	    if( !pFile->Seek( pCur->GetOfsData(), e ) )
		return false;

	    cb = MemHandleSize( pCur->m_hData );
	    pv = MemHandleLock( pCur->m_hData );
	    if( !pv )
	    {
		e = SE_INTERNALERROR;
		return false;
	    }

	    bool const fWrote = pFile->Write( pv, cb, e );

	    MemHandleUnlock( pCur->m_hData );

	    if( !fWrote )
		return false;
	}

	return true;
}

#ifdef DEBUG
bool
RecordList::IsRecordEqual( unsigned ii, RecordList const* precords ) const
{
	Assert( int( ii ) >= 0 );
	Assert( ii < GetCount() );
	Assert( ii < precords->GetCount() );

	bool fEqual = false;

	MemHandle const h1 = m_pprg[ii]->m_hData;
	MemHandle const h2 = precords->m_pprg[ii]->m_hData;

	UInt32 const cb1 = MemHandleSize( h1 );
	UInt32 const cb2 = MemHandleSize( h2 );

	if( cb1 == cb2 )
	{
	    MemPtr p1 = MemHandleLock( h1 );
	    MemPtr p2 = MemHandleLock( h2 );

	    fEqual = !memcmp( p1, p2, cb1 );

	    MemHandleUnlock( h2 );
	    MemHandleUnlock( h1 );
	}

	return fEqual;
}
#endif

/*
 * Database.
 */

Database::Database()
{
	m_hAppInfo = 0;
	m_cOpen = 0;
	m_errLast = 0;

	m_records.SetDatabase( this );
}

Database::~Database()
{
	Assert( !m_cOpen );

	if( m_hAppInfo )
	    MemHandleFree( m_hAppInfo );
}

LocalID
Database::GetAppInfo()
{
	return m_hAppInfo ? MemHandleToLocalID( m_hAppInfo ) : 0;
}

void
Database::SetAppInfo( LocalID lid )
{
	m_hAppInfo = lid ? MemHandle( MemLocalIDToGlobal( lid, 0 ) ) : 0;
}

char const*
Database::GetCategoryStr( UInt16 category )
{
	char const* psz;

	Assert( m_hAppInfo );
	Assert( dmAllCategories != category );
	Assert( dmUnfiledCategory != category );

	Assert( category >= 0 && category < dmRecNumCategories );

	AppInfoPtr pappinfo = AppInfoPtr( MemHandleLock( m_hAppInfo ) );

	// This is only valid until some action that alters the categories.
	psz = pappinfo->categoryLabels[category];

	MemHandleUnlock( m_hAppInfo );

	return psz;
}

Boolean
Database::FLookupCategory( char const* psz, UInt16* pcategory )
{
	Boolean fFound = false;

	Assert( m_hAppInfo );

	AppInfoPtr pappinfo = AppInfoPtr( MemHandleLock( m_hAppInfo ) );

	for( UInt8 ii = 0; ii < DimensionOf( pappinfo->categoryLabels ); ii++ )
	{
	    if( ( psz && pappinfo->categoryLabels[ii][0] && StrCompare( psz, pappinfo->categoryLabels[ii] ) == 0 ) ||
		( !psz && !pappinfo->categoryLabels[ii][0] ) )
	    {
		*pcategory = ii;
		fFound = true;
		break;
	    }
	}

	MemHandleUnlock( m_hAppInfo );

	return fFound;
}

Boolean
Database::CategoryHasRecords( UInt16 category )
{
	for( unsigned ii = m_records.GetCount(); ii--; )
	    if( ( m_records.GetEntry( ii )->GetAttr() & dmRecAttrCategoryMask ) == category )
		return true;

	return false;
}

int
Database::Read( LPCTSTR pszFile, IProgress* const pProgress )
{
	int		nRet = -1;
	unsigned	cbSize = 0;
	MemHandle	hAppInfo = 0;
	MemPtr		pvAppInfo = 0;
	UInt32		cbAppInfo;
	SysError	e = SE_OK;
	SPNew<FileIO>	spFile;
	ScopedProgressHolder sph( pProgress );

	Assert( !m_cOpen );

	m_cOpen++;

	// Open file.

	spFile = FileIO::Create();

	if( !spFile )
	    goto LError;

	spFile->Set( pszFile );
	if( !spFile->Open( FM_READ, e ) )
	    goto LError;

	// Get file size.

	cbSize = spFile->Size( e );
	if( e )
	    goto LError;

	// Read header.

	if( !spFile->Read( &m_header, sizeof( m_header ), e ) )
	    goto LError;

	sph.Init( m_header.NumRecords(), PROGRESS_NOWAIT|PROGRESS_PREDICTIVEWAIT, CAPTION_OPENING );

	// Read record entries.

	if( m_records.ReadEntries( spFile, m_header.NumRecords(), pProgress ) < 0 )
	    goto LError;

	// Read AppInfo.

	if( m_header.GetOfsAppInfo() )
	{
	    // Calculate size of AppInfo block.

	    cbAppInfo = m_header.GetAppInfoSize( m_records.GetEntry( 0 ), cbSize );

	    if( !cbAppInfo )
		goto LError;

	    // Allocate AppInfo block.

	    hAppInfo = DmNewHandle( cbAppInfo );

	    if( !hAppInfo )
		goto LError;

	    pvAppInfo = MemHandleLock( hAppInfo );

	    // Seek to AppInfo data.

	    if( !spFile->Seek( m_header.GetOfsAppInfo(), e ) )
		goto LError;

	    // Read AppInfo data.

	    if( !spFile->Read( pvAppInfo, cbAppInfo, e ) )
		goto LError;
	}

	// Read record data.

	if( m_records.ReadData( spFile, cbSize, pProgress ) < 0 )
	    goto LError;

	// Success, take ownership of appinfo block, and link into list.

	m_hAppInfo = hAppInfo;
	hAppInfo = 0;

	nRet = 0;

LError:
	if( pvAppInfo )
	    MemHandleUnlock( MemPtrRecoverHandle( pvAppInfo ) );
	if( hAppInfo )
	    MemHandleFree( hAppInfo );
	if( nRet < 0 )
	    Reset();
	return nRet;
}

bool
Database::Write( LPCTSTR pszFile, SysError& e, IProgress* const pProgress )
{
	UInt32		ofs = 0;
	UInt32		cbAppInfo;
	int		ii;
	SPNew<FileIO>	spFile;
	ScopedProgressHolder sph( pProgress );

	Assert( m_cOpen );

	// Open file.

	spFile = FileIO::Create();

	if( !spFile )
	{
	    e = SE_OUTOFMEMORY;
	    return false;
	}

	spFile->Set( pszFile );
	if( !spFile->Open( FM_WRITE, e ) )
	    return false;

	sph.Init( m_records.GetCount(), PROGRESS_NOWAIT|PROGRESS_PREDICTIVEWAIT, CAPTION_SAVING );

	// Massage offsets.

	ofs = sizeof( m_header );

	m_header.SetNumRecords( m_records.GetCount() );
	ofs += sizeof( RawRecordEntry ) * m_records.GetCount();
	ofs = ( ( ofs + 3 ) & ~3 );
	Assert( ( ofs % 4 ) == 0 );

	cbAppInfo = m_hAppInfo ? MemHandleSize( m_hAppInfo ) : 0;
	m_header.SetOfsAppInfo( cbAppInfo ? ofs : 0 );
	if( cbAppInfo )
	    ofs += cbAppInfo;

	for( ii = 0; ii < m_records.GetCount(); ii++ )
	{
	    m_records.GetEntry( ii )->SetOfsData( ofs );
	    ofs += m_records.GetEntry( ii )->GetSize();
	    //ofs = ( ( ofs + 1 ) & ~1 );
	}

	// Write header.

	if( !spFile->Write( &m_header, sizeof( m_header ), e ) )
	    return false;

	// Write record entries.

	if( !m_records.WriteEntries( spFile, e, pProgress ) )
	    return false;

	// Write AppInfo.

	if( cbAppInfo )
	{
	    void const* pvAppInfo;

	    // Seek to AppInfo data.

	    if( !spFile->Seek( m_header.GetOfsAppInfo(), e ) )
		return false;

	    // Write AppInfo data.

	    pvAppInfo = MemHandleLock( m_hAppInfo );
	    bool const fWrote = spFile->Write( pvAppInfo, cbAppInfo, e );
	    MemHandleUnlock( m_hAppInfo );
	    if( !fWrote )
		return false;
	}

	// Write record data.

	if( !m_records.WriteData( spFile, e, pProgress ) )
	    return false;

	// Verify the data written matches the data in memory.

#ifdef DEBUG
	{
	    spFile->Close();

	    SPNew<Database> spdb = new Database;
	    int const err = spdb->Read( pszFile );

	    if( err < 0 )
	    {
		Panic1( "VERIFY FAILED:\r\nRead failed (%d).", err );
		spdb->Close();
		e = SE_CORRUPT;
		return false;
	    }

	    if( !IsEqual( spdb ) )
	    {
		StrT s;

		PanicSz( "VERIFY FAILED:\r\nData written does not match data in memory.  Renaming file extension to .xxx for safe keeping..." );

		s.Set( pszFile );
		s.SetLength( s.Length() - 3 );
		s.Append( TEXT( "xxx" ) );

		SysError eTmp;
		spFile->Rename( s.Text(), eTmp );

		spdb->Close();
		e = SE_CORRUPT;
		return false;
	    }

	    spdb->Close();
	}
#endif

	// Success.

	return true;
}

#if 0
unsigned
Database::FindRecord( void const *pRec )
{
	// Brute force search for record; this is for emergency use only!

	for( unsigned ii = 0; ii < m_records.GetCount(); ii++ )
	    if( MemEmergencyPtrFromHandle( m_records.GetEntry( ii )->m_hData ) == pRec )
		return ii;

	return noRecordSelected;
}
#endif

/*
 * DataMgr.
 */

Err
Database::DmDatabaseInfo( Char* nameP, UInt16* attributesP,
			  UInt16* versionP, UInt32* crDateP,
			  UInt32* modDateP, UInt32* bckUpDateP,
			  UInt32* modNumP, LocalID* appInfoIDP,
			  LocalID* sortInfoIDP, UInt32* typeP,
			  UInt32* creatorP )
{
	Assert( !sortInfoIDP );

	SUPPRESS_UNUSED_VAR( sortInfoIDP );

	if( nameP )
	    StrCopy( nameP, m_header.GetName() );
	if( attributesP )
	    *attributesP = m_header.GetAttr();
	if( versionP )
	    *versionP = m_header.GetVersion();
	if( crDateP )
	    *crDateP = m_header.GetCreateDate();
	if( modDateP )
	    *modDateP = m_header.GetModifyDate();
	if( bckUpDateP )
	    *bckUpDateP = m_header.GetBackupDate();
	if( modNumP )
	    *modNumP = m_header.GetModNum();
	if( appInfoIDP )
	    *appInfoIDP = GetAppInfo();
	if( typeP )
	    *typeP = m_header.GetType();
	if( creatorP )
	    *creatorP = m_header.GetCreator();

	return 0;
}

Err
Database::DmSetDatabaseInfo( Char const* nameP, UInt16* attributesP,
			     UInt16* versionP, UInt32* crDateP,
			     UInt32* modDateP, UInt32* bckUpDateP,
			     UInt32* modNumP, LocalID* appInfoIDP,
			     LocalID* sortInfoIDP, UInt32* typeP,
			     UInt32* creatorP )
{
	Assert( !sortInfoIDP );

	SUPPRESS_UNUSED_VAR( sortInfoIDP );

	if( nameP )
	    m_header.SetName( nameP );
	if( attributesP )
	    m_header.SetAttr( *attributesP );
	if( versionP )
	    m_header.SetVersion( *versionP );
	if( crDateP )
	    m_header.SetCreateDate( *crDateP );
	if( modDateP )
	    m_header.SetModifyDate( *modDateP );
	if( bckUpDateP )
	    m_header.SetBackupDate( *bckUpDateP );
	if( modNumP )
	    m_header.SetModNum( *modNumP );
	if( appInfoIDP )
	    SetAppInfo( *appInfoIDP );
	if( typeP )
	    m_header.SetType( *typeP );
	if( creatorP )
	    m_header.SetCreator( *creatorP );

	return 0;
}

Err
Database::DmOpenDatabaseInfo( UInt16* openCountP, UInt16* modeP, Boolean* resDBP )
{
	Assert( m_cOpen );

	Assert( !modeP );
	Assert( !resDBP );

	SUPPRESS_UNUSED_VAR( modeP );
	SUPPRESS_UNUSED_VAR( resDBP );

	if( openCountP )
	    *openCountP = UInt16( m_cOpen );

	return 0;
}

void
Database::Initialize( Char const* nameP, UInt32 creator,
		      UInt32 type, Boolean resDB )
{
	Assert( nameP );
	Assert( creator );
	Assert( type );
	Assert( !resDB );

	SUPPRESS_UNUSED_VAR( resDB );

	Assert( !m_cOpen );

	Reset();

	m_cOpen++;

	SetDbName( nameP );
	SetCreator(creator);
	SetType(type);

	time_t timeNow;

	timeNow = __localtime();	// Seconds since 1970/01/01 00:00:00.
	timeNow += c_UnixToPalm;	// Convert to    1904/01/01 00:00:00.

	CASSERT( sizeof( time_t ) == sizeof( UInt32 ) );

	m_header.SetCreateDate( timeNow );
	m_header.SetModifyDate( timeNow );
}

void
Database::Reset()
{
	m_header.Reset();
	m_records.Reset();
}

void
Database::Close()
{
	Assert( m_cOpen );

	m_cOpen--;
}

Err
Database::DmRecordInfo( unsigned index, UInt16* attrP, UInt32* uniqueIDP, LocalID* chunkIDP )
{
	Assert( m_cOpen );
	Assert( !chunkIDP );

	SUPPRESS_UNUSED_VAR( chunkIDP );

	RecordEntry const* p = m_records.GetEntry( index );

	if( !p )
	{
	    Panic1( "'index' out of range (%u)", index );
	    return dmErrIndexOutOfRange;
	}

	if( attrP )
	    *attrP = p->GetAttr();
	if( uniqueIDP )
	    *uniqueIDP = p->GetUid();

	return 0;
}

Err
Database::DmSetRecordInfo( unsigned index, UInt16* attrP, UInt32* uniqueIDP )
{
	Assert( m_cOpen );

	RecordEntry* p = m_records.GetEntry( index );

	if( !p )
	{
	    Panic1( "'index' out of range (%u)", index );
	    return dmErrIndexOutOfRange;
	}

	if( attrP )
	    p->SetAttr( UChar( ( *attrP ) & 0xff ) );
	if( uniqueIDP )
	    p->SetUid( *uniqueIDP );

	return 0;
}

Err
Database::DmAttachRecord( unsigned* atP, MemHandle newH, MemHandle* oldHP )
{
	Assert( m_cOpen );
	Assert( atP );
	Assert( newH );

	if( *atP > m_records.GetCount() )
	    *atP = m_records.GetCount();

	if( !oldHP )
	    return m_records.Insert( *atP, newH );

	if( *atP >= m_records.GetCount() )
	    return dmErrIndexOutOfRange;

	*oldHP = m_records.GetEntry( *atP )->m_hData;

	m_records.GetEntry( *atP )->m_hData = newH;

	return 0;
}

Err
Database::DmDetachRecord( unsigned index, MemHandle* oldHP )
{
	Assert( m_cOpen );
	return m_records.Remove( index, oldHP );
}

Err
Database::DmMoveRecord( unsigned from, unsigned to )
{
	Assert( m_cOpen );

	Err err = m_records.Move( from, to );

	if( !err )
	    m_header.IncModNumber();

	return err;
}

MemHandle
Database::DmNewRecord( unsigned* atP, UInt32 size )
{
	Assert( m_cOpen );

	Err err = 0;
	MemHandle h = 0;

	h = DmNewHandle( size );
	if( h )
	{
	    err = DmAttachRecord( atP, h, 0 );
	    if( !err )
	    {
		RecordEntry* p;

		p = m_records.GetEntry( *atP );
		Assert( p );

		p->SetAttr( p->GetAttr() | dmRecAttrBusy | dmRecAttrDirty );

		m_header.IncModNumber();
	    }
	    else
		Assert( err == memErrNotEnoughSpace || err == dmErrMemError );
	}
	else
	    err = dmErrMemError;

	m_errLast = err;
	return h;
}

Err
Database::DmRemoveRecord( unsigned index )
{
	Assert( m_cOpen );

	Err err;
	MemHandle h;

	err = DmDetachRecord( index, &h );
	if( err )
	    return err;

	if( h )
	    err = MemHandleFree( h );

	m_header.IncModNumber();

	return err;
}

MemHandle
Database::DmNewHandle( UInt32 size )
{
	Assert( m_cOpen );

	m_errLast = 0;

	if( !size )
	{
	    m_errLast = dmErrInvalidParam;
	    return 0;
	}

	MemHandle const h = MemHandleNew( size );

	if( !h )
	    m_errLast = dmErrMemError;

	return h;
}

Err
Database::DmFindRecordByID( UInt32 uniqueID, unsigned* indexP )
{
	m_errLast = m_records.FindByID( uniqueID, indexP );

	return m_errLast;
}

MemHandle
Database::DmQueryRecord( unsigned index )
{
	Assert( m_cOpen );

	m_errLast = 0;

	if( index >= m_records.GetCount() )
	{
	    m_errLast = dmErrIndexOutOfRange;
	    return 0;
	}

	return m_records.GetEntry( index )->m_hData;
}

MemHandle
Database::DmGetRecord( unsigned index )
{
	Assert( m_cOpen );

	m_errLast = 0;

	if( index >= m_records.GetCount() )
	{
	    m_errLast = dmErrIndexOutOfRange;
	    return 0;
	}

	RecordEntry* const p = m_records.GetEntry( index );
	Assert( p );

	if( p->GetAttr() & dmRecAttrDelete )
	{
	    m_errLast = dmErrRecordDeleted;
	    return 0;
	}

	if( p->GetAttr() & dmRecAttrBusy )
	{
	    m_errLast = dmErrRecordBusy;
	    return 0;
	}

	p->SetAttr( p->GetAttr() | dmRecAttrBusy );

	return p->m_hData;
}

MemHandle
Database::DmResizeRecord( unsigned index, UInt32 newSize )
{
	Assert( m_cOpen );

	m_errLast = MemHandleResize( m_records.GetEntry( index )->m_hData, newSize );
	return m_records.GetEntry( index )->m_hData;
}

Err
Database::DmReleaseRecord( unsigned index, Boolean dirty )
{
	Assert( m_cOpen );

	RecordEntry* const p = m_records.GetEntry( index );

	if( !p )
	{
	    Panic1( "'index' out of range (%u)", index );
	    return dmErrIndexOutOfRange;
	}

	Assert( p->GetAttr() & dmRecAttrBusy );
	p->SetAttr( p->GetAttr() & ~dmRecAttrBusy );

	if( dirty )
	{
	    p->SetAttr( p->GetAttr() | dmRecAttrDirty );
	    m_header.IncModNumber();
	}

	return 0;
}

#if 0
Err
Database::DmMoveCategory( UInt16 toCategory, UInt16 fromCategory, Boolean dirty )
{
	for( unsigned ii = m_records.GetCount(); ii--; )
	{
	    RecordEntry* const pre = m_records.GetEntry( ii );

	    if( ( pre->GetAttr() & dmRecAttrCategoryMask ) == fromCategory )
	    {
		UInt8 attr = pre->GetAttr() & ~dmRecAttrCategoryMask;

		if( dirty )
		    attr |= dmRecAttrDirty;

		pre->SetAttr( attr | UInt8( toCategory ) );
	    }
	}

	m_header.IncModNumber();

	return 0;
}
#endif

Err
Database::DmQuickSort( DmComparF* compar, Int32 other )
{
	Assert( m_cOpen );

	return m_records.QuickSort( compar, other );
}

unsigned
Database::DmFindSortPosition( void* newRecord, DmComparF* compar, Int32 other )
{
	Assert( m_cOpen );

	unsigned	cRecords;
	MemHandle	hRecord;
	MemPtr		r;
	unsigned	kmin;		// Position in the database.
	unsigned	probe;		// Position in the database.
	unsigned	ii;		// Position in the database.
	Int16		result;		// Result of comparing two records.
	Boolean		binarySearch = true;

	result = -1;			// For the case when there aren't any records.
	kmin = probe = 0;
	cRecords = m_records.GetCount();
	hRecord = 0;

	while( cRecords > 0 )
	{
	    ii = cRecords / 2;
	    probe = kmin + ii;

	    // Compare the two records.  Treat deleted records as greater.
	    // If the records are equal look at the following position.

	    do
	    {
		if( hRecord )
		    MemHandleUnlock( hRecord );

		hRecord = DmQueryRecord( probe );
		if( !hRecord )
		{
		    if( binarySearch )
		    {
			// Delete record is greater.
			result = -1;
			break;
		    }
		    else
		    {
			// MemHandle incremental search past the end of normal
			// data (into deleted records or off end of the
			// database).  Return position after last normal
			// record.
			return probe;
		    }
		}
		else
		{
		    r = MemHandleLock( hRecord );

		    result = compar( newRecord, r, other );

		    // If equal look at the following position.

		    if( result == 0 )
		    {
			binarySearch = false;
			probe++;
		    }
		    else
		    {
			// If we've stopped binary searching then we've found 
			// our position.

			if( !binarySearch )
			{
			    MemHandleUnlock( hRecord );
			    return probe;
			}
			break;
		    }
		}
	    }
	    while( true );

	    Assert( result );

	    // More likely than < 0 because of deleted records.

	    if( result > 0 )
	    {
		kmin = probe + 1;
		cRecords = cRecords - ii - 1;
	    }
	    else
	    {
		cRecords = ii;
	    }
	}

	if( hRecord )
	    MemHandleUnlock( hRecord );
	return ( result < 0 ) ? probe : probe + 1;
}

#ifdef DEBUG
bool
Database::IsEqual( Database const* pdb ) const
{
	CASSERT( sizeof( m_header ) == 78 );

	// Compare whether the headers match.

	if( memcmp( &m_header, &pdb->m_header, sizeof( m_header ) ) )
	    return false;

	// Compare whether the records match.  The number of records must be
	// the same if the headers matched.

	Assert( DmNumRecords() == pdb->DmNumRecords() );

	ScopedProgressHolder sph;
	sph.Create();
	sph.Init( m_records.GetCount(), PROGRESS_NOWAIT|PROGRESS_PREDICTIVEWAIT, CAPTION_VERIFYING );

	for( unsigned ii = 0; ii < DmNumRecords(); ii++ )
	{
	    sph.Update( ii + 1 );

	    if( !m_records.IsRecordEqual( ii, &pdb->m_records ) )
		return false;
	}

	// The databases seem to match.

	return true;
}
#endif

#endif // !OS_PALM

