/*
 * _pdb.h - Database class to abstract access to a Palm .PDB database file.
 */

#pragma once

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

void AssertNoRecordLocks( DmOpenRef pdb, LPCTSTR pszContext );

#ifndef OS_PALM

/*
 * OutInterface.
 */

#define PURE = 0
#define IMPL

#define DeclareOutInterfaceMembers( IPURE ) \
	virtual int	SetOutputFilename( char const* pszOutFile ) IPURE; \
	virtual int	Begin( Database* pdb ) IPURE; \
	virtual int	OutputCategory( Database* pdb, char const* pszCategory, UInt8 id, UInt8 index ) IPURE; \
	virtual int	OutputAppInfo( Database* pdb, AppInfoPtr pappinfo, UInt32 cb ) IPURE; \
	virtual int	OutputStore( Database* pdb, const char* pszStore, UInt8 id ) IPURE; \
	virtual int	OutputItem( Database* pdb, ItemPtr pitem ) IPURE; \
	virtual int	Finished( Database* pdb ) IPURE; \

class Item;
typedef Item ItemType;
typedef Item* ItemPtr;

class OutInterface
{
    public:
	DeclareOutInterfaceMembers( PURE );
};

/*
 * PdbHeader.
 */

#pragma pack( push, 1 )

struct RecordEntry;
class FileIO;

class PdbHeader
{
    public:
			PdbHeader();

	void		Reset() { memset( this, 0, sizeof( *this ) ); }

	void		IncModNumber() { m_nModification = MacDWord( MacDWord( m_nModification ) + 1 ); }
	UInt32		GetNextUid();

	UInt16		NumRecords() const { return MacWord( m_cRecords ); }
	UInt32		GetOfsAppInfo() const { return MacDWord( m_ofsAppInfo ); }
	UInt32		GetOfsSortInfo() const { return MacDWord( m_ofsSortInfo ); }
	const Char*	GetName() const { return &m_szName[0]; }
	UInt32		GetType() const { return MacDWord( m_type ); }
	UInt32		GetCreator() const { return MacDWord( m_creator ); }
	UInt16		GetAttr() const { return MacWord( m_attr ); }
	UInt16		GetVersion() const { return MacWord( m_ver ); }
	UInt32		GetCreateDate() const { return MacDWord( m_dateCreate ); }
	UInt32		GetModifyDate() const { return MacDWord( m_dateModify ); }
	UInt32		GetBackupDate() const { return MacDWord( m_dateBackup ); }
	UInt32		GetModNum() const { return MacDWord( m_nModification ); }

	UInt32		GetAppInfoSize( const RecordEntry* pRecordZero, UInt32 cbFileSize ) const;

	void		SetNumRecords( UInt16 c ) { m_cRecords = MacWord( c ); }
	void		SetOfsAppInfo( UInt32 ofs ) { m_ofsAppInfo = MacDWord( ofs ); }
	void		SetName( const Char* pszName ) { Assert( StrLen( pszName ) < DimensionOf( m_szName ) ); StrCopy( m_szName, pszName ); }
	void		SetType( UInt32 type ) { m_type = MacDWord( type ); }
	void		SetCreator( UInt32 creator ) { m_creator = MacDWord( creator ); }
	void		SetAttr( UInt16 attr ) { m_attr = MacWord( attr ); }
	void		SetVersion( UInt16 ver ) { m_ver = MacWord( ver ); }
	void		SetCreateDate( UInt32 dateCreate ) { m_dateCreate = MacDWord( dateCreate ); }
	void		SetModifyDate( UInt32 dateModify ) { m_dateModify = MacDWord( dateModify ); }
	void		SetBackupDate( UInt32 dateBackup ) { m_dateBackup = MacDWord( dateBackup ); }
	void		SetModNum( UInt32 nModification ) { m_nModification = MacDWord( nModification ); }

    private:
	Char		m_szName[32];		// name of database on device
	UInt16		m_attr;			// database attributes
	UInt16		m_ver;			// version number
	UInt32		m_dateCreate;		// creation date
	UInt32		m_dateModify;		// modification date
	UInt32		m_dateBackup;		// last backup date
	UInt32		m_nModification;	// modification number
	UInt32		m_ofsAppInfo;		// offset to AppInfo (from 0)
	UInt32		m_ofsSortInfo;		// RESERVED: must be 0
	UInt32		m_type;			// database type
	UInt32		m_creator;		// database creator id
	UInt32		m_ulSeed;		// unique id seed
	UInt32		m_idNextRecordList;	// RESERVED: must be 0
	UInt16		m_cRecords;		// number of records
};
CASSERT( sizeof( PdbHeader ) == 78 );

#pragma pack( pop )

/*
 * RawRecordEntry.
 */

struct RawRecordEntry
{
    public:
	UInt32		GetOfsData() const { return MacDWord( m_ofsData ); }
	UChar		GetAttr() const { return m_attr; }
	UInt32		GetUid() const;

	void		SetOfsData( UInt32 ofs ) { m_ofsData = MacDWord( ofs ); }
	void		SetAttr( UChar attr ) { m_attr = attr; }
	void		SetUid( UInt32 uid );

    private:
	UInt32		m_ofsData;		// offset to data (from 0)
	UChar		m_attr;			// record attributes
	UChar		m_uid[3];		// unique id
};
CASSERT( sizeof( RawRecordEntry ) == 8 );

/*
 * RecordEntry.
 */

struct RecordEntry
{
			~RecordEntry() { MemHandleFree( m_hData ); }

	UInt32		GetOfsData() const { return m_raw.GetOfsData(); }
	UChar		GetAttr() const { return m_raw.GetAttr(); }
	UInt32		GetUid() const { return m_raw.GetUid(); }
	UInt32		GetSize() const;

	void		SetOfsData( UInt32 ofs ) { m_raw.SetOfsData( ofs ); }
	void		SetAttr( UChar attr ) { m_raw.SetAttr( attr ); }
	void		SetUid( UInt32 uid ) { m_raw.SetUid( uid ); }

	RawRecordEntry	m_raw;
	MemHandle	m_hData;
};

/*
 * RecordList.
 */

struct ProgressData;

class RecordList
{
    public:
			RecordList();
			~RecordList();

	void		SetDatabase( Database* pdb ) { m_pdb = pdb; }

	void		Reset() { FreeArray( &m_pprg, &m_c ); m_cAlloc = 0; }

	UInt16		GetCount() const { return m_c; }
	RecordEntry*	GetEntry( unsigned ii );

	Err		Insert( unsigned ii, MemHandle h );
	Err		Move( unsigned from, unsigned to );
	Err		Remove( unsigned ii, MemHandle* phOld );

	Err		FindByID( UInt32 uniqueID, unsigned* indexP );

	Err		QuickSort( DmComparF* compar, Int32 other );

	// File access.

	int		ReadEntries( FileIO* pFile, UInt16 cRecords, IProgress* pProgress = 0 );
	int		ReadData( FileIO* pFile, UInt32 cbFile, IProgress* pProgress = 0 );

	bool		WriteEntries( FileIO* pFile, SysError& e, IProgress* pProgress = 0 );
	bool		WriteData( FileIO* pFile, SysError& e, IProgress* pProgress = 0 );

	// Diagnostic.

#ifdef DEBUG
	bool		IsRecordEqual( unsigned ii, RecordList const* precords ) const;
#endif

    private:
	void		FreeArray( RecordEntry*** ppprg, UInt16* pc );

    private:
	Database*	m_pdb;
	RecordEntry**	m_pprg;
	UInt16		m_c;
	UInt16		m_cAlloc;
};

inline RecordEntry*
RecordList::GetEntry( unsigned ii )
{
	return ( int( ii ) >= 0 && ii < m_c ) ? m_pprg[ii] : 0;
}

/*
 * Database.
 */

class Database
{
    public:
			Database();
	virtual		~Database();

	void		Initialize( Char const* nameP, UInt32 creator,
				    UInt32 type, Boolean resDB );

	int		Read( LPCTSTR pszFile, IProgress* pProgress = 0 );
	bool		Write( LPCTSTR pszFile, SysError& e, IProgress* pProgress = 0 );
	void		Close();

	const Char*	GetDbName() { return m_header.GetName(); }
	UInt32		GetNextUid() { return m_header.GetNextUid(); }

	PdbHeader const* GetHeader() const { return &m_header; }

	void		SetDbName( Char const* psz ) { m_header.SetName( psz ); }
	void		SetType( UInt32 type ) { m_header.SetType( type ); }
	void		SetCreator( UInt32 creator ) { m_header.SetCreator( creator ); }

	LocalID		GetAppInfo();
	void		SetAppInfo( LocalID lid );
	const char*	GetCategoryStr( UInt16 category );
	Boolean		FLookupCategory( char const* psz, UInt16 *pcategory );
	Boolean		CategoryHasRecords( UInt16 category );

	RecordEntry*	GetEntry( unsigned index ) { return m_records.GetEntry( UInt16( index ) ); }

#if 0
	unsigned	FindRecord( void const *pRec );
#endif

	// DataMgr.

	Err		DmGetLastErr() { return m_errLast; }

	Err		DmDatabaseInfo( Char* nameP, UInt16* attributesP,
					UInt16* versionP, UInt32* crDateP,
					UInt32* modDateP, UInt32* bckUpDateP,
					UInt32* modNumP, LocalID* appInfoIDP,
					LocalID* sortInfoIDP, UInt32* typeP,
					UInt32* creatorP );
	Err		DmSetDatabaseInfo( Char const* nameP, UInt16* attributesP,
					   UInt16* versionP, UInt32* crDateP,
					   UInt32* modDateP, UInt32* bckUpDateP,
					   UInt32* modNumP, LocalID* appInfoIDP,
					   LocalID* sortInfoIDP, UInt32* typeP,
					   UInt32* creatorP );
	Err		DmOpenDatabaseInfo( UInt16* openCountP, UInt16* modeP,
					    Boolean* resDBP );

	unsigned	DmNumRecords() const { return m_records.GetCount(); }

	Err		DmRecordInfo( unsigned index, UInt16* attrP, UInt32* uniqueIDP, LocalID* chunkIDP );
	Err		DmSetRecordInfo( unsigned index, UInt16* attrP, UInt32* uniqueIDP );

	Err		DmAttachRecord( unsigned* atP, MemHandle newH, MemHandle* oldHP );
	Err		DmDetachRecord( unsigned index, MemHandle* oldHP );
	Err		DmMoveRecord( unsigned from, unsigned to );

	MemHandle	DmNewRecord( unsigned* atP, UInt32 size );
	Err		DmRemoveRecord( unsigned index );
	MemHandle	DmNewHandle( UInt32 size );

	Err		DmFindRecordByID( UInt32 uniqueID, unsigned* indexP );

	MemHandle	DmQueryRecord( unsigned index );
	MemHandle	DmGetRecord( unsigned index );
	MemHandle	DmResizeRecord( unsigned index, UInt32 newSize );
	Err		DmReleaseRecord( unsigned index, Boolean dirty );

#if 0
	Err		DmMoveCategory( UInt16 toCategory, UInt16 fromCategory, Boolean dirty );
#endif

	Err		DmQuickSort( DmComparF* compar, Int32 other );
	unsigned	DmFindSortPosition( void* newRecord, DmComparF* compar, Int32 other );

    protected:
	void		Reset();

#ifdef DEBUG
	bool		IsEqual( Database const* pdb ) const;
#endif

    private:
	PdbHeader	m_header;
	RecordList	m_records;

	MemHandle	m_hAppInfo;

	int		m_cOpen;		// Open count.
	Err		m_errLast;		// Last error.

	Database*	m_pNext;
	Database*	m_pPrev;
};

#endif // !OS_PALM

