/*
 * _types.h - Palm compatibility data types.
 */

#pragma once

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

/*
 * Palm OS compatibility data types.
 */

#ifndef OS_PALM

typedef signed char Boolean;

typedef char Char;
typedef unsigned char UChar;
CASSERT( sizeof( Char ) == 1 );
CASSERT( sizeof( UChar ) == 1 );

typedef signed char Int8;
typedef unsigned char UInt8;
CASSERT( sizeof( Int8 ) == 1 );
CASSERT( sizeof( UInt8 ) == 1 );

typedef signed short Int16;
typedef unsigned short UInt16;
CASSERT( sizeof( Int16 ) == 2 );
CASSERT( sizeof( UInt16 ) == 2 );

typedef signed int Int32;
typedef unsigned int UInt32;
CASSERT( sizeof( Int32 ) == 4 );
CASSERT( sizeof( UInt32 ) == 4 );

typedef UInt16 Err;

typedef UInt32 LocalID;			// Local (card relative) chunk ID.

typedef void* MemPtr;			// Global pointer.
typedef struct _opaque* MemHandle;	// Global handle.

class Database;
typedef Database* DmOpenRef;

UInt32 const c_UnixToPalm = 2082844800;	// Difference in seconds between 1970 and 1904.

#define firstYear		1904
#define numberOfYears		128
#define lastYear		( firstYear + numberOfYears - 1 )

struct DateTimeType
{
	Int16 second;
	Int16 minute;
	Int16 hour;
	Int16 day;
	Int16 month;
	Int16 year;
	Int16 weekDay;		// Days since Sunday (0 to 6)
};

typedef DateTimeType* DateTimePtr;

struct TimeType
{
	UInt8 hours;
	UInt8 minutes;
};

typedef TimeType* TimePtr;

#define noTime	-1		// The entire TimeType is -1 if there isn't a time.

struct PalmDateType
{
#ifdef OS_MAC
	UInt16 year	:7;	// Years since 1904 (MAC format).
	UInt16 month	:4; 
	UInt16 day	:5;
#else
	UInt16 day	:5;
	UInt16 month	:4; 
	UInt16 year	:7;	// Years since 1904 (MAC format).
#endif
};

typedef PalmDateType* PalmDatePtr;

// Convert a date in a DateType structure to an UInt16.
#define DateToInt( date )	( *( UInt16* ) &date )

#endif // !OS_PALM

/*
 * Byte Order translations.
 */

#if defined( OS_MAC ) || defined( OS_PALM )

#define MacWord( x )	( x )
#define MacDWord( x )	( x )

#else // !( OS_MAC || OS_PALM )

#define MacWord( x )	( ( ( ( x ) & 0xff ) << 8 ) | ( ( ( x ) & 0xff00 ) >> 8 ) )
#define MacDWord( x )	( ( ( ( x ) & 0xff000000 ) >> 24 ) | \
			  ( ( ( x ) & 0x00ff0000 ) >> 8 ) | \
			  ( ( ( x ) & 0x0000ff00 ) << 8 ) | \
			  ( ( ( x ) & 0x000000ff ) << 24 ) )

#endif // !( OS_MAC || OS_PALM )

inline void
CopyMacWord( void* to, void* from )
{
	*( UInt16* )to = MacWord( *( UInt16* )from );
}

inline UInt16
AddMacWord( UInt16 x, Int16 y )
{
	x = MacWord( x );
	x = x + UInt16( y );
	return MacWord( x );
}

inline UInt16
OrMacWord( UInt16 x, UInt16 y )
{
	x = MacWord( x );
	x |= y;
	return MacWord( x );
}

inline UInt16
AndMacWord( UInt16 x, UInt16 y )
{
	x = MacWord( x );
	x &= y;
	return MacWord( x );
}

#define IMPL_GET_WORD( var )	UInt16 get_##var() const { return MacWord( var ); }
#define IMPL_SET_WORD( var )	void set_##var( UInt16 x ) { var = MacWord( x ); }
#define IMPL_ADD_WORD( var )	void add_##var( Int16 x ) { var = AddMacWord( var, x ); }
#define IMPL_OR_WORD( var )	void or_##var( UInt16 x ) { var = OrMacWord( var, x ); }
#define IMPL_AND_WORD( var )	void and_##var( UInt16 x ) { var = AndMacWord( var, x ); }

#define IMPL_WORD( var ) \
	IMPL_GET_WORD( var ) \
	IMPL_SET_WORD( var ) \
	IMPL_ADD_WORD( var ) \

inline void
CopyMacDWord( void* to, void* from )
{
	UInt32 tmp;

	memcpy( &tmp, from, sizeof( tmp ) );
	tmp = MacDWord( tmp );
	memcpy( to, &tmp, sizeof( tmp ) );
}

inline UInt32
AddMacDWord( UInt32 x, Int32 y )
{
	x = MacDWord( x );
	x += y;
	return MacDWord( x );
}

inline UInt32 OrMacDWord( UInt32 x, UInt32 y )
{
	x = MacDWord( x );
	x |= y;
	return MacDWord( x );
}

#define IMPL_GET_DWORD( var )	UInt32 get_##var() const { return MacDWord( var ); }
#define IMPL_SET_DWORD( var )	void set_##var( UInt32 x ) { var = MacDWord( x ); }
#define IMPL_ADD_DWORD( var )	void add_##var( Int32 x ) { var = AddMacDWord( var, x ); }
#define IMPL_OR_DWORD( var )	void or_##var( UInt32 x ) { var = OrMacDWord( var, x ); }

#define IMPL_DWORD( var ) \
	IMPL_GET_DWORD( var ) \
	IMPL_SET_DWORD( var ) \
	IMPL_ADD_DWORD( var ) \

#define IMPL_ZERO( var )	void zero_##var() { memset( &var, 0, sizeof( var ) ); }

