/*
 * DbOpen2.cpp - Database implementation for HS2 Palm based databases.
 */

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#include "precomp.h"

#include "str.h"
#include "sys.h"
#include "storelist.h"
#include "item.h"
#include "dbopen.h"
#include "number.h"
#include "BitTable.h"
#include "palmlayer.h"

#include "globalhs2.h"
#include "columns.h"
#include "data1x.h"
#include "storeshs2.h"
#include "db.h"
#include "pdbhs2.h"

#ifdef _WIN32_WCE
#include "apppref.h"
#include "pref.h"
#endif
#include "progress.h"

#ifdef DEBUG
//#define USE_TRACEDB
//#define USE_TRACECOUNTS
//#define USE_TRACETAG
#endif

#ifdef USE_TRACEDB
	#define TraceDB		DbgMsg
#else
	#define TraceDB( x )
#endif
#ifdef USE_TRACECOUNTS
	#undef TraceCounts
	#define TraceCounts DbgMsg
#else
	#define TraceCounts(x)
#endif
#ifdef USE_TRACETAG
	#define TraceTAG	DbgMsg
#else
	#define TraceTAG( x )
#endif

/*
 * Misc.
 */

CASSERT( storeAllStores == allStores );
CASSERT( storeUnfiledStore == noStore );

#define CheckLocks( pdb, psz, ul )		syntaxsafe
#define CheckAppInfoLocks( pdb, max, psz, ul )	syntaxsafe

inline void
__itoa( int x, char* pszOut, size_t cchMax, int radix )
{
#if defined( OS_WINDOWSMOBILE )
	SUPPRESS_UNUSED_VAR( cchMax );
	_itoa( x, pszOut, radix );
#else
	_itoa_s( x, pszOut, cchMax, radix );
#endif
}

void
SkipPrefix( Char const** ppsz )
{
	if( MemCmp( *ppsz, "HS2_", 4 ) == 0 )
	    ( *ppsz ) += 4;
}

#ifdef UNICODE
void
SkipPrefix( const TCHAR** ppsz )
{
	if( _tcsncmp( *ppsz, TEXT( "HS2_" ), 4 ) == 0 )
	    ( *ppsz ) += 4;
}
#endif

/*
 * Convert category representations.
 *
 * HS3 expects categories to be in 0..N-1, and has special values for "All"
 * and "Unfiled".  However HS2 and the Palm database format expect categories
 * to be in 1..N and use 0 for "Unfiled", and have a special value for "All".
 *
 * These functions convert between the HS3 category representation and the
 * Palm data manager category representation.
 */

UInt16
FromDmCategory( UInt16 const category )
{
	if( category == dmAllCategories )
	    return catAllCategories;
	if( category == dmUnfiledCategory )
	    return catUnfiledCategory;
	return category - 1;
}

UInt16
ToDmCategory( UInt16 const category )
{
	if( category == catAllCategories )
	    return dmAllCategories;
	if( category == catUnfiledCategory )
	    return dmUnfiledCategory;
	return category + 1;
}

/*
 * DatabasePrefs methods.
 */

void
DatabasePrefs::SetMultipleCategories( MultipleCategories const* const pmulti )
{
	Assert( pmulti );

	m_fMultipleCategories_Unfiled = !!( pmulti->m_Unfiled );

	// Subtle:  initializing cat to catAllCategories might seem
	// unnecessary because the loop below will always set it when there
	// are multiple categories.  But it also needs to be catAllCategories
	// when pmulti is completely empty.

	unsigned cMultiple = !!m_fMultipleCategories_Unfiled;
	WORD cat = WORD( m_fMultipleCategories_Unfiled ? catUnfiledCategory : catAllCategories );

	// Set the bitflag array of categories.

	ZeroMemory( &m_rggrfMultipleCategories, sizeof( m_rggrfMultipleCategories ) );

	CASSERT( c_cMaxCategories <= DimensionOf( pmulti->m_rg ) );

	for( unsigned ii = 0; ii < c_cMaxCategories; ii++ )
	    if( pmulti->m_rg[ii] )
	    {
		unsigned const index = ii / 8;
		BYTE const mask = BYTE( 1 << ( ii % 8 ) );

		Assert( mask );

		m_rggrfMultipleCategories[index] |= mask;

		cMultiple++;
		cat = ( cMultiple > 1 ) ? catAllCategories : pmulti->m_rg[ii];
	    }

	// Set the other flags appropriately.

	m_fMultipleCategories = !!cMultiple;
	category = cat;
}

Boolean
DatabasePrefs::GetMultipleCategories( MultipleCategories* const pmulti ) const
{
	if( pmulti )
	{
	    pmulti->SetAll( false );
	    pmulti->m_Unfiled = !!m_fMultipleCategories_Unfiled;

	    CASSERT( c_cMaxCategories <= DimensionOf( pmulti->m_rg ) );

	    for( unsigned ii = 0; ii < c_cMaxCategories; ii++ )
	    {
		unsigned const index = ii / 8;
		BYTE const mask = BYTE( 1 << ( ii % 8 ) );

		Assert( mask );

		if( m_rggrfMultipleCategories[index] & mask )
		    pmulti->m_rg[ii] = true;
	    }
	}

	return m_fMultipleCategories;
}

void
DatabasePrefs::SetMultipleStores( MultipleStores const* const pmulti )
{
	Assert( pmulti );

	m_fMultipleStores_Unfiled = !!( pmulti->m_Unfiled );

	// Subtle:  initializing store to storeAllStores might seem
	// unnecessary because the loop below will always set it when there
	// are multiple stores.  But it also needs to be storeAllStores when
	// pmulti is completely empty.

	unsigned cMultiple = !!m_fMultipleStores_Unfiled;
	BYTE store = BYTE( m_fMultipleStores_Unfiled ? storeUnfiledStore : storeAllStores );

	// Set the bitflag array of stores.

	ZeroMemory( &m_rggrfMultipleStores, sizeof( m_rggrfMultipleStores ) );

	CASSERT( c_cMaxStores <= DimensionOf( pmulti->m_rg ) );

	for( unsigned ii = 0; ii < c_cMaxStores; ii++ )
	    if( pmulti->m_rg[ii] )
	    {
		unsigned const index = ii / 8;
		BYTE const mask = BYTE( 1 << ( ii % 8 ) );

		Assert( mask );

		m_rggrfMultipleStores[index] |= mask;

		cMultiple++;
		store = ( cMultiple > 1 ) ? storeAllStores : pmulti->m_rg[ii];
	    }

	// Set the other flags appropriately.

	m_fMultipleStores = !!cMultiple;
	idStore = store;
}

Boolean
DatabasePrefs::GetMultipleStores( MultipleStores* const pmulti ) const
{
	if( pmulti )
	{
	    pmulti->SetAll( false );
	    pmulti->m_Unfiled = !!m_fMultipleStores_Unfiled;

	    CASSERT( c_cMaxStores <= DimensionOf( pmulti->m_rg ) );

	    for( unsigned ii = 0; ii < c_cMaxStores; ii++ )
	    {
		unsigned const index = ii / 8;
		BYTE const mask = BYTE( 1 << ( ii % 8 ) );

		Assert( mask );

		if( m_rggrfMultipleStores[index] & mask )
		    pmulti->m_rg[ii] = true;
	    }
	}

	return m_fMultipleStores;
}

/*
 * Item helpers and accessors.
 */

UInt16
DbOpenHs2::GetItemFlags( unsigned record, Err* perr ) const
{
	MemHandle h;
	UInt8* p;
	UInt16 flags;

	h = DmQueryRecord( m_pdb, record );

	if( perr )
	{
	    //*perr = dmErrRecordDeleted;
	    *perr = m_pdb->DmGetLastErr();
	}

	Assert2( h, "GetItemFlags\n  DmQueryRecord\n\nrecord == %u\nerr == %x",
		 record, m_pdb->DmGetLastErr() );

	if( !h )
	    return 0;

	p = ( UInt8* )MemHandleLock( h );

	if( tagFlags == *p )
	{
	    p++;
	    CASSERT( SameSize( flags, ( ( ItemPtr )0 )->m_flags ) );
	    flags = ( p[0] << 8 ) | p[1];
	}
	else
	{
	    Assert( tagMasterStoreList == *p || tagCategoryList == *p );
	    flags = flagMetadataRecord;
	}

	MemHandleUnlock( h );

	CheckLocks( m_pdb, __FILE__, __LINE__ );

	return flags;
}

/*
 * Xfer helpers.
 */

enum HS2XferMode { xferSkip, xferLoad, xferSave };

inline HS2XferMode
operator|( HS2XferMode const a, HS2XferMode const b )
{
	return HS2XferMode( UInt8( a ) | UInt8( b ) );
}

inline HS2XferMode
operator&=( HS2XferMode &me, HS2XferMode const &mask )
{
	return me = HS2XferMode( UInt8( me ) & UInt8( mask ) );
}

inline HS2XferMode
operator|=( HS2XferMode &me, HS2XferMode const &mask )
{
	return me = HS2XferMode( UInt8( me ) | UInt8( mask ) );
}


#if 0
#define DECLARE_XFER_FN( typ ) \
	/*Err Xfer##typ( typ *pItem, MemPtr pRec, UInt16 *pOffset, HS2XferMode mode );*/ \
	static Err Xfer##typ( typ *pItem, MemPtr pRec, UInt16 *pOffset, HS2XferMode mode ) \
	{ \
	    Err err = 0; \
	    if( mode == xferLoad ) \
		MemMove( pItem, ( UInt8* )pRec + *pOffset, sizeof( typ ) ); \
	    else if( mode == xferSave ) \
		CE( DmWrite( pRec, *pOffset, pItem, sizeof( typ ) ) ); \
	    /* xferSkip just increments *pOffset */ \
	    ( *pOffset ) += sizeof( typ ); \
	Error: \
	    return err; \
	}
#endif

#define DECLARE_XFER_FN( typ ) \
	static Err Xfer##typ( typ *pItem, MemPtr pRec, UInt16 *pOffset, HS2XferMode mode ) \
	{ \
	    if( mode & xferLoad ) \
	    { \
		register UInt16 cb = sizeof( typ ); \
		register UInt8* pb = ( UInt8* )pRec + *pOffset; \
		MemMove( pItem, pb, cb ); \
		if( sizeof( typ ) == 2 ) CopyMacWord( pItem, pItem ); \
		if( sizeof( typ ) == 4 ) CopyMacDWord( pItem, pItem ); \
	    } \
	    else if( mode & xferSave ) \
	    { \
		typ xx; \
		MemMove( &xx, pItem, sizeof( typ ) ); \
		if( sizeof( typ ) == 2 ) CopyMacWord( &xx, &xx ); \
		if( sizeof( typ ) == 4 ) CopyMacDWord( &xx, &xx ); \
		DmWrite( pRec, *pOffset, &xx, sizeof( typ ) ); \
	    } \
	    /* xferSkip just increments *pOffset */ \
	    ( *pOffset ) += sizeof( typ ); \
	    return 0; \
	}

#if 0
#define DECLARE_XFER_FN( typ ) \
	/*Err Xfer##typ( typ *pItem, MemPtr pRec, UInt16 *pOffset, HS2XferMode mode );*/ \
	static Err Xfer##typ( typ *pItem, MemPtr pRec, UInt16 *pOffset, HS2XferMode mode ) \
	{ \
	    if( mode & xferLoad ) \
		MemMove( pItem, ( UInt8* )pRec + *pOffset, sizeof( typ ) ); \
	    else if( mode & xferSave ) \
		DmWrite( pRec, *pOffset, pItem, sizeof( typ ) ); \
	    /* xferSkip just increments *pOffset */ \
	    ( *pOffset ) += sizeof( typ ); \
	    return 0; \
	}
#endif

DECLARE_XFER_FN( UInt16 )
DECLARE_XFER_FN( UInt32 )
DECLARE_XFER_FN( DateTimeType )

static Err
XferUInt8( UInt8 *pb, MemPtr pRec, UInt16 *pOffset, HS2XferMode mode )
{
	if( mode & xferLoad )
	    *pb = *( ( ( UInt8* )pRec ) + *pOffset );
	else if( mode & xferSave )
	    DmWrite( pRec, *pOffset, pb, sizeof( *pb ) );
	// xferSkip just increments *pOffset.
	( *pOffset )++;
	return 0;
}

static Err
XferUnits4( StrT* ps, MemPtr pRec, UInt16* pOffset, HS2XferMode mode )
{
	Char szUnits[5];

	if( mode & xferLoad )
	{
	    MemMove( szUnits, ( ( UInt8* )pRec ) + *pOffset, 4 );
	    szUnits[4] = 0;
	    ps->SetA( szUnits );
	}
	else if( mode & xferSave )
	{
	    // Never write the old format.
	    Panic();
	}

	// xferSkip just increments *pOffset.
	( *pOffset ) += 4;
	return 0;
}

static Err
XferUnitsEx( StrT* ps, MemPtr pRec, UInt16* pOffset, HS2XferMode mode )
{
	UInt8 cch;
	Char szUnits[cchMaxUnit];

	if( mode & xferSave )
	{
	    Assert( ps->Length() );
	    Assert( ps->Length() < cchMaxUnit );

#ifdef UNICODE
	    cch = UInt8( WideToNarrow( ps->Text(), ps->Length(), szUnits, DimensionOf( szUnits ) - 1 ) );
	    szUnits[cch] = 0;
#else
	    cch = ps->Length();
	    MemMove( szUnits, ps->Text(), cch + 1 );
#endif

	    cch++;
	    XferUInt8( &cch, pRec, pOffset, xferSave );

	    DmWrite( pRec, *pOffset, szUnits, cch );
	}
	else
	{
	    // xferLoad, xferSkip.

	    XferUInt8( &cch, pRec, pOffset, xferLoad );

	    Assert( cch > 0 );

	    if( mode & xferLoad )
	    {
#ifdef UNICODE
		WCHAR wz[cchMaxUnit];
		ps->Set( wz, NarrowToWide( ( Char const* )( ( UInt8 const* )pRec + *pOffset ), cch - 1, wz, DimensionOf( wz ) - 1 ) );
#else
		ps->Set( ( Char* )( ( UInt8* )pRec + *pOffset ), cch - 1 );
#endif
	    }
#ifdef DEBUG
	    else
		Assert( mode == xferSkip );
#endif
	}

	( *pOffset ) += UInt16( 0 ) + cch;

	return 0;
}

static Err
XferAisle1( StrT* ps, MemPtr pRec, UInt16* pOffset, HS2XferMode mode )
{
	if( mode & xferLoad )
	{
	    TCHAR sz[cchMaxAisle];
	    BYTE aisle = *( LPBYTE( pRec ) + *pOffset );

	    sz[0] = 0;

	    if( aisle )
		if( aisle <= MAX_NUMERIC_AISLES )
		{
		    _itot_s( aisle, RgC( sz ), 10 );
		}
		else if( aisle <= MAX_AISLES )
		{
		    sz[0] = 'A' + ( aisle - ( MAX_NUMERIC_AISLES + 1 ) );
		    sz[1] = 0;
		}

	    ps->Set( sz );
	}
	else if( mode & xferSave )
	{
	    // Never write the old format.
	    Panic();
	}

	// xferSkip just increments *pOffset.
	( *pOffset )++;
	return 0;
}

static Err
XferAisleEx( StrT* ps, MemPtr pRec, UInt16* pOffset, HS2XferMode mode )
{
	UInt8 cch;
	Char szAisle[cchMaxAisle];

	if( mode & xferSave )
	{
	    Assert( ps->Length() );

#ifdef UNICODE
	    cch = UInt8( WideToNarrow( ps->Text(), ps->Length(), szAisle, DimensionOf( szAisle ) - 1 ) );
	    szAisle[cch] = 0;
#else
	    cch = ps->Length();
	    MemMove( szAisle, ps->Text(), cch + 1 );
#endif

	    cch++;
	    XferUInt8( &cch, pRec, pOffset, xferSave );

	    DmWrite( pRec, *pOffset, szAisle, cch );
	}
	else
	{
	    // xferLoad, xferSkip.

	    XferUInt8( &cch, pRec, pOffset, xferLoad );

	    Assert( cch > 0 );

	    if( mode & xferLoad )
	    {
#ifdef UNICODE
		WCHAR wz[cchMaxAisle];
		ps->Set( wz, NarrowToWide( ( Char const* )( ( UInt8 const* )pRec + *pOffset ), cch - 1, wz, DimensionOf( wz ) - 1 ) );
#else
		ps->Set( ( Char* )( ( UInt8* )pRec + *pOffset ), cch - 1 );
#endif
	    }
#ifdef DEBUG
	    else
		Assert( mode == xferSkip );
#endif
	}

	( *pOffset ) += UInt16( 0 ) + cch;

	return 0;
}

static Err
XferString( StrT* ps, MemPtr pRec, UInt16 *pOffset, HS2XferMode mode )
{
	Err		err = 0;
	UInt8		cbShort;
	UInt16		cbLong;

#ifdef DEBUG
	if( mode & ( xferLoad|xferSave ) )
	    Assert( ps );
#endif

	if( mode & xferSave )
	{
	    Assert( ps );
	    Assert( ps->GetCbXferPalm() < 65000 );
	    cbLong = UInt16( ps->GetCbXferPalm() );
	    cbShort = ( cbLong >= 255 ) ? UInt8( 255 ) : UInt8( cbLong );
	    XferUInt8( &cbShort, pRec, pOffset, xferSave );
	    if( cbShort == 255 )
		XferUInt16( &cbLong, pRec, pOffset, mode );

	    LPCTSTR pszBegin = ps->Text();
	    LPCTSTR pszEnd;
	    UInt16 cbTo = 0;
	    LPSTR pszTo = LPSTR( pRec ) + *pOffset;
	    while( *pszBegin )
	    {
		pszEnd = pszBegin;

#ifdef OS_LF
		pszEnd = ps->Text() + ps->Length();
#else
		while( *pszEnd && *pszEnd != '\r' )
		    pszEnd++;
#endif

		if( pszEnd - pszBegin )
		{
		    UInt16 const cb = UInt16( TCharToNarrow( pszBegin, pszEnd - pszBegin, pszTo, cbLong - cbTo ) );
		    pszTo += cb;
		    cbTo += UInt16( 0 ) + cb;
		}

#if defined( OS_CRLF )
		if( *pszEnd == '\r' )
		    pszEnd++;
#elif defined( OS_CR )
		if( *pszEnd == '\r' )
		{
		    *( pszTo++ ) = '\n';
		    cbTo++;
		    pszEnd++;
		}
#endif

		pszBegin = pszEnd;
	    }

	    Assert( cbTo == cbLong );

	    ( LPSTR( pRec ) + *pOffset )[cbLong] = 0;
	    cbLong++;			// Null terminated (new string format).
	}
	else
	{
	    // xferLoad, xferSkip.
	    XferUInt8( &cbShort, pRec, pOffset, xferLoad );
	    if( cbShort == 255 )
		XferUInt16( &cbLong, pRec, pOffset, xferLoad );
	    else
		cbLong = cbShort;
	    if( mode & xferLoad )
	    {
		LPTSTR pszAlloc;
		unsigned cchWide = NarrowToTChar( LPSTR( pRec ) + *pOffset, cbLong, 0, 0 );

#ifdef OS_CRLF
		for( LPCSTR pszWalk = LPCSTR( pRec ) + *pOffset; *pszWalk; pszWalk++ )
		    if( *pszWalk == '\n' )
			cchWide++;
#endif

		pszAlloc = LPTSTR( MemAlloc( ( cchWide + 1 ) * sizeof( *pszAlloc ) ) );
		CPE( pszAlloc );

		LPSTR pszBegin = LPSTR( pRec ) + *pOffset;
		LPSTR pszEnd;
		unsigned cchTo = 0;
		LPTSTR pszTo = pszAlloc;
		while( *pszBegin )
		{
		    pszEnd = pszBegin;

#ifdef OS_LF
		    pszEnd = pszBegin + cbLong;
#else
		    while( *pszEnd && *pszEnd != '\n' )
			pszEnd++;
#endif

		    if( pszEnd - pszBegin )
		    {
			Assert( pszEnd - pszBegin < 65000 );
			UInt16 const cb = UInt16( pszEnd - pszBegin );
			unsigned const cch = NarrowToTChar( pszBegin, cb, pszTo, cchWide - cchTo );
			pszTo += cch;
			cchTo += cch;
			pszBegin = pszEnd;
		    }

#if defined( OS_CRLF )
		    if( *pszEnd == '\n' )
		    {
			*( pszTo++ ) = '\r';
			*( pszTo++ ) = '\n';
			cchTo += 2;
			pszBegin++;
		    }
#elif defined( OS_CR )
		    if( *pszEnd == '\n' )
		    {
			*( pszTo++ ) = '\r';
			cchTo++;
			pszBegin++;
		    }
#endif
		}

		Assert( cchTo == cchWide );

		ps->Attach( pszAlloc, cchWide, cchWide + 1 );
	    }
#ifdef DEBUG
	    else
		Assert( mode == xferSkip );
#endif

	    cbLong++;			// Null terminator (new string format).
	}

	( *pOffset ) += UInt16( 0 ) + cbLong;

LError:
	return err;
}

static UInt16
CbXferString( StrT const* ps )
{
	UInt16 cbSize = 0;

	if( ps && ps->Length() )
	{
	    Assert( ps->GetCbXferPalm() < 65000 );
	    cbSize = UInt16( ps->GetCbXferPalm() );
	    if( cbSize >= 255 )
		cbSize += sizeof( UInt16 ); // Additional length word when >= 255.
	    //$ review: (chrisant) Will this ever need to know about the xfer
	    // mode and xfer_oldstringformat?
	    cbSize += 1 + 1 + 1;	// Tag byte, plus mandatory length byte, plus null terminator (new string format).
	}

	return cbSize;
}

static Err
XferStores( Item* pitem, MemPtr pRec, UInt16* pOffset, HS2XferMode mode )
{
	Err		err = 0;
	UInt8		cStores;

	Assert( pitem );

	if( mode & xferSave )
	{
	    Assert( pitem->m_stores.Count() );
	    cStores = UInt8( pitem->m_stores.Count() );
	    XferUInt8( &cStores, pRec, pOffset, xferSave );
	    DmWrite( pRec, *pOffset, pitem->m_stores.Array(), cStores );
	}
	else
	{
	    XferUInt8( &cStores, pRec, pOffset, xferLoad );
	    if( mode & xferLoad )
	    {
		// Clever optimization for CopyStoreWorker and similar
		// functions, to avoid realloc simply from adding one store.
		CPE( pitem->m_stores.Alloc( cStores + 1 ) );
		pitem->m_stores.Copy( cStores, ( Store* )ByteOffset( pRec, *pOffset ) );
	    }
	}

	( *pOffset ) += UInt16( 0 ) + cStores;

LError:
	return err;
}

static Err
XferPerStore( Item* pitem, MemPtr pRec, UInt16* pOffset, HS2XferMode mode )
{
	Err err = 0;
	UInt8 cPerStore;

	Assert( pitem );

	if( mode & xferSave )
	{
	    PanicSz( "UNREACHABLE" );
	}
	else
	{
	    XferUInt8( &cPerStore, pRec, pOffset, xferLoad );

	    if( mode & xferLoad )
	    {
		PerStoreHs2List perstore;

		CPE( perstore.Copy( cPerStore, reinterpret_cast<PerStore0*>( ByteOffset( pRec, *pOffset ) ) ) );
		perstore.MacAdjust();

		perstore.UpConvert( &pitem->m_perstore );
	    }

	    ( *pOffset ) += cPerStore * sizeof( PerStore0 );
	}

LError:
	return err;
}

static void
XferOnePerStoreEx( PerStore2* p, MemPtr pRec, UInt16* pOffset, HS2XferMode mode )
{
	XferUInt8( &p->m_idStore, pRec, pOffset, mode );

	if( mode & xferLoad )
	    memcpy( p->m_szAisle, static_cast<UInt8*>( pRec ) + *pOffset, sizeof( p->m_szAisle ) );
	else if( mode & xferSave )
	    DmWrite( pRec, *pOffset, p->m_szAisle, sizeof( p->m_szAisle ) );

	// xferSkip just increments *pOffset.
	( *pOffset ) += sizeof( p->m_szAisle );

	XferUInt32( reinterpret_cast<UInt32*>( &p->m_lPrice ), pRec, pOffset, mode );
}

static Err
XferPerStoreEx( Item* pitem, MemPtr pRec, UInt16* pOffset, HS2XferMode mode )
{
	Err err = 0;
	UInt8 cPerStore;

	Assert( pitem );

	if( mode & xferSave )
	{
	    Assert( pitem->m_perstore.Count() );

	    cPerStore = UInt8( pitem->m_perstore.Count() );
	    XferUInt8( &cPerStore, pRec, pOffset, xferSave );

	    PerStoreHs2List perstore( pitem->m_perstore );
	    PerStore2 const* pPerStore = perstore.Array();

	    while( cPerStore-- )
	    {
		XferOnePerStoreEx( const_cast<PerStore2*>( pPerStore ), pRec, pOffset, mode );
		pPerStore++;
	    }
	}
	else
	{
	    XferUInt8( &cPerStore, pRec, pOffset, xferLoad );

	    if( mode & xferLoad )
	    {
		PerStoreHs2List perstore;

		CPE( perstore.Copy( cPerStore, reinterpret_cast<PerStore2*>( ByteOffset( pRec, *pOffset ) ) ) );
		perstore.MacAdjust();

		perstore.UpConvert( &pitem->m_perstore );
	    }

	    ( *pOffset ) += cPerStore * sizeof( PerStore2 );
	}

LError:
	return err;
}

static Err
XferCategory( UInt16* pCategory, MemPtr pRec, UInt16* pOffset, HS2XferMode mode )
{
	// Item::m_category is a UInt16 for legacy reasons, but the packed
	// category field is only a UInt8 (and would never need to store a
	// UInt16).
	if( mode & xferLoad )
	{
	    *pCategory = *( reinterpret_cast<UInt8*>( pRec ) + *pOffset );
	}
	else if( mode & xferSave )
	{
	    CASSERT( MAX_CATEGORIES <= 200 );
	    Assert( *pCategory < MAX_CATEGORIES );
	    UInt8 const category = UInt8( *pCategory );
	    DmWrite( pRec, *pOffset, &category, sizeof( UInt8 ) );
	}
	// xferSkip just increments *pOffset.
	( *pOffset )++;
	return 0;
}

inline UInt16
CbAutoOrders( WORD c )
{
	return c * WORD( 1 + 4 );
}

static void
XferOneAutoOrder( AutoOrder* p, MemPtr pRec, UInt16* pOffset, HS2XferMode mode )
{
	XferUInt8( &p->m_idStore, pRec, pOffset, mode );
	XferUInt32( reinterpret_cast<UInt32*>( &p->m_ulOrder ), pRec, pOffset, mode );
}

static Err
XferAutoOrder( AutoOrderList* pList, MemPtr pRec, UInt16* pOffset, HS2XferMode mode )
{
	Err err = 0;
	UInt8 cAutoOrder;

	Assert( pList );

	if( mode & xferSave )
	{
	    cAutoOrder = UInt8( pList->Count() );
	    Assert( cAutoOrder );
	    XferUInt8( &cAutoOrder, pRec, pOffset, xferSave );
	}
	else
	{
	    XferUInt8( &cAutoOrder, pRec, pOffset, xferLoad );
	}

	if( mode == xferSkip )
	{
	    *pOffset += CbAutoOrders( cAutoOrder );
	}
	else
	{
#ifdef DEBUG
	    UInt16 const offsetExpected = *pOffset + CbAutoOrders( cAutoOrder );
#endif

	    if( mode & xferLoad )
	    {
		Assert( pList->FPristine() );
		if( !pList->AllocForXfer( cAutoOrder ) )
		    return memErrNotEnoughSpace;
	    }

	    for( AutoOrder const* p = pList->Array(); cAutoOrder--; p++ )
		XferOneAutoOrder( const_cast<AutoOrder*>( p ), pRec, pOffset, mode );

	    Assert( offsetExpected == *pOffset );
	}

	return err;
}

/*
 * Custom Category List.
 */

bool
FindMetadataRecord( DmOpenRef pdb, UInt8 tagFind, unsigned* precord, MemHandle* phandle, void** ppv )
{
	Assert( tagMasterStoreList == tagFind ||
		tagCategoryList == tagFind );
	Assert( precord );
	Assert( phandle );

	bool fSeenStores = false;

	*precord = 0;
	*phandle = 0;
	*ppv = 0;

	for( unsigned ii = 0; ii < 2; ii++ )
	{
	    MemHandle const h = DmGetRecord( pdb, ii );

	    if( !h )
		break;

	    UInt8* const pb = ( UInt8* )MemHandleLock( h );
	    UInt8 const tagRecord = *pb;

	    if( tagRecord == tagFind )
	    {
		*precord = ii;
		*phandle = h;
		*ppv = pb;
		return true;
	    }

	    MemHandleUnlock( h );
	    DmReleaseRecord( pdb, ii, false );

	    if( tagRecord != tagMasterStoreList && tagRecord != tagCategoryList )
		break;

	    if( tagRecord == tagMasterStoreList )
		fSeenStores = true;
	}

	if( tagMasterStoreList == tagFind )
	    *precord = 0;
	else if( tagCategoryList == tagFind )
	    *precord = fSeenStores ? 1 : 0;

	return false;
}

void
DbOpenHs2::GetCategoryList()
{
	Assert( !m_pCategories );

	unsigned record;
	MemHandle h;
	CustomCategoryList *pCategories = 0;

	m_pCategories = new CustomCategoryList;

	if( FindMetadataRecord( m_pdb, tagCategoryList, &record, &h, &pCategories ) )
	{
	    Assert( MemHandleSize( h ) == sizeof( *m_pCategories ) );

	    memcpy( m_pCategories, pCategories, sizeof( *m_pCategories ) );

	    MemHandleUnlock( h );
	    DmReleaseRecord( m_pdb, record, false );
	}
}

void
DbOpenHs2::PutCategoryList()
{
	Assert( m_prefs.fCustomCategories );
	Assert( m_pCategories );

	// If the categories aren't actually dirty, there's nothing to do.

	if( !m_fCategoriesDirty )
	    return;

	// If we don't have a list and we don't need a list, there's nothing
	// to do.

	unsigned record;
	MemHandle h;
	CustomCategoryList* pCategories;
	bool const fHaveList = FindMetadataRecord( m_pdb, tagCategoryList, &record, &h, &pCategories );
	bool const fNeedList = !!m_pCategories->Count();

	if( fHaveList || fNeedList )
	{
	    // If there is no list and there needs to be one, create the record.

	    if( !fHaveList && fNeedList )
	    {
		h = DmNewRecord( m_pdb, &record, sizeof( *pCategories ) );

		if( h )
		{
		    m_cMetadataRecords++;

		    Assert( m_cMetadataRecords <= 2 );

		    pCategories = ( CustomCategoryList* )MemHandleLock( h );

		    if( m_pbits )
			m_pbits->NewRecord( record, false );
		}
		else
		    Panic1( "PutCategoryList\n  DmNewRecord failed (%d)", m_pdb->DmGetLastErr() );
	    }

	    // If we have a pointer to the list record, update it.

	    if( pCategories )
	    {
		Assert( MemHandleSize( h ) == sizeof( *pCategories ) );

		memcpy( pCategories, m_pCategories, sizeof( *pCategories ) );	// Copy even when !fNeedList, in case 

		MemHandleUnlock( h );
		DmReleaseRecord( m_pdb, record, true );
	    }

	    // If there is a list record but there doesn't need to be, remove the
	    // record.

	    if( fHaveList && !fNeedList )
	    {
		m_cMetadataRecords--;

		DmRemoveRecord( m_pdb, record );

		if( m_pbits )
		    m_pbits->RemoveRecord( record );
	    }

	    SetDirty();
	}

	m_fCategoriesDirty = false;
}

/*
 * Sort Helpers.
 */

#ifdef DEBUG
static Boolean s_fAlreadySortContext = false;
#endif

static DbOpenHs2* s_popenSort;

SortContext::SortContext( DbOpenHs2* popen )
{
#ifdef DEBUG
	Assert( !s_fAlreadySortContext );
	s_fAlreadySortContext = true;
#endif

	s_popenSort = popen;
}

SortContext::~SortContext()
{
	s_popenSort = 0;

#ifdef DEBUG
	s_fAlreadySortContext = false;
#endif
}

// -- Sort Comparators ----------------------------------------------------

typedef int ( *SortCmpPfn )( MemPtr pA, MemPtr pB );

static Boolean s_fEmptyString;

inline void
CopyBytes( MemPtr _pDst, MemPtr _pSrc, UInt16 cb )
{
	register UInt8 *pDst = ( UInt8* )_pDst;
	register const UInt8 *pSrc = ( UInt8 const* )_pSrc;

	while( cb-- )
	    *( pDst++ ) = *( pSrc++ );
}

static int
StringTypeCmp( MemPtr pA, MemPtr pB )
{
	const Char *pszA = pA ? ( Char const* )pA : "";
	const Char *pszB = pB ? ( Char const* )pB : "";
	register int iSort;

	// Make empty strings sort after non-empty strings.

	if( !*pszA )
	{
	    if( !*pszB )
		return 0;
	    iSort = 1;
	}
	else if( !*pszB )
	    iSort = -1;
	else
	    return StrCaselessCompareSort( pszA, pszB );

	// Make empty strings *always* sort after non-empty strings, even for
	// "descending" order.

	Assert( !*pszA || !*pszB );
	s_fEmptyString = true;

	return iSort;
}

//#define FULL_NEGATIVE_COMPARISONS
#ifdef FULL_NEGATIVE_COMPARISONS
#define IsBeginNumeric( psz ) ( *psz <= '9' && ( *psz >= '0' || ( ( *psz == '-' || *psz == '+' ) && psz[1] >= '0' && psz[1] <= '9' ) ) )
#else
#define IsBeginNumeric( psz ) ( *psz <= '9' && *psz >= '0' )
#endif

inline Char
EnsureUpper( Char ch )
{
	if( ch >= 'a' && ch <= 'z' )
	    return ch - 'a' + 'A';

	return ch;
}

static Int32
ParseNum( Char const*& psz )
{
	Int32 num = 0;
#ifdef FULL_NEGATIVE_COMPARISONS
	Boolean fNegate = false;

	if( *psz == '+' )
	    psz++;

	if( *psz == '-' )
	{
	    fNegate = true;
	    psz++;
	}
#endif

	while( *psz <= '9' && *psz >= '0' )
	{
	    num *= 10;
	    num += ( *psz ) - '0';
	    psz++;
	}

#ifdef FULL_NEGATIVE_COMPARISONS
	if( fNegate )
	    return -num;
#endif

	return num;
}

static int
StrAlphaNumCompare( Char const* pszA, Char const* pszB )
{
	while( true )
	{
	    if( IsBeginNumeric( pszA ) && IsBeginNumeric( pszB ) )
	    {
		register Int32 numA = ParseNum( pszA );	// Passed by reference.
		register Int32 numB = ParseNum( pszB );	// Passed by reference.

		if( numA < numB )
		    return -1;
		else if( numA > numB )
		    return 1;
	    }
	    else
	    {
		register int result = int( UInt8( EnsureUpper( *pszA ) ) ) - int( UInt8( EnsureUpper( *pszB ) ) );

		if( result || !*pszA )
		    return result;

		pszA++;
		pszB++;
	    }
	}
}

static int
AisleStringCmp( MemPtr pA, MemPtr pB )
{
	Char const* pszA = pA ? ( Char const* )pA : "";
	Char const* pszB = pB ? ( Char const* )pB : "";

	return StrAlphaNumCompare( pszA, pszB );
}

static int
CmpSortPriority( MemPtr pA, MemPtr pB )
{
	return *( ( Int8* )pA ) - *( ( Int8* )pB );
}

static int
CmpSortQuantity( MemPtr pA, MemPtr pB )
{
	Int32 lA = 0;
	Int32 lB = 0;

	if( pA )
	{
	    CopyBytes( &lA, pA, sizeof( lA ) );
	    lA = MacDWord( lA );
	}
	if( pB )
	{
	    CopyBytes( &lB, pB, sizeof( lB ) );
	    lB = MacDWord( lB );
	}

	if( lA < lB )
	    return -1;
	if( lA > lB )
	    return 1;

	return 0;
}

static int
CmpSortDate( MemPtr pA, MemPtr pB )
{
	static UInt16 const c_dateNone = dateNone;
	PalmDateType date1;
	PalmDateType date2;

	CASSERT( SameSize( c_dateNone, date1 ) );

	CopyBytes( &date1, pA ? pA : ( MemPtr )&c_dateNone, sizeof( date1 ) );
	CopyBytes( &date2, pB ? pB : ( MemPtr )&c_dateNone, sizeof( date2 ) );

	*( ( UInt16* )&date1 ) = MacWord( *( ( UInt16* )&date1 ) );
	*( ( UInt16* )&date2 ) = MacWord( *( ( UInt16* )&date2 ) );

	register int result;

	result = int( date1.year ) - int( date2.year );
	if( result != 0 )
	{
	    if( *( UInt16* )&date1 == dateNone )
		return 1;
	    if( *( UInt16* )&date2 == dateNone )
		return -1;
	    return result;
	}

	result = int( date1.month ) - int( date2.month );
	if( result != 0 )
	    return result;

	result = int( date1.day ) - int( date2.day );

	return result;
}

static int
CmpSortPrice( MemPtr pA, MemPtr pB )
{
	Int32 lA = 0;
	Int32 lB = 0;

	if( pA )
	{
	    CopyBytes( &lA, pA, sizeof( lA ) );
	    lA = MacDWord( lA );
	}
	if( pB )
	{
	    CopyBytes( &lB, pB, sizeof( lB ) );
	    lB = MacDWord( lB );
	}

	if( lA < lB )
	    return -1;
	if( lA > lB )
	    return 1;

	return 0;
}

static int
CmpSortCustomCategory( MemPtr pA, MemPtr pB )
{
	Assert( s_popenSort->Categories() );

	if( !pA && pB )
	    return 1;
	if( !pB && pA )
	    return -1;
	if( !pA && !pB )
	    return 0;

	return StrCaselessCompare( s_popenSort->Categories()->CategoryFromId( *( UInt8* )pA )->m_szCategory,
				   s_popenSort->Categories()->CategoryFromId( *( UInt8* )pB )->m_szCategory );
}

static int
CmpSortEntryOrder( MemPtr pA, MemPtr pB )
{
	UInt32 ulA = 0;
	UInt32 ulB = 0;

	if( pA )
	{
	    CopyBytes( &ulA, pA, sizeof( ulA ) );
	    ulA = MacDWord( ulA );
	}
	if( pB )
	{
	    CopyBytes( &ulB, pB, sizeof( ulB ) );
	    ulB = MacDWord( ulB );
	}

	if( ulA < ulB )
	    return -1;
	if( ulA > ulB )
	    return 1;

	return 0;
}

static int
CmpSortAlarm( MemPtr pA, MemPtr pB )
{
	UInt32 ulA = 0;
	UInt32 ulB = 0;

	if( pA )
	{
	    CopyBytes( &ulA, pA, sizeof( ulA ) );
	    ulA = MacDWord( ulA );
	}
	if( pB )
	{
	    CopyBytes( &ulB, pB, sizeof( ulB ) );
	    ulB = MacDWord( ulB );
	}

	// Ascending sort puts empty alarms at the bottom.

	if( ulA < ulB )
	    return ulA ? -1 : 1;
	if( ulA > ulB )
	    return ulB ? 1 : -1;

	return 0;
}

static int
CmpSortIcon( MemPtr pA, MemPtr pB )
{
	UInt16 nA = 0;
	UInt16 nB = 0;

	if( pA )
	{
	    CopyBytes( &nA, pA, sizeof( nA ) );
	    nA = MacWord( nA );
	}
	if( pB )
	{
	    CopyBytes( &nB, pB, sizeof( nB ) );
	    nB = MacWord( nB );
	}

	// Ascending sort puts empty icons at the bottom.

	if( nA < nB )
	    return nA ? -1 : 1;
	if( nA > nB )
	    return nB ? 1 : -1;

	return 0;
}

static int
CmpCompletedStatus( MemPtr pA, MemPtr pB )
{
	UInt16 flagsA = 0;
	UInt16 flagsB = 0;

	if( pA )
	{
	    CopyBytes( &flagsA, pA, sizeof( flagsA ) );
	    flagsA = MacWord( flagsA );
	}
	if( pB )
	{
	    CopyBytes( &flagsB, pB, sizeof( flagsB ) );
	    flagsB = MacWord( flagsB );
	}

	if( ( flagsA & ( flagNeed|flagPurchased ) ) != ( flagsB & ( flagNeed|flagPurchased ) ) )
	{
	    if( ( flagsA & flagNeed ) != ( flagsB & flagNeed ) )
		return ( flagsA & flagNeed ) ? -1 : 1;
	    else if( ( flagsA & flagPurchased ) != ( flagsB & flagPurchased ) )
		return ( flagsA & flagPurchased ) ? 1 : -1;
	}

	return 0;
}

static int
CmpAutoOrder( MemPtr _pA, MemPtr _pB )
{
	DWORD ulA = 0;
	DWORD ulB = 0;
	BYTE const _store = s_popenSort->Inline_GetStore();
	BYTE const store = ( _store == noStore ) ? allStores : _store;

	Assert( store == allStores ||
		store < MAX_STORES );

	if( _pA )
	{
	    UInt8* pA = static_cast<UInt8*>( _pA );
	    for( BYTE cA = *( pA++ ); cA--; pA += sizeof( UInt8 ) + sizeof( DWORD ) )
	    {
		if( store == *pA )
		{
		    CopyBytes( &ulA, pA + sizeof( UInt8 ), sizeof( ulA ) );
		    ulA = MacDWord( ulA );
		    break;
		}
	    }
	}
	if( _pB )
	{
	    UInt8* pB = static_cast<UInt8*>( _pB );
	    for( BYTE cB = *( pB++ ); cB--; pB += sizeof( UInt8 ) + sizeof( DWORD ) )
	    {
		if( store == *pB )
		{
		    CopyBytes( &ulB, pB + sizeof( UInt8 ), sizeof( ulB ) );
		    ulB = MacDWord( ulB );
		    break;
		}
	    }
	}

	if( ulA < ulB )
	    return -1;
	if( ulA > ulB )
	    return 1;
	return 0;
}

inline UInt8*
SkipString( UInt8* p )
{
	p++;

	if( *( ( UInt8* )p ) < 255 )
	    return p + sizeof( UInt8 ) + *( ( UInt8* )p ) + 1;

	UInt16 cch;

	p++;

	CopyBytes( &cch, p, sizeof( cch ) );

	cch = MacWord( cch );

	return p + sizeof( UInt16 ) + cch + 1;
}

inline UInt8*
GetString( UInt8* p )
{
	p++;

	if( *( ( UInt8* )p ) < 255 )
	    return p + sizeof( UInt8 );

	return p + sizeof( UInt8 ) + sizeof( UInt16 );
}

static Boolean const c_rgIsStringTag[] =
{
	false,	//tagEND,
	true,	//tagDescription,
	true,	//tagNote,
	false,	//tagFlags,
	false,	//tagQuantity,
	false,	//tagPrice,
	false,	//tagPriority,
	false,	//tagDate,
	false,	//tagAlarm,
	false,	//tagUnits,
	false,	//tagMasterStoreList,
	false,	//tagStores,
	false,	//tagAisle,
	false,	//tagEntryOrder,
	false,	//tagPerStore,
	true,	//tagCustomText,
	false,	//tagAlarmMidi,
	false,	//tagIcon,
	true,	//tagUnitsEx,
	true,	//tagAisleEx,
	false,	//tagPerStoreEx,
	false,	//tagCategoryList,
	false,	//tagCategory,
	false,	//tagAutoOrder,
};
CASSERT( DimensionOf( c_rgIsStringTag ) == tagMAX );

static UInt8 s_iRotateUnits = 0;
static UInt8 s_iRotateAisles = 0;
static UInt8 s_iRotatePerStore = 0;

MemPtr
GetMemberPtr( MemPtr _p, UInt8 tag )
{
	UInt8* p = ( UInt8* )_p;

	AssertSz( tagFlags == *p, "record is corrupt" );

	Assert( tagUnits != tag );
	Assert( tagAisle != tag );

#ifdef DEBUG
	UInt16 ii = 0;
	UInt8 rgTags[32];
#endif

	while( tag != *p )
	{
#ifdef DEBUG
	    rgTags[ii++] = *p;
#endif

	    switch( *p )
	    {
	    case tagEND:
		if( tagIcon == tag )
		{
		    //$ todo: (chrisant) category icon?!?
		}
		return 0;

	    case tagFlags:
	    case tagIcon:
	    case tagDate:
		p += 1 + sizeof( UInt16 );
		break;

	    case tagAlarm:
	    case tagAlarmMidi:
	    case tagQuantity:
	    case tagEntryOrder:
	    case tagPrice:
		p += 1 + sizeof( UInt32 );
		break;

	    case tagUnits:
		if( tagUnitsEx == tag )
		{
		    static Char s_szUnits[2][5];
		    register Char* psz = s_szUnits[s_iRotateUnits];

		    s_iRotateUnits ^= 1;

		    psz[4] = 0;
		    CopyBytes( psz, p + 1, sizeof( UInt32 ) );

		    *( ( UInt32* )psz ) = MacDWord( *( ( UInt32* )psz ) );

		    return psz;
		}
		p += 1 + sizeof( UInt32 );
		break;

	    case tagAisle:
		if( tagAisleEx == tag )
		{
		    static Char s_szAisles[2][cchMaxAisle];
		    register Char* psz = s_szAisles[s_iRotateAisles];

		    s_iRotateAisles ^= 1;

		    UInt8 aisle = p[1];

		    psz[0] = 0;

		    if( aisle )
			if( aisle <= MAX_NUMERIC_AISLES )
			{
			    __itoa( aisle, psz, DimensionOf( s_szAisles[0] ), 10 );
			}
			else if( aisle <= MAX_AISLES )
			{
			    psz[0] = 'A' + ( aisle - ( MAX_NUMERIC_AISLES + 1 ) );
			    psz[1] = 0;
			}

		    return psz;
		}
		p += 1 + sizeof( UInt8 );
		break;

	    case tagPriority:
	    case tagCategory:
		p += 1 + sizeof( UInt8 );
		break;

	    case tagDescription:
	    case tagNote:
	    case tagCustomText:
		p = SkipString( p );
		break;

	    case tagUnitsEx:
	    case tagAisleEx:
	    case tagStores:
		p += 1 + sizeof( UInt8 ) + p[1];
		break;

	    case tagPerStore:
	    case tagPerStoreEx:
		{
		    bool const fEx = ( tagPerStoreEx == *p );
		    p++;

		    Assert( FImplies( tagPrice == tag || tagAisleEx == tag, s_popenSort ) );

		    if( ( tagPrice == tag || tagAisleEx == tag ) &&
			s_popenSort->Inline_GetPerStorePrices() )
		    {
			static Int32 s_rglPrice[2];
			static Char s_rgszAisle[2][cchMaxAisle];
			register UInt8* pRet;

			// special scope to force compiler to use the
			// PerStoreList construction/destructor semantics we
			// want, in case the compiler isn't smart enough to
			// have only one destruction path.
			{
			    PerStoreHs2List perstore;

			    //$ review: (chrisant) This is pretty inefficient...
			    if( fEx )
				perstore.Copy( *p, ( PerStore2* )( p + 1 ) );
			    else
				perstore.Copy( *p, ( PerStore0* )( p + 1 ) );

			    if( tagPrice == tag )
			    {
				s_rglPrice[s_iRotatePerStore] = perstore.GetPrice( s_popenSort->Inline_GetStore() );
				pRet = ( UInt8* )&s_rglPrice[s_iRotatePerStore];
			    }
			    else // tagAisleEx == tag
			    {
				Char const* pszAisle = perstore.GetAisle( s_popenSort->Inline_GetStore() );
				UInt16 const cch = Min( UInt16( strlen( pszAisle ) ), UInt16( DimensionOf( s_rgszAisle[0] ) - 1 ) );
				memcpy( s_rgszAisle[s_iRotatePerStore], pszAisle, cch );
				s_rgszAisle[s_iRotatePerStore][cch] = 0;
				pRet = ( UInt8* )&s_rgszAisle[s_iRotatePerStore];
			    }
			}

			s_iRotatePerStore ^= 1;
			return pRet;
		    }
		    p += sizeof( UInt8 ) + ( UInt16( *p ) * ( fEx ? sizeof( PerStore2 ) : sizeof( PerStore0 ) ) );
		}
		break;

	    case tagAutoOrder:
		{
		    p++;
		    p += sizeof( UInt8 ) + CbAutoOrders( *p );
		}
		break;

	    default:
#ifdef DEBUG
		StrT s;
		for( UInt16 jj = ( Int16( ii - 3 ) >= 0 ) ? ( ii - 3 ) : 0; jj < ii; jj++ )
		    s.Printf( TEXT( ", %u" ), UInt16( rgTags[jj] ) );
		DbgMsg(( 20, TEXT( "last few tags%s -> bad" ), s.Text() ));
#endif
		PanicSz( "record is corrupt" );
		return 0;
	    }
	}

	if( c_rgIsStringTag[*p] )
	    return GetString( p );

	return p + 1;

#undef p
}

#define tagNone tagEND

// -- Sort Tables ---------------------------------------------------------

#ifdef DEBUG
#define SORTENTRY( sort, mask, dirty )		{ sort, mask, dirty },
#define SORTCMPENTRY( sortType, pfn )		{ pfn, sort##sortType, tag##sortType },
#define SORTCMPENTRY_( sortType, pfn, tag )	{ pfn, sort##sortType, tag },
#else
#define SORTENTRY( sort, mask, dirty )		{ mask, dirty },
#define SORTCMPENTRY( sortType, pfn )		{ pfn, tag##sortType },
#define SORTCMPENTRY_( sortType, pfn, tag )	{ pfn, tag },
#endif

struct SortMasks
{
#ifdef DEBUG
	SortType	sort;
#endif
	MaskType	mask;
	DirtyType	dirty;
};

static SortMasks const c_rgSortMasks[] =
{
	SORTENTRY( sortNone,		0,			0 )
	SORTENTRY( sortPriority,	maskPriority,		dirtyPrio )
	SORTENTRY( sortDate,		maskDate,		dirtyDate )
	SORTENTRY( sortAisle,		maskAisle|maskPerStore,	dirtyAisle|dirtyPerStore )
	SORTENTRY( sortCategory,	maskCATEGORY,		dirtyCategory )
	SORTENTRY( sortDescription,	maskDescription,	dirtyDescription )
	SORTENTRY( sortQuantity,	maskQuantity,		dirtyQuantity )
	SORTENTRY( sortUnits,		maskUnits,		dirtyUnits )
	SORTENTRY( sortPrice,		maskPrice|maskPerStore,	dirtyPrice|dirtyPerStore )
	SORTENTRY( sortEntryOrder,	maskEntryOrder,		dirtyEntryOrder )
	SORTENTRY( sortCustomText,	maskCustomText,		dirtyCustomText )
	SORTENTRY( sortAlarm,		maskAlarm,		dirtyAlarm )
	SORTENTRY( sortIcon,		maskIcon,		dirtyIcon )
};
CASSERT( DimensionOf( c_rgSortMasks ) == sortMAX );

struct SortCmpInfo
{
	SortCmpPfn	m_pfn;
#ifdef DEBUG
	SortType	m_sort;
#endif
	UInt8		m_tag;
};

static SortCmpInfo const c_rgSortCmpInfo[] =
{
	SORTCMPENTRY_( None, 0, tagEND )
	SORTCMPENTRY( Priority, CmpSortPriority )
	SORTCMPENTRY( Date, CmpSortDate )
	SORTCMPENTRY_( Aisle, AisleStringCmp, tagAisleEx )
	SORTCMPENTRY( Category, CmpSortCustomCategory )
	SORTCMPENTRY( Description, StringTypeCmp )
	SORTCMPENTRY( Quantity, CmpSortQuantity )
	SORTCMPENTRY_( Units, StringTypeCmp, tagUnitsEx )
	SORTCMPENTRY( Price, CmpSortPrice )
	SORTCMPENTRY( EntryOrder, CmpSortEntryOrder )
	SORTCMPENTRY( CustomText, StringTypeCmp )
	SORTCMPENTRY( Alarm, CmpSortAlarm )
	SORTCMPENTRY( Icon, CmpSortIcon )
};
CASSERT( DimensionOf( c_rgSortCmpInfo ) == sortMAX );

#ifdef DEBUG
static void
AssertSortHelpers( void )
{
	int ii;
	static Boolean fVerified = false;

	if( !fVerified )
	{
	    fVerified = true;

	    for( ii = DimensionOf( c_rgSortMasks ); ii--; )
		Assert2( c_rgSortMasks[ii].sort == ii,
			 "ERROR: c_rgSortMasks[%d].sort is %d",
			 ii, c_rgSortMasks[ii].sort );

	    for( ii = DimensionOf( c_rgSortCmpInfo ); ii--; )
		Assert2( c_rgSortCmpInfo[ii].m_sort == ii,
			 "ERROR: c_rgSortCmpInfo[%d].sort is %d",
			 ii, c_rgSortCmpInfo[ii].m_sort );
	}
}
#else
#define AssertSortMaskArray() syntaxsafe
#endif

// -- Sort Functions ------------------------------------------------------

MaskType
DbOpenHs2::GetUnpackSortMask()
{
	MaskType mask = 0;
	Int16 ii;
	SortOrderInfo *pSO = &m_prefs.rgOrder[FAltSort()];

#ifdef DEBUG
	if( IfGlobals() )
	    AssertSortHelpers();
#endif

	if( m_prefs.get_flags() & dpfSortDisabled )
	    return 0;

	for( ii = Int16( pSO->dim_m_sort() ); ii--; )
	    mask |= c_rgSortMasks[pSO->get_m_sort( ii )].mask;
	if( pSO->m_fCompletedAtBottom )
	    mask |= maskFlags;
	return mask;
}

DirtyType
DbOpenHs2::GetDirtyMask()
{
	DirtyType dirty = 0;

	if( CanLearnOrder( FAltSort() ) )
	{
	    dirty = dirtyNeed|dirtyAutoOrder;
	}
	else
	{
	    Int16 ii;
	    SortOrderInfo* pSO = &m_prefs.rgOrder[FAltSort()];

#ifdef DEBUG
	    if( IfGlobals() )
		AssertSortHelpers();
#endif

	    for( ii = Int16( pSO->dim_m_sort() ); ii--; )
		dirty |= c_rgSortMasks[pSO->get_m_sort( ii )].dirty;
	    if( pSO->m_fCompletedAtBottom )
		dirty |= dirtyNeed;
	}

	return dirty;
}

Int16
DbOpenHs2::CmpFindSort( MemPtr _pA, MemPtr _pB, Int32 other )
{
#define pA ( ( UInt8 const* )_pA )
#define pB ( ( UInt8 const* )_pB )

	SUPPRESS_UNUSED_VAR( other );

	ASSERTGLOBALS;

	// Sort master store list before any other records.

	if( *pA == tagMasterStoreList || *pB == tagMasterStoreList )
	{
	    if( *pA == tagMasterStoreList )
	    {
		if( *pB == tagMasterStoreList )
		{
		    PanicSz( "there should be only one master store list; how can we be comparing it to itself?" );
		    return 0;
		}
		return -1;
	    }
	    if( *pB == tagMasterStoreList )
	    {
		return 1;
	    }
	}

	// Sort category list after master store list but before any other
	// records.

	if( *pA == tagCategoryList || *pB == tagCategoryList )
	{
	    if( *pA == tagCategoryList )
	    {
		if( *pB == tagCategoryList )
		{
		    PanicSz( "there should be only one category list; how can we be comparing it to itself?" );
		    return 0;
		}
		return -1;
	    }
	    if( *pB == tagCategoryList )
	    {
		return 1;
	    }
	}

	// AutoOrder or CompletedAtBottom perform special comparisons before
	// applying the regular sort order.

	int iSort = 0;
	bool const fAltSort = s_popenSort->FAltSort();
	DatabasePrefs const* const pPrefs = &s_popenSort->m_prefs;
	SortOrderInfo const* const pSO = &pPrefs->rgOrder[fAltSort];

	if( pPrefs->m_rgfAutoOrder[fAltSort] )
	{
	    // AutoOrder uses a special sort order.  But if multiple items are
	    // equivalent (particularly in the case of order value 0) then
	    // fall back to using the sort order from pSO.

	    iSort = CmpCompletedStatus( GetMemberPtr( _pA, tagFlags ), GetMemberPtr( _pB, tagFlags ) );
	    if( iSort )
		goto Out;
	    iSort = CmpAutoOrder( GetMemberPtr( _pA, tagAutoOrder ), GetMemberPtr( _pB, tagAutoOrder ) );
	    if( iSort )
		goto Out;
	}
	else if( pSO->m_fCompletedAtBottom )
	{
	    // Sort completed items at bottom.

	    iSort = CmpCompletedStatus( GetMemberPtr( _pA, tagFlags ), GetMemberPtr( _pB, tagFlags ) );
	    if( iSort )
		goto Out;
	}

	// Try each sort order until we get an inequality (iSort != 0).

	Assert( 0 == iSort );

	for( UInt16 ii = 0; ii < NUM_SORT_FIELDS; ii++ )
	{
	    register SortType const sort = pSO->get_m_sort( ii );

	    Assert( UInt16( sort ) < UInt16( sortMAX ) );

	    register UInt8 tag = c_rgSortCmpInfo[sort].m_tag;

	    s_fEmptyString = false;

	    if( tag )
		iSort = ( *c_rgSortCmpInfo[sort].m_pfn )( GetMemberPtr( _pA, tag ), GetMemberPtr( _pB, tag ) );

	    if( iSort )
	    {
		if( pSO->m_rgfDescend[ii] && !s_fEmptyString )
		    iSort = -iSort;
		break;
	    }
	}

Out:
	if( iSort < 0 )
	    return -1;
	if( iSort > 0 )
	    return 1;
	return 0;
#undef pB
#undef pA
}

/*
 * AppInfo.
 */

static ThisAppInfoPtr
ResizeAppInfo( DmOpenRef pdb, ThisAppInfoPtr pAppInfo )
{
	LocalID idAppInfo;
	MemHandle h;
#ifdef DEBUG
	Err err;
#endif

	h = MemPtrRecoverHandle( pAppInfo );

	while( MemHandleLockCount( h ) )
	    MemHandleUnlock( h );

#ifdef DEBUG
	err =
#endif
	MemHandleResize( h, sizeof( *pAppInfo ) );
	pAppInfo = ThisAppInfoPtr( MemHandleLock( h ) );

	Assert1( !err, "resize failed, err = 0x%x", err );

	idAppInfo = MemHandleToLocalID( h );
	DmSetDatabaseInfo( pdb, 0, 0, 0, 0, 0, 0, 0, &idAppInfo, 0, 0, 0 );

	return pAppInfo;
}

Err
DbOpenHs2::InitAppInfo()
{
	Err		err = 0;
	MemHandle	h;
	LocalID		idAppInfo;
	ThisAppInfoPtr	pAppInfo = 0;
	UInt8		b;
	Boolean		fResize = true;

#ifdef DEBUG
	Assert( this );
	//AssertSz( m_prefs, "pPrefs is null - must contain default values" );
	Assert( m_pdb );
#endif

	// We have a DmOpenRef and we want the database's app info block ID.
	// Get the database's dbID and cardNo and then use them to get the
	// appInfoID.
	CE( DmDatabaseInfo( m_pdb, 0, 0, 0, 0, 0, 0, 0, &idAppInfo, 0, 0, 0 ) );

	if( !idAppInfo )
	{
	    // If no appinfo block exists, create one.
	    fResize = false;
	    DbgMsg(( 3, TEXT( "creating new AppInfo block" ) ));
	    h = DmNewHandle( m_pdb, sizeof( ThisAppInfoType ) );
	    CPE( h );

	    idAppInfo = MemHandleToLocalID( h );
	    DmSetDatabaseInfo( m_pdb, 0, 0, 0, 0, 0, 0, 0, &idAppInfo, 0, 0, 0 );
	    pAppInfo = ThisAppInfoPtr( MemLocalIDToLockedPtr( idAppInfo, CARDNO ) );

	    // Initialize.
	    DmSet( pAppInfo, 0, sizeof( *pAppInfo ), 0 );

	    b = 15;
	    CASSERT( SameSize( pAppInfo->appinfo.lastUniqID, b ) );
	    DmWrite( pAppInfo, OffsetOf( *pAppInfo, appinfo.lastUniqID ), &b, sizeof( b ) );

	    CASSERT( SameSize( pAppInfo->appinfo.categoryUniqIDs[0], b ) );
	    for( b = 0; b < 16; b++ )
		DmWrite( pAppInfo, OffsetOf( *pAppInfo, appinfo.categoryUniqIDs[0] ) + b, &b, sizeof( b ) );

	    TCharToNarrowString( UnfiledString(), RgC( pAppInfo->appinfo.categoryLabels[0] ) );
	}
	else
	{
	    pAppInfo = ThisAppInfoPtr( MemLocalIDToLockedPtr( idAppInfo, CARDNO ) );
	    if( fResize )
		pAppInfo = ResizeAppInfo( m_pdb, pAppInfo );
	}

	// Set up default preferences.
	DmWriteElementEx( pAppInfo, prefs, m_prefs );

	SetDirty();

LError:
	__MemPtrUnlock( pAppInfo );
	return err;
}

Err
DbOpenHs2::ReadAppInfo( UInt16 flags )
{
	Err		err = 0;
	LocalID		idAppInfo = 0;
	ThisAppInfoPtr	pAppInfo = 0;
	DatabasePrefs*	pPrefs;
	UInt16		version;
	Boolean		fResetCounts = false;

	CASSERT( newlistDefault == newlistShopping );

	Assert( this );
	Assert( m_pdb );

	// Defaults.

	pPrefs = &m_prefs;
	ZeroMemory( pPrefs, sizeof( *pPrefs ) );
	Assert( flags >= 0 && flags < newlistMAX );

	//$ WARNING: (chrisant) Watch out for elements added after DBVersion
	// 14, that occur in the first 72 bytes of the struct!  their defaults
	// must be set again later.
	pPrefs->set_category( dmAllCategories );
	pPrefs->set_view( viewAllNotPurchased );
	pPrefs->set_wExportFlags( exportNote );
	pPrefs->rgbRowColor.index = 0;
	pPrefs->rgbRowColor.r = 0xff;
	pPrefs->rgbRowColor.g = 0xff;
	pPrefs->rgbRowColor.b = 0x99;
	pPrefs->rgOrder[0].m_fCompletedAtBottom = true;
	pPrefs->idStore = allStores;
	pPrefs->set_flags( dpfDecimals );
	TCharToNarrowString( DefaultCurrencyString(), RgC( pPrefs->szCurrency ) );
	pPrefs->cMaxLinesPerItem = -1;

	switch( flags )
	{
	case newlistTodo:
	    pPrefs->rgOrder[0].set_m_sort( 0, sortDate );
	    pPrefs->rgOrder[0].set_m_sort( 1, sortPriority );
	    pPrefs->rgOrder[0].set_m_sort( 2, sortDescription );
	    pPrefs->set_columns( colNeed|colPrio|colName|colDate );
	    break;

	case newlistDated:
	    pPrefs->rgOrder[0].set_m_sort( 0, sortDate );
	    pPrefs->rgOrder[0].set_m_sort( 1, sortDescription );
//	    pPrefs->columns = colNeed|colName|colDate;
	    pPrefs->set_columns( colNeed|colName|colPrice|colDate );
	    pPrefs->or_flags( dpfStampToday );
	    break;

	case newlistChecklist:
	    pPrefs->rgOrder[0].set_m_sort( 0, sortDescription );
	    pPrefs->set_columns( colNeed|colName );
	    break;

	case newlistShoppingPerStore:
	    pPrefs->or_flags( dpfPerStorePrices );
	    // fall thru
	case newlistShopping:
	    pPrefs->rgOrder[0].set_m_sort( 0, sortDescription );
//	    pPrefs->columns = colNeed|colName|colPrice;
	    pPrefs->set_columns( colNeed|colName|colQuantity|colPrice );
	    pPrefs->set_wExportFlags( exportQuantity|exportPrice|exportNote );
	    break;

#ifdef DEBUG
	default:
	    Panic1( "ReadAppInfo, newlist flags = %d", flags );
#endif
	}

	MemMove( &pPrefs->rgOrder[1], &pPrefs->rgOrder[0], sizeof( pPrefs->rgOrder[0] ) );

	// Get appinfo block.

	if( m_pdb )
	{
	    CE( DmDatabaseInfo( m_pdb, 0, 0, &version, 0, 0, 0, 0, &idAppInfo, 0, 0, 0 ) );

	    if( !idAppInfo || DBVersion != version )
	    {
		UInt16 cb;
		DatabasePrefs14* p14 = ( DatabasePrefs14* )pPrefs;

		if( version < DBVersion_movedtodatabaseprefs )
		    CE( dmErrDatabaseProtected );

		if( DBVersion != version )
		{
		    SetDirty();

		    // We're upgrading the database, so to preserve the old
		    // version make a backup on the next Flush().
		    m_fMakeBackup = true;

		    if( idAppInfo )
		    {
			pAppInfo = ThisAppInfoPtr( MemLocalIDToLockedPtr( idAppInfo, CARDNO ) );
			CPE( pAppInfo );

			// Try to upgrade.

			cb = 0;
			switch( version )
			{
			case 12:	// version #12 is 72 bytes
					// #12 and earlier had a bug in Undo
					// that didn't correctly update the
					// bold counts.
					// fall thru...
			case 13:	// version #13 is 72 bytes
			case 14:	// version #14 is 72 bytes
			    cb = 72;
			    break;
			case 15:	// version #15 is 92 bytes
			case 16:	// version #16 is 92 bytes
			case 17:	// version #17 is 92 bytes
			case 18:	// version #18 is 92 bytes
			case 19:	// version #19 is 92 bytes
			case 20:	// version #20 is bogus
			case 21:
			case 22:
			case 23:
			case 24:
			    cb = 92;
			    break;

			case 25:
			case 26:
			case 27:
			case 28:
			    cb = 124;
			    break;

			case 29:
			    cb = 154;
			    break;

			CASSERT( DBVersion == 30 );

			    // INSERT PREVIOUS DBVERSION HERE WHEN
			    // INCREMENTING THE CURRENT DBVERSION.

			default:
			    Panic1( "Unknown version %d", version );
			}

			if( version < DBVersion_RECALCITEMCOUNTS )
			    fResetCounts = true;

			if( cb )
			{
			    // Copy the existing prefs into memory.

			    void* pvFrom = &pAppInfo->prefs;
			    if( version < DbVersion_alarmcache )
			    {
				// Prior to the alarm cache changes, the prefs
				// were at a different location.
				pvFrom = &( ( ThisAppInfoType20* )pAppInfo )->prefs;
			    }
			    MemMove( pPrefs, pvFrom, cb );

			    if( version < DBVersion_separatesorts )
			    {
				p14->fSortCompletedAtBottomAll = p14->fSortCompletedAtBottom;
				MemMove( p14->sortAll, p14->sort, sizeof( p14->sort ) );
				MemMove( p14->rgfDescendAll, p14->rgfDescend, sizeof( p14->rgfDescend ) );
			    }

			    if( version < DBVersion_rationalsortinfo )
			    {
				pPrefs->rgOrder[0].m_fCompletedAtBottom = !!p14->fSortCompletedAtBottom;
				pPrefs->rgOrder[1].m_fCompletedAtBottom = !!p14->fSortCompletedAtBottomAll;

				MemMove( pPrefs->rgOrder[0].addr_m_sort(), p14->sort, sizeof( p14->sort ) );
				MemMove( pPrefs->rgOrder[1].addr_m_sort(), p14->sortAll, sizeof( p14->sortAll ) );
				MemMove( pPrefs->rgOrder[0].m_rgfDescend, p14->rgfDescend, sizeof( p14->rgfDescend ) );
				MemMove( pPrefs->rgOrder[1].m_rgfDescend, p14->rgfDescendAll, sizeof( p14->rgfDescendAll ) );

				//$ WARNING: (chrisant) If any of these are
				// reused in the future, the defaults must be
				// reset somewhere.
				pPrefs->zero_reserved2();
				pPrefs->fontPlusOne = 0;
				pPrefs->zero_reserved5();
				pPrefs->zero_reserved6();
			    }

			    if( version < DbVersion_rowcolor )
			    {
				pPrefs->rgbRowColor.index = 0;
				pPrefs->rgbRowColor.r = 0xff;
				pPrefs->rgbRowColor.g = 0xff;
				pPrefs->rgbRowColor.b = 0x99;
			    }
			    if( version < DbVersion_alarmcache )
				DmSet( pAppInfo, OffsetOf( *pAppInfo, alarmCache ), sizeof( pAppInfo->alarmCache ), 0 );

			    if( version < DbVersion_categoryicons )
				pPrefs->zero_rgCategoryIcons_deprecated();

			    if( version < DbVersion_customcategories )
				pPrefs->fCustomCategories = false;

			    if( version < DbVersion_persistmultiples )
			    {
				pPrefs->m_fMultipleCategories = false;
				pPrefs->m_fMultipleCategories_Unfiled = false;
				pPrefs->m_fMultipleStores = false;
				pPrefs->m_fMultipleStores_Unfiled = false;
				ZeroMemory( pPrefs->m_rggrfMultipleCategories, sizeof( pPrefs->m_rggrfMultipleCategories ) );
				ZeroMemory( pPrefs->m_rggrfMultipleStores, sizeof( pPrefs->m_rggrfMultipleStores ) );
			    }

			    if( version < DbVersion_autoorder )
			    {
				ZeroMemory( pPrefs->m_rgfAutoOrder, sizeof( pPrefs->m_rgfAutoOrder ) );
				pPrefs->m_fLearning = false;
			    }
			}

			// Unlock so we can call InitAppInfo (which resizes
			// the appinfo block).
			__MemPtrUnlock( pAppInfo );
			pAppInfo = 0;
		    }

		    // Stamp database with current version.
		    version = DBVersion;
		    DmSetDatabaseInfo( m_pdb, 0, 0, &version, 0, 0, 0, 0, 0, 0, 0, 0 );
		}
		Assert( !pAppInfo );
		CE( InitAppInfo() );
		if( fResetCounts )
		    m_prefs.or_flags( dpfResetCounts );
	    }
	    else
	    {
		// Appinfo exists, and database is correct version.
		pAppInfo = ThisAppInfoPtr( MemLocalIDToLockedPtr( idAppInfo, CARDNO ) );
		CPE( pAppInfo );

		// Read info.
		*pPrefs = pAppInfo->prefs;
	    }

	    // Finally initialize side data structures to reflect the prefs
	    // state.

	    m_prefs.GetMultipleCategories( &m_multiCategories );
	    m_prefs.GetMultipleStores( &m_multiStores );
	}

LError:
	__MemPtrUnlock( pAppInfo );
	return err;
}

void
DbOpenHs2::WriteAppInfo()
{
	LocalID		idAppInfo;
	ThisAppInfoPtr	pAppInfo = 0;
	UInt16		version;

	Assert( this );

	if( m_pdb )
	{
	    Assert( !m_fReadonly );

	    CheckAppInfoLocks( m_pdb, 3, __FILE__, __LINE__ );

	    //$ review: (chrisant) Wish I could verify popen->m_prefs somehow.
	    CEv( m_pdb->DmDatabaseInfo( 0, 0, &version, 0, 0, 0, 0, &idAppInfo, 0, 0, 0 ) );

	    Assert( DBVersion == version );
	    Assert( idAppInfo );

	    pAppInfo = ThisAppInfoPtr( MemLocalIDToLockedPtr( idAppInfo, CARDNO ) );
	    CPEv( pAppInfo );

	    // BackupBuddy Certification, only write the appinfo if it's
	    // changed.

	    if( DBVersion != version ||
		MemPtrSize( pAppInfo ) != sizeof( *pAppInfo ) ||
		MemCmp( &pAppInfo->prefs, &m_prefs, sizeof( m_prefs ) ) != 0 )
	    {
		SetDirty();

		// Resize, in case we upgraded the db version.
		pAppInfo = ResizeAppInfo( m_pdb, pAppInfo );

		// FORCE RESORTING ON THE PALM DEVICE.
#ifndef OS_PALM
		m_prefs.or_flags( dpfForceResort );
#endif

		// Write per-database info.
		DmWriteElementEx( pAppInfo, prefs, m_prefs );

		CheckAppInfoLocks( m_pdb, 3, __FILE__, __LINE__ );
	    }
LError:
	    __MemPtrUnlock( pAppInfo );
	}
}

void
DbOpenHs2::IncrementChangeCounter()
{
	m_nChangeCounter++;

	if( m_nChangeCounter < 0 )
	    m_nChangeCounter = 0;
}

/*
 * Pack and Unpack.
 */

#ifdef DEBUG
static UInt16 s_cUnpack;

void
DbgResetUnpackCounter()
{
	s_cUnpack = 0;
}

UInt16
DbgGetUnpackCounter()
{
	return s_cUnpack;
}
#endif

Item*
DbOpenHs2::UnPackRecord( unsigned record ) const
{
	Assert( this );
	Assert( m_pdb );
	Assert( record != noRecordSelected );

	MemHandle hRec;
	ItemPtr pitemRet = 0;
	SPNew<Item> spitem;
	unsigned dmRecord = record += m_cMetadataRecords;

	TraceDB(( 5, TEXT( "unpacking record %u" ), record ));

	hRec = DmQueryRecord( m_pdb, dmRecord );

	if( hRec )
	{
	    spitem = new Item;

	    if( spitem )
	    {
		pitemRet = UnPackRecordPtr( MemHandleLock( hRec ), maskALL, spitem, record );

		if( !pitemRet )
		    Panic1( "Unable to unpack record %u.", record );
		else
		    spitem.Transfer();

		MemHandleUnlock( hRec );
	    }
	}

	CheckLocks( m_pdb, __FILE__, __LINE__ );

	return pitemRet;
}

typedef Err ( *MapFnPtr )( MemPtr pv, MemPtr pRec, UInt16 *pOffset, HS2XferMode mode );

#define cchMAXDBGNAME 32
struct MapEntry
{
	MaskType	m_mask;
	MapFnPtr	m_pfn;
	UInt16		m_offset;
#ifdef DEBUG
	UInt16		m_tagDbg;
//	char		m_szDbgName[cchMAXDBGNAME];
#endif
};

CASSERT( sizeof( "hello" ) == 6 );

#ifdef OS_PALM

#ifdef DEBUG
#define TAGMAP(_tag, _xfer, _member) dc.l mask##_tag; jmp _xfer; dc.w struct(ItemType._member); dc.w tag##_tag; //dc.b #_tag; ds.b cchMAXDBGNAME-sizeof(#_tag);
#define TAGMAP2(_tag, _mask, _xfer, _member) dc.l mask##_mask; jmp _xfer; dc.w struct(ItemType._member); dc.w tag##_tag; //dc.b #_tag; ds.b cchMAXDBGNAME-sizeof(#_tag);
#define TAGMAPex(_tag, _xfer, _offset) dc.l mask##_tag; jmp _xfer; dc.w _offset; dc.w tag##_tag; //dc.b #_tag; ds.b cchMAXDBGNAME-sizeof(#_tag);
#else
#define TAGMAP(_tag, _xfer, _member) dc.l mask##_tag; jmp _xfer; dc.w struct(ItemType._member);
#define TAGMAP2(_tag, _mask, _xfer, _member) dc.l mask##_mask; jmp _xfer; dc.w struct(ItemType._member);
#define TAGMAPex(_tag, _xfer, _offset) dc.l mask##_tag; jmp _xfer; dc.w _offset;
#endif

// This asm voodoo is necessary in order for it to work when Globals are not
// available, for instance during sysAppLaunchCmdFind.
#define BEGIN_MAPTABLE	static void asm MapTable(void) {
#define END_MAPTABLE	}
#define c_rgMap		( reinterpret_cast<MapEntry const*>( MapTable ) )

#else // !OS_PALM

#ifdef DEBUG
#define TAGMAP( _tag, _xfer, _member ) { mask##_tag, MapFnPtr( _xfer ), UInt16( OffsetOf0( ItemType, _member ) ), tag##_tag },
#define TAGMAP2( _tag, _mask, _xfer, _member ) { mask##_mask, MapFnPtr( _xfer ), UInt16( OffsetOf0( ItemType, _member ) ), tag##_tag },
#define TAGMAPex( _tag, _xfer, _offset ) { mask##_tag, MapFnPtr( _xfer ), UInt16( _offset ), tag##_tag },
#define TAGMAPex2( _tag, _mask, _xfer, _offset ) { mask##_mask, MapFnPtr( _xfer ), UInt16( _offset ), tag##_tag },
#else
#define TAGMAP( _tag, _xfer, _member ) { mask##_tag, MapFnPtr( _xfer ), UInt16( OffsetOf0( ItemType, _member ) ) },
#define TAGMAP2( _tag, _mask, _xfer, _member ) { mask##_mask, MapFnPtr( _xfer ), UInt16( OffsetOf0( ItemType, _member ) ) },
#define TAGMAPex( _tag, _xfer, _offset ) { mask##_tag, MapFnPtr( _xfer ), UInt16( _offset ) },
#define TAGMAPex2( _tag, _mask, _xfer, _offset ) { mask##_mask, MapFnPtr( _xfer ), UInt16( _offset ) },
#endif

#define BEGIN_MAPTABLE	static MapEntry c_rgMap[] = {
#define END_MAPTABLE	};

#endif // !OS_PALM

BEGIN_MAPTABLE
TAGMAPex( END, 0, -1 )
TAGMAPex( Description, XferString, Item::offsetDescription )
TAGMAPex( Note, XferString, Item::offsetNote )
TAGMAP( Flags, XferUInt16, m_flags )
TAGMAPex( Quantity, XferUInt32, Item::offsetQuantity )
TAGMAPex( Price, XferUInt32, Item::offsetPrice )
TAGMAPex( Priority, XferUInt8, Item::offsetPriority )
TAGMAP( Date, XferUInt16, m_date )
TAGMAPex( Alarm, XferUInt32, Item::offsetAlarm )
TAGMAPex( Units, XferUnits4, Item::offsetUnits )
TAGMAPex( MasterStoreList, 0, -1 )
TAGMAPex( Stores, XferStores, 0 )
TAGMAPex( Aisle, XferAisle1, Item::offsetAisle )
TAGMAPex( EntryOrder, XferUInt32, Item::offsetEntryOrder )
TAGMAPex( PerStore, XferPerStore, 0 )
TAGMAPex( CustomText, XferString, Item::offsetCustomText )
TAGMAPex( AlarmMidi, XferUInt32, Item::offsetAlarmMidi )
TAGMAPex( Icon, XferUInt16, Item::offsetIcon )
TAGMAPex2( UnitsEx, Units, XferUnitsEx, Item::offsetUnits )
TAGMAPex2( AisleEx, Aisle, XferAisleEx, Item::offsetAisle )
TAGMAPex2( PerStoreEx, PerStore, XferPerStoreEx, 0 )
TAGMAPex( CategoryList, 0, -1 )
TAGMAPex2( Category, CATEGORY, XferCategory, Item::offsetCategory )
TAGMAPex( AutoOrder, XferAutoOrder, Item::offsetAutoOrder )
END_MAPTABLE

#ifdef OS_PALM

inline MapEntry const*
Glue_GetMapEntry( UInt8 tag )
{
	return &reinterpret_cast<MapEntry const*>( MapTable ) )[tag];
}

MapEntry const*
GetMapEntry( UInt8 tag )
{
	return Glue_GetMapEntry( tag );
}

inline MapFnPtr
Glue_GetFunctionPtr( UInt16 index )
{
	register MapEntry const* pme = Glue_GetMapEntry( index );
	return MapFnPtr( ( ( UInt8* )pme ) + pme->m_jmp.addressOffset + 2 + sizeof( JmpInstructionType ) );
}

MapFnPtr
GetFunctionPtr( UInt16 index )
{
	return Glue_GetFunctionPtr( index );
}

#else // !OS_PALM

inline const MapEntry*
GetMapEntry( UInt8 tag )
{
	return &c_rgMap[tag];
}

inline MapFnPtr
GetFunctionPtr( UInt16 index )
{
	return c_rgMap[index].m_pfn;
}

#endif // !OS_PALM

#ifdef TraceTAGs
#undef TraceTAG
#define TraceTAG( tag ) DbgMsg(( 5, "found %hs", GetMapEntry( tag )->m_szDbgName ))
#endif

ItemPtr
UnPackRecordPtr( MemPtr pRec, MaskType mask, ItemPtr pitem, unsigned recDebug )
{
	ItemPtr		pitemRet = 0;
	UInt16		wOffset = 0;
	UInt8		tag;
	UInt16		flags = 0;

	SUPPRESS_UNUSED_VAR( recDebug );

	Assert( mask != 0 );

#ifdef DEBUG
	if( IfGlobals() )
	{
	    static Boolean fVerified = false;
	    if( !fVerified )
	    {
		fVerified = true;
		for( tag = 0; tag < tagMAX; ++tag )
		    Assert( GetMapEntry( tag )->m_tagDbg == tag );
	    }

	    s_cUnpack++;
	}

	Assert( pitem );
	pitem->AssertClear();
#endif

	pitem->m_dirtySort = 0;

	if( tagMasterStoreList == *( ( UInt8* )pRec ) || tagCategoryList == *( ( UInt8* )pRec ) )
	{
	    flags |= flagMetadataRecord;
	    goto Done;
	}

#ifdef DEBUG
	if( mask == maskALL )
	    flags |= flagCANPACK;
#endif
	mask &= maskKNOWN;

	do
	{
	    tag = *( ( UInt8* )pRec + wOffset );
	    ++wOffset;

	    Assert1( tag < tagMAX, "bogus tag encountered (0x%x)", UInt16( tag ) );
	    TraceTAG( tag );
	    if( tag >= tagMAX )
	    {
		Panic2( "Record %u corrupt at byte %u.\n", recDebug, wOffset - 1 );
		goto LError;
	    }

	    if( tag == tagEND )
		break;

	    MapEntry const* const pme = GetMapEntry( tag );
	    HS2XferMode const mode = ( mask & pme->m_mask ) ? xferLoad : xferSkip;

	    CEv( ( GetFunctionPtr( tag ) )( ( void* )( ( ( UInt8* )pitem ) + pme->m_offset ), pRec, &wOffset, mode ) );

	    mask &= ~pme->m_mask;

#ifdef DEBUG
	    if( tag == tagDescription && mode == xferLoad )
	    {
		Assert( pitem->GetDescription()->Length() );
		TraceDB(( 5, TEXT( "loaded \"%s\"" ), pitem->GetDescription()->Text() ));
	    }
#endif

	    switch( tag )
	    {
	    case tagAlarm:
		flags |= flagHasAlarm;
		pitem->InitAlarmOrig();
		break;
	    case tagCategory:
		//$ review: (chrisant) Ideally we should assert this, but we
		// don't have the database pointer so we can't get at the
		// pref.
		//Assert( m_prefs.fCustomCategories );
		flags |= flagCATEGORY;
		break;
	    }
	}
	while( mask );

	if( pitem->GetState() == ITEMSTATE_NEED )
	    pitem->m_flags |= flagWasNeeded;

Done:
	pitem->m_flags |= flags;

#ifdef DEBUG
	AutoOrderListValidity const aol = pitem->TestAutoOrderListValid();
	Assert1( aol == aolValid, "CORRUPT:  AutoOrderList invalid code %u", aol );
#endif

	pitemRet = pitem;
	pitem = 0;

LError:
	if( pitem )
	    pitem->Clear();
	if( pitemRet )
	    Assert( ( pitemRet->GetState() == ITEMSTATE_NEED ) == !!( pitemRet->m_flags & flagWasNeeded ) );
	return pitemRet;
}

bool
DbOpenHs2::FAlreadySorted( unsigned dmRecord )
{
	Assert( dmRecord >= m_cMetadataRecords );

	bool			fRet = false;
	MemHandle		hRec = DmQueryRecord( m_pdb, dmRecord );
	MemPtr			pRec = MemHandleLock( hRec );
	MemHandle		h;
	Int16			nPrev;
	Int16			nNext;

	// Compare to preceding record.

	nPrev = -1;			// Assume we're not sorted.
	if( dmRecord == m_cMetadataRecords )
	    nPrev = 1;
	else
	{
	    dmRecord--;
	    h = DmQueryRecord( m_pdb, dmRecord );
	    if( h )
	    {
		nPrev = CmpFindSort( pRec, MemHandleLock( h ), 0 );
		MemHandleUnlock( h );
	    }
	    dmRecord++;
	}

	// If 'pitem' < the preceding record, then clearly we're not sorted.

	if( nPrev < 0 )
	    goto LOut;

	// Compare to following record.

	nNext = 1;			// Assume we're not sorted.
	if( dmRecord < DmNumRecords( m_pdb ) )
	{
	    h = DmQueryRecord( m_pdb, dmRecord );
	    if( h )
	    {
		nNext = CmpFindSort( pRec, MemHandleLock( h ), 0 );
		MemHandleUnlock( h );
	    }
	}

	// We are sorted iff: equal to both neighbors, or sorted relative to
	// neighbors but not equal to either neighbor.

	fRet = ( nPrev > 0 && nNext < 0 ) || ( nPrev == 0 && nNext == 0 );

LOut:
	MemHandleUnlock( hRec );
	return fRet;
}

SysError
DbOpenHs2::PackRecord( unsigned record, ItemPtr* ppitem, UInt16 wPackFlags, UInt16* pces )
{
	Err		err = 0;
	UInt16		cbSize;
	MemHandle	hRec = 0;
	MemPtr		pRec = 0;
	UInt16		wOffset = 0;
	ItemPtr		pitem = 0;
	bool const	fUpdateCounts = !( wPackFlags & packNoCounts );
	bool const	fCanUndo = !( wPackFlags & packNoUndo );
#ifdef DEBUG
	bool const	fUndoing = !!( wPackFlags & packUndoing );
#endif
	UInt16		attr;
	UInt32		uid = 0;
	bool		fMove;
	bool		fWasNeeded;
	bool		fDelete = false;
	UInt16		flags;
	ItemPtr		pitemUndo = 0;
	UInt16		ces;
	const MapEntry*	pme;
	UInt8*		pTag;
	UInt8		rgTags[tagMAX];
	bool const	fExists = ( ( record != noRecordSelected ) && !( wPackFlags & packInsertAtIndex ) );
	unsigned	dmRecord = ( record != noRecordSelected ) ? record + m_cMetadataRecords : record;

	CASSERT( SameSize( flags, ( *ppitem )->m_flags ) );

#ifdef DEBUG
	Assert( m_pdb );
	Assert( !m_fReadonly );
	Assert( m_prefs.fCustomCategories );
	Assert( ppitem );
	Assert( ( !*ppitem ) || ( ( *ppitem )->m_flags & flagCANPACK ) );
	Assert( FImplies( fUndoing, !fCanUndo ) );
	Assert( FImplies( fCanUndo, !fUndoing ) );

//	DbgMsg(( 5, "record %u\n  %d == packForceSort\n  %d == packNoSort",
//		 record, F( wPackFlags & packForceSort ), F( wPackFlags & packNoSort ) ));

	AssertNoRecordLocks( TEXT( "before pack" ) );

#ifdef TraceTAGs
	#undef TraceTAG
	#define TraceTAG( x ) DbgMsg(( 5, "writing %hs", #x ))
#endif // TraceTAGs
#endif // DEBUG

	if( !pces )
	    pces = &ces;
	*pces = 0;

	if( !*ppitem )
	{
	    // Just free ppitem.
	    goto LError;
	}

	SetDirty();

	TraceDB(( 5, TEXT( "PACKING record %d" ), record ));

	pitem = *ppitem;
	fWasNeeded = fExists && ( pitem->m_flags & flagWasNeeded );

	if( fExists )
	{
	    DmRecordInfo( m_pdb, dmRecord, &attr, &uid, 0 );

	    // Try to prevent accidental overwrite/deletion of Stores list.
	    Assert( !hRec );
	    Assert( !pRec );
	    hRec = DmQueryRecord( m_pdb, dmRecord );
	    pRec = MemHandleLock( hRec );
	    AssertSz( tagMasterStoreList != *( UInt8* )pRec, "bad write to Stores list" );
	    AssertSz( tagCategoryList != *( UInt8* )pRec, "bad write to Categories list" );
	    MemHandleUnlock( hRec );
	    pRec = 0;
	    hRec = 0;
	}

	if( fCanUndo && fExists )
	{
	    if( uid != 0 )
	    {
#if 0
LRetry:
#endif
		pitemUndo = UnPackRecord( record );

#if 0
		if( !pitemUndo && DBClearOneUndo( this ) )
		    goto LRetry;
#endif
	    }

	    if( !pitemUndo )
		ClearUndo();
	}

	// Maybe delete record.

	if( pitem->FEmpty() && !( wPackFlags & packNoDelete ) )
	{
	    // The actual deletion occurs further below...

	    fDelete = true;
	    *pces |= cesDeleted;
	    m_recAfterPack = noRecordSelected;

	    if( !fExists )
		goto LError;	// Can't both create and delete a record in the same call.

	    goto LSetupUndo;
	}

	// Fixup, in case date is accidentally 0.

	if( !DateToInt( pitem->m_date ) )
	    *( UInt16* )&pitem->m_date = dateNone;

	// Calc size for record.

	CASSERT( SameSize( pitem->m_date, UInt16 ) );
	cbSize = 1 + sizeof( pitem->m_flags ) +
		 1 + sizeof( pitem->m_prio ) +
		 1 + sizeof( pitem->m_ulEntryOrder ) +
		 1;
	cbSize += UInt16( 0 ) + CbXferString( pitem->GetDescription() );
	cbSize += UInt16( 0 ) + CbXferString( pitem->GetCustomText() );
	cbSize += UInt16( 0 ) + CbXferString( pitem->GetNote() );
	cbSize += UInt16( 0 ) + CbXferString( pitem->GetAisle() );
	if( DateToInt( pitem->m_date ) != dateNone )
	    cbSize += 1 + sizeof( pitem->m_date );
	if( pitem->m_flags & flagHasAlarm )
	    cbSize += 1 + sizeof( pitem->m_alarm );
	if( pitem->m_uidAlarmMidi )
	    cbSize += 1 + sizeof( pitem->m_uidAlarmMidi );
	if( pitem->m_lPrice )
	    cbSize += 1 + sizeof( pitem->m_lPrice );
	if( pitem->m_lQuantity )
	    cbSize += 1 + sizeof( pitem->m_lQuantity );
	if( pitem->GetUnits()->Length() )
	    cbSize += UInt16( 1 + 1 ) + UInt16( pitem->GetUnits()->Length() ) + UInt16( 1 );
	pitem->m_flags &= ~flagHasStore;
	if( pitem->m_stores.Count() )
	{
	    cbSize += 1 + 1 + pitem->m_stores.Count();
	    pitem->m_flags |= flagHasStore;		// IMPORTANT!
	}
	if( pitem->m_perstore.Count() )
	    cbSize += 1 + 1 + pitem->m_perstore.Count() * sizeof( PerStore2 );
	if( pitem->m_icon )
	    cbSize += 1 + sizeof( pitem->m_icon );
	if( m_prefs.fCustomCategories && pitem->m_category != catUnfiledCategory )
	    cbSize += 1 + sizeof( UInt8 );
	if( pitem->CountAutoOrders() )
	    cbSize += 1 + sizeof( UInt8 ) + CbAutoOrders( pitem->CountAutoOrders() );

	// Entry order.

	if( !pitem->m_ulEntryOrder )
	{
	    m_prefs.set_ulEntryOrder( m_prefs.get_ulEntryOrder() + 1 );
	    pitem->m_ulEntryOrder = m_prefs.get_ulEntryOrder();
	}
	else if( !fExists && pitem->m_ulEntryOrder > m_prefs.get_ulEntryOrder() )
	{
	    m_prefs.set_ulEntryOrder( pitem->m_ulEntryOrder );
	}

	// Ensure record exists.

	if( !fExists )
	{
	    hRec = DmNewRecord( m_pdb, &dmRecord, cbSize );
	    Assert2( hRec, "error creating record (cbSize=%d, err=%d)", cbSize, m_pdb->DmGetLastErr() );

	    record = dmRecord - m_cMetadataRecords;
	}
	else
	{
	    // Make sure record is at least big enough to hold the new
	    // contents (but don't shrink it yet; in case something crashes
	    // below let's be as non-destructive as possible along the way).

	    hRec = DmQueryRecord( m_pdb, dmRecord );

	    if( hRec && MemHandleSize( hRec ) < cbSize )
		hRec = DmResizeRecord( m_pdb, dmRecord, cbSize );

	    if( hRec )
		hRec = DmGetRecord( m_pdb, dmRecord );
	}

	if( !hRec )
	{
	    err = m_pdb->DmGetLastErr();
	    goto LError;
	}

	// Get attributes and unique ID.

	DmRecordInfo( m_pdb, dmRecord, &attr, &uid, NULL );

	// Lock down pointer.

	pRec = MemHandleLock( hRec );

	// Write flags (special because masked against flagMaskPersisted).

	TraceTAG( tagFlags );
	rgTags[0] = tagFlags;
	XferUInt8( rgTags, pRec, &wOffset, xferSave );
	flags = pitem->m_flags & flagMaskPersisted;
	XferUInt16( &flags, pRec, &wOffset, xferSave );

	// Build list of other things to write out.

	pTag = rgTags;
	// ---- BEGIN IMMUTABLE ORDER ----
	if( pitem->m_stores.Count() )
	    *( pTag++ ) = tagStores;		// Stores go second to boost performance of GetFilteredRecord.
	if( pitem->m_perstore.Count() )
	    *( pTag++ ) = tagPerStoreEx;	// Perstores go third to help ZapStore.
	if( pitem->m_flags & flagHasAlarm )
	    *( pTag++ ) = tagAlarm;		// Alarm goes fourth to help Alarm functions.
	if( pitem->m_uidAlarmMidi )
	    *( pTag++ ) = tagAlarmMidi;		// Alarm sound goes fifth to help Alarm functions.
	if( pitem->m_icon )
	    *( pTag++ ) = tagIcon;		// And the icon goes next, also to help Alarm functions.
	*( pTag++ ) = tagEntryOrder;
	*( pTag++ ) = tagPriority;
	if( m_prefs.fCustomCategories && pitem->m_category != catUnfiledCategory )
	    *( pTag++ ) = tagCategory;
	if( pitem->GetDescription()->Length() )
	{
	    // Ideally we'd have liked description to come immediately after
	    // the Alarm, to help the Alarm functions, but entryorder and
	    // priority already came ahead of it.
	    *( pTag++ ) = tagDescription;
	    TraceDB(( 5, TEXT( "wrote \"%s\"" ), pitem->GetDescription()->Text() ));
	}
	// ---- END IMMUTABLE ORDER ----
	if( pitem->GetCustomText()->Length() )
	    *( pTag++ ) = tagCustomText;
	if( pitem->m_lQuantity )
	    *( pTag++ ) = tagQuantity;
	if( pitem->GetUnits()->Length() )
	    *( pTag++ ) = tagUnitsEx;
	if( *( UInt16* )&pitem->m_date != dateNone )
	    *( pTag++ ) = tagDate;
	if( pitem->m_lPrice )
	    *( pTag++ ) = tagPrice;
	if( pitem->GetAisle()->Length() )
	    *( pTag++ ) = tagAisleEx;
	if( pitem->GetNote()->Length() )
	    *( pTag++ ) = tagNote;
	if( pitem->CountAutoOrders() )
	    *( pTag++ ) = tagAutoOrder;
	*pTag = tagEND;

	// Write out data.

	pTag = rgTags;
	while( 1 )
	{
	    TraceTAG( *pTag );

	    XferUInt8( pTag, pRec, &wOffset, xferSave );
	    if( *pTag == tagEND )
		break;

	    pme = GetMapEntry( *pTag );
	    ( GetFunctionPtr( *pTag ) )( MemPtr( ( ( UInt8* )pitem ) + pme->m_offset ), pRec, &wOffset, xferSave );

	    ++pTag;
	}

	Assert2( wOffset == cbSize, "wOffset == %d\ncbSize == %d", wOffset, cbSize );

	// Resize the record to shrink it if necessary.

	MemHandleUnlock( hRec );
	pRec = 0;

	DmReleaseRecord( m_pdb, dmRecord, true );

	hRec = DmResizeRecord( m_pdb, dmRecord, cbSize );
	Assert3( hRec, "error resizing record %d (cbSize=%d, err=%d)", record, cbSize, m_pdb->DmGetLastErr() );

	hRec = 0;

	// Update the bit table.

	if( m_pbits )
	{
	    if( !fExists )
		m_pbits->NewRecord( record, true );

	    m_pbits->RefreshRecord( record );

	    if( !m_pbits->InView( record ) )
		*pces |= cesFiltered;
	}

	// Move the record into sorted position.  Doing this after having
	// written the record allows us to reuse the same optimized comparison
	// function used by SortDatabase.

	if( wPackFlags & packForceSort )
	{
	    TraceDB(( 5, TEXT( "force sort" ) ));
	    fMove = true;
	}
	else if( wPackFlags & packNoSort )
	{
	    TraceDB(( 5, TEXT( "no sort" ) ));
	    fMove = false;
	}
	else if( fExists )
	{
	    TraceDB(( 5, TEXT( "stable sort" ) ));
	    // m_dirtySort allows us to only sort when one of the sort key
	    // fields has been modified.
	    fMove = !!( GetDirtyMask() & pitem->m_dirtySort );
	}
	else
	{
	    TraceDB(( 5, TEXT( "new item sort" ) ));
	    fMove = !( wPackFlags & packInsertAtIndex );
	}

	TraceDB(( 5, TEXT( "fMove == %d" ), fMove ));
	if( fMove )
	{
	    unsigned const recordFrom = record;
	    SysError const eSort = SortRecord( recordFrom, &record );

	    //$ review: (chrisant) HS2 assumed this always succeeded, and I'm
	    // not going to address this in the legacy HS2 PDB format code
	    // since it's just shuffling things around in memory.
	    Assert( !eSort );
	    SUPPRESS_UNUSED_VAR( eSort );

	    dmRecord = record + m_cMetadataRecords;

	    if( recordFrom != record )
		*pces |= cesMoved;
	}

	// PackRecord doesn't return the updated index, instead it can be
	// retrieved via GetRecordAfterPack.

	m_recAfterPack = record;

LSetupUndo:
	Assert( !( !fExists && fDelete ) );

	// Update bold counts for stores if appropriate.

	bool const fIsNeeded = ( pitem->GetState() == ITEMSTATE_NEED );

	if( fUpdateCounts &&
	    ( pitem->m_fStoresDIRTY || fDelete || fWasNeeded != fIsNeeded ) )
	{
	    ASSERTGLOBALS;

	    // Update bold counts for stores.

	    if( fWasNeeded )
	    {
		StoreList *p;

		if( pitem->m_fStoresDIRTY )
		{
		    p = &pitem->m_storesORIG;
		    TraceCounts(( 5, TEXT( "UpdateCounts( -1 )\nORIG stores = %d" ), p->Count() ));
		}
		else
		{
		    p = &pitem->m_stores;
		    TraceCounts(( 5, TEXT( "UpdateCounts( -1 )\nstores = %d" ), p->Count() ));
		}

		UpdateCounts( p, -1 );
	    }

	    if( fIsNeeded && !fDelete )
	    {
		TraceCounts(( 5, TEXT( "UpdateCounts( +1 )\nstores = %d" ), pitem->m_stores.Count() ));
		UpdateCounts( &pitem->m_stores, 1 );
	    }
	}

	// The call to SetupUndo() must occur before actually deleting the
	// record, otherwise it remembers the wrong record and doesn't restore
	// it correctly.

	if( fCanUndo )
	{
	    if( !fExists || ( pitemUndo && !pitemUndo->FEmpty() ) )
		SetupUndo( &pitemUndo, record, fDelete, !fExists );
	}

	if( fDelete && fExists )
	{
	    DbgMsg(( 5, TEXT( "deleting record %d" ), record ));
	    DmRemoveRecord( m_pdb, dmRecord );
	    if( m_pbits )
		m_pbits->RemoveRecord( record );
	}

LError:
#ifdef DEBUG
	if( err )
	    DbgMsg(( 12, TEXT( "PackRecord failure:\n  err = %d" ), m_pdb->DmGetLastErr() ));
#endif
	if( pitemUndo )
	{
	    delete pitemUndo;
	    pitemUndo = 0;
	}
	Assert( !pRec );
	Assert( !hRec );
	if( pitem && !err && ( fDelete || ( pitem->m_dirtySort & dirtyAlarm ) ) )
	{
	    if( !fDelete && !!( pitem->m_flags & flagHasAlarm ) )
		UpdateAlarm( m_pdb, pitem->m_alarm, true );
	    else
	    {
		UpdateAlarm( m_pdb, pitem->m_alarmORIG, false );
#if 0
		DeleteAlarmIfPosted( m_pdb, uid );
#endif
	    }
	}

	delete *ppitem;
	*ppitem = 0;

#ifdef DEBUG
	AssertNoRecordLocks( TEXT( "after pack" ) );
	CheckLocks( m_pdb, __FILE__, __LINE__ );
#endif

	//$ todo: (chrisant) Return a more specific error.
	return err ? SE_UNKNOWN : SE_OK;
}

bool
DbOpenHs2::GetRecordUniqueId( unsigned record, DWORD* pdwUniqueId ) const
{
	Assert( pdwUniqueId );
	Assert( noRecordSelected != record );
	Assert( int( record ) >= 0 && record < Count() );

	UInt32 uniqueId;
	UInt16 const dmRecord = UInt16( record + m_cMetadataRecords );

	CASSERT( SameSize( uniqueId, *pdwUniqueId ) );

	if( m_pdb->DmRecordInfo( dmRecord, 0, &uniqueId, 0 ) )
	    return false;

	*pdwUniqueId = uniqueId;

	return true;
}

bool
DbOpenHs2::FindRecordByUniqueId( DWORD dwUniqueId, unsigned* precord ) const
{
	Assert( precord );

	CASSERT( SameSize( UInt32, dwUniqueId ) );

	if( m_pdb->DmFindRecordByID( UInt32( dwUniqueId ), precord ) )
	    return false;

	Assert( *precord >= m_cMetadataRecords );

	*precord -= m_cMetadataRecords;

	return true;
}

bool
DbOpenHs2::GetItemQuickInfo( unsigned record, ItemQuickInfo* pqi ) const
{
	Assert( pqi );
	Assert( noRecordSelected != record );
	Assert( int( record ) >= 0 && record < Count() );

	Err err;
	UInt16 const dmRecord = UInt16( record + m_cMetadataRecords );
	UInt16 const flags = GetItemFlags( dmRecord, &err );

	if( err && dmErrRecordDeleted != err )
	    return false;

	Assert( !( flags & flagMetadataRecord ) );

	if( flags & flagMetadataRecord )
	    return false;

	pqi->Initialize( flags, ( dmErrRecordDeleted == err ) );

	return true;
}

/*
 * Alarms.
 */

enum CheckNextAlarmType
{
	HotSyncCheck,			// Try to be low impact.
	NormalCheck,			// Rescan if db has an alarm.
	ForceCheck,			// Rescan no matter what.
};

static UInt32
GetRecordAlarm( DmOpenRef pdb, unsigned record, UInt32 currentTime )
{
	UInt32 alarm = 0;
	MemHandle h = DmQueryRecord( pdb, record );

	if( h )
	{
	    UInt8* const pRecBegin = ( UInt8* )MemHandleLock( h );

#if 0
	    DbgMsg_Alarms(( 5, "scan rec %u", record ));
#endif

	    if( tagMasterStoreList != pRecBegin[0] &&
		tagCategoryList != pRecBegin[0] )
	    {
		MemPtr const p = GetMemberPtr( pRecBegin, tagAlarm );

		if( p )
		{
		    memcpy( &alarm, p, sizeof( alarm ) );

#if 0
		    DbgMsg_Alarms(( 30, "record %u,\nalarm = %lu\ncurrentTime = %lu\n%s",
				    record, alarm, currentTime, alarm < currentTime ? "clear" : "OK" ));
#endif

		    if( alarm < currentTime )
			alarm = 0;
		}
	    }

	    MemHandleUnlock( h );
	}

	return alarm;
}

static UInt32
CheckRecordsForNextAlarm( DmOpenRef pdb, UInt32 currentTime )
{
	UInt32 nextAlarm = 0;

	for( unsigned record = DmNumRecords( pdb ); record--; )
	{
	    register UInt32 alarm;

	    alarm = GetRecordAlarm( pdb, record, currentTime + 1 );

	    if( alarm && ( !nextAlarm || alarm < nextAlarm ) )
		nextAlarm = alarm;
	}

	return nextAlarm;
}

void
CheckOneDatabaseForNextAlarm( DmOpenRef pdb, UInt32 currentTime, UInt32* pnextAlarm, CheckNextAlarmType check )
{
	UInt16 version;
	LocalID idAppInfo;

	if( !DmDatabaseInfo( pdb, 0, 0, &version, 0, 0, 0, 0, &idAppInfo, 0, 0, 0 ) &&
	    version >= DbVersion_alarmcache && idAppInfo )
	{
	    ThisAppInfoPtr const pAppInfo = ( ThisAppInfoPtr )MemLocalIDToLockedPtr( idAppInfo, CARDNO );
	    register UInt32 alarm = pAppInfo->alarmCache.get_m_seconds();

	    if( alarm || ForceCheck == check )
	    {
		if( HotSyncCheck != check || alarm <= TimGetSeconds() )
		{
#ifdef DEBUG
		    Char szDb[cchMaxDBName];
		    DmDatabaseInfo( pdb, szDb, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 );
		    DbgMsg_Alarms(( 10, "Full scan for '%s'...", szDb ));
#endif

		    alarm = CheckRecordsForNextAlarm( pdb, currentTime );

		    DbgMsg_Alarms(( 10, "updating next db alarm %lu", alarm ));
		    pAppInfo->alarmCache.set_m_seconds( alarm );

		}

		if( !*pnextAlarm || alarm < *pnextAlarm )
		    *pnextAlarm = alarm;
	    }

	    __MemPtrUnlock( pAppInfo );
	}
}

void
UpdateAlarm( DmOpenRef pdb, UInt32 alarmSeconds, bool fSet )
{
	Assert( pdb );

	// Get AppInfo pointer.

	UInt16 version;
	LocalID idAppInfo;

	DmDatabaseInfo( pdb, 0, 0, &version, 0, 0, 0, 0, &idAppInfo, 0, 0, 0 );

	Assert( version >= DbVersion_alarmcache );

	ThisAppInfoPtr const pAppInfo = ThisAppInfoPtr( MemLocalIDToLockedPtr( idAppInfo, CARDNO ) );

	Assert( pAppInfo );

	// This is the alarm date/time to set (or unset).

	UInt32 currentTime = pAppInfo ? pAppInfo->alarmCache.get_m_seconds() : 0;

	if( alarmSeconds && fSet )
	{
	    Assert( pAppInfo );

	    // If an earlier/equal alarm exists, or if the alarm is in the
	    // past, then no further work is needed.

	    if( ( currentTime && currentTime <= alarmSeconds ) ||
		( alarmSeconds < TimGetSeconds() ) )
	    {
		DbgMsg(( 5, "setting && earlier/equal alarm exists in this db" ));
		goto LOut;
	    }

	    // This becomes the new earliest alarm.
	}
	else
	{
	    // If an earlier alarm exists, this cannot have been the earliest,
	    // so no further work is needed.

	    if( currentTime && currentTime < alarmSeconds )
	    {
		DbgMsg_Alarms(( 5, "clearing, but db has earlier alarm" ));
		goto LOut;
	    }

	    // Look for other earliest alarm in this db.

	    alarmSeconds = 0;

	    if( pdb )
		CheckOneDatabaseForNextAlarm( pdb, TimGetSeconds(), &alarmSeconds, NormalCheck );
	}

	// Update the alarm cache in the database AppInfo block.  Luckily the
	// AppInfo block can be modified even if the database is open readonly.

	if( pAppInfo )
	{
	    DbgMsg(( 5, "earliest alarm for this db = %lu", alarmSeconds ));
	    pAppInfo->alarmCache.set_m_seconds( alarmSeconds );
	}

	//$ todo: (chrisant) NOT SUPPORTED YET.  This is where code to
	// interact with an Alarm Manager API would go, for registering the
	// next alarm to be able to fire.

LOut:
	__MemPtrUnlock( pAppInfo );
}

/*
 * DbOpenHs2 - Database wrapper for HS2 Palm databases.
 */

DbOpenHs2::DbOpenHs2( LPCTSTR pszFile )
{
	Assert( !m_fDirty );
	Assert( !m_fUserDirty );
	Assert( !m_fPrepared );
	Assert( !m_fMakeBackup );
	Assert( !m_fFlushing );
	Assert( !m_fCompleteReasons );
	Assert( !m_fFilteringDisabled );
	Assert( !m_nChangeCounter );

#ifdef OS_WINDOWS
	m_codepage = GetACP();
#endif

	if( pszFile )
	{
#ifdef OS_WINDOWSMOBILE
	    Assert( ::GetFileName( pszFile ) != pszFile );	// WM does not have the concept of a working directory, so paths should never be relative.
#endif
	    m_sFile.Set( pszFile );
	}
}

DbOpenHs2::~DbOpenHs2()
{
	Close();
}

SysError
DbOpenHs2::NewDatabase( LPCTSTR pszFile, LPCTSTR pszDisplayName, DbOpen** ppdb )
{
	Assert( pszFile );
	Assert( *pszFile );
	Assert( pszDisplayName );
	Assert( ppdb );

	StrA sName;
	SPNew<DbOpenHs2> spdb;

	*ppdb = 0;

	sName.SetT( pszDisplayName );
	if( sName.Length() >= dmDBNameLength )
	    sName.SetLength( dmDBNameLength - 1 );

	if( FileExists( pszFile ) )
	    return SE_FILEEXISTS;

	spdb = DmNewDatabase( pszFile, sName.Text() );
	if( !spdb )
	    return SE_UNKNOWN;

	spdb->PrepareForPPC();

	*ppdb = spdb.Transfer();

	return SE_OK;
}

SysError
DbOpenHs2::OpenDatabase( LPCTSTR pszFile, DbOpen** ppdb )
{
	Assert( pszFile );
	Assert( ppdb );

	*ppdb = 0;

	SysError e = SE_OK;
	SPNew<IProgress> spProgress;
	CreateProgressObject( &spProgress );
	SPNew<DbOpenHs2> spdb = DmOpenDatabase( pszFile, e, spProgress );

	if( e )
	    return e;

	Assert( spdb );

	// Prepare for use on PPC devices (e.g. resort the database using PPC
	// locale settings, but without marking the database dirty).

	spdb->PrepareForPPC();

	*ppdb = spdb.Transfer();

	return e;
}

void
DbOpenHs2::SetDirty()
{
	IncrementChangeCounter();
	m_fDirty = true;

	if( m_fPrepared )
	    m_fUserDirty = true;
}

void
DbOpenHs2::SetStoresDirty()
{
	if( !m_fStoresDirty )
	{
	    m_fStoresDirty = true;
	    SetDirty();
	}
}

void
DbOpenHs2::SetCategoriesDirty()
{
	if( !m_fCategoriesDirty )
	{
	    m_fCategoriesDirty = true;
	    SetDirty();
	}
}

void
DbOpenHs2::PrepareForPPC()
{
	/*
	 * Because the PPC platform sort order can be different than on the
	 * Palm platform, we always resort the database.  Note that even if we
	 * didn't always resort, we would still need to resort when the
	 * database has the dpfForceResort flag set (and it would be pointless
	 * to clear the dpfForceResort flag on the PPC since we always set it
	 * during save on the PPC).
	 *
	 * Because this is unconditional, we don't want this to dirty the
	 * database or open it for writing.  Otherwise the simple act of
	 * opening a database would always alter it.
	 *
	 * And now we take other preparatory actions here as well, such as
	 * converting from the Palm OS categories to our custom categories.
	 */

	// Prepare the store list.

	GetMasterStoreList();
	SortStores( false );

	Assert( m_pStores );

	// Prepare the category list.  If the database does not yet use our
	// custom category implementation, then convert the database at this
	// time.

	if( m_prefs.fCustomCategories )
	{
	    GetCategoryList();
	    m_pCategories->Sort();
	    SetCategoriesDirty();
	}
	else
	{
	    //$ todo: (chrisant) If this fails we're in trouble.
	    CategoriesConvertFromNative();
	}

	Assert( m_pCategories );

	// Sort the database.  This needs to happen after the categories
	// themselves have been sorted because the database might be sorted by
	// category.

	SortDatabaseInternal( false );

	// Repair duplicate store entries (that's all the old "Repair Items"
	// command did in HS2, even though people thought it did more).  Also
	// conditionally recalculate the bold stores.

	if( RepairDuplicateStoreEntries() ||
	    ( m_prefs.get_flags() & dpfResetCounts ) )
	{
	    m_prefs.and_flags( UInt16( ~dpfResetCounts ) );
	    ResetStoreItemCounts( false );
	}

	// The database is finished being prepared, and is ready for use.  Any
	// alterations made beyond this point mark the database as dirty and
	// needing to be saved (any alterations made prior to this point have
	// not marked the database as needing to saved).

	m_fPrepared = true;

	Assert( !IsDirty() );
}

bool
DbOpenHs2::CanFlush( bool *pfFlushSupported )
{
	if( pfFlushSupported )
	    *pfFlushSupported = true;

	// If a Flush is currently progress don't let another start
	// reentrantly.

	if( m_fFlushing )
	    return false;

	return IsDirty();
}

template<class T>
class AutoRevert
{
    public:
	AutoRevert( T& variable, T const& value ) : m_pVariable( &variable ), m_oldValue( variable) { *m_pVariable = value; }
	~AutoRevert() { *m_pVariable = m_oldValue; }
    private:
	T* const	m_pVariable;
	T const		m_oldValue;
};

#ifdef _WIN32_WCE
struct SaveSentinel
{
	SaveSentinel( TCHAR const* pszFile ) { SetStringPref( piSaveSentinel, pszFile ); }
	~SaveSentinel() { SetStringPref( piSaveSentinel, 0 ); }
};
#endif

SysError
DbOpenHs2::Flush()
{
	Assert( m_pdb );

	Assert( m_sFile.Length() > 4 );
	Assert( LStrCmpI( m_sFile.Text() + m_sFile.Length() - 4, TEXT( ".pdb" ) ) == 0 );
	Assert( FImplies( !m_fDirty, !m_fStoresDirty ) );
	Assert( FImplies( !m_fDirty, !m_fCategoriesDirty ) );

	// On Windows Mobile it is possible for certain window messages to
	// trigger the list view to make reentrant Flush() calls:  inside
	// Flush() we create a progress object, which peeks WM_PAINT out of
	// the message queue, which dispatches various system messages, which
	// triggers the list view to make a second Flush() call while we're
	// already inside the first one.  It would be error prone to require
	// callers to avoid reentrancy, so we're just going to bail out of
	// reentrant calls by returning success.

	if( !CanFlush( 0 ) )
	{
	    if( m_fFlushing )
		DbgMsg(( 10, "REENTRANT FLUSH" ));
	    return SE_OK;
	}

#ifdef _WIN32_WCE
	// Some utilities or OEM features aggressively kill applications
	// within a second or two when the user taps the [X] button.  This
	// kills HS in the middle of a save.  This sentinel at least enables
	// HS to detect that it happened, so it can alert the user.
	SaveSentinel ss( m_sFile.Text() );
#endif

	Assert( !m_fFlushing );
	Assert( IsDirty() );
	AutoRevert<bool> flushing( m_fFlushing, true );

	SysError e = SE_OK;
	bool fExisted = FileExists( m_sFile.Text() );

	// Make sure we can write to the file:  this should catch when the
	// file is readonly or has restricted permissions.

	if( fExisted )
	{
	    SPNew<FileIO> spFile = FileIO::Create();
	    spFile->Set( m_sFile.Text() );
	    if( !spFile->Open( FM_READWRITE, e ) )
	    {
		Assert1( false, "REASON:  Save failed because could not open file for write.  (SysError %u)", e );
		return e;
	    }
	}

	// Update the app info block.

	WriteAppInfo();

	// Update the store and category lists (these check internally whether
	// the lists are dirty).

	PutMasterStoreList();
	PutCategoryList();

	// Use a two-phase write to avoid data loss.

	StrT sOld;
	StrT sNew;
	LPCTSTR pszName = GetFileName( m_sFile.Text() );

	Assert( pszName );

	sOld.Set( m_sFile.Text(), pszName - m_sFile.Text() );
	sOld.Append( TEXT( "~" ) );
	sOld.Append( pszName );

	sNew.Set( sOld );
	sNew.Append( TEXT( ".tmp" ) );
	sOld.Append( TEXT( ".bak" ) );

	bool const fMakingBackup = m_fMakeBackup && fExisted;

	if( fMakingBackup )
	{
	    sOld.Set( m_sFile );
	    sOld.SetLength( sOld.Length() - 4 );
	    sOld.Append( TEXT( "-bak.bak" ) );
	}

	// Write the database to the new file.

	__DeleteFile( sNew.Text() );

	SPNew<IProgress> spProgress;
	CreateProgressObject( &spProgress );
	if( !m_pdb->Write( sNew.Text(), e, spProgress ) )
	{
	    Assert1( false, "REASON:  Save failed because could not write new file.  (SysError %u)", e );
	    return e;
	}

	// Rename the original file.

	if( fExisted )
	{
	    __DeleteFile( sOld.Text() );
	    e = __RenameFile( m_sFile.Text(), sOld.Text() );

	    if( e )
	    {
		Assert1( false, "REASON:  Save failed because could not rename original file.  (SysError %u)", e );
		return e;
	    }

	    m_fMakeBackup = false;
	}

	// Rename the new file into place.  If this fails, we try to rename
	// the original file back into place, but keep the newly saved file
	// (the .tmp file) just in case.

	e = __RenameFile( sNew.Text(), m_sFile.Text() );

	if( e )
	{
	    Assert1( false, "REASON:  Save failed because could not rename new file.  (SysError %u)", e );

	    if( fExisted )
	    {
		__RenameFile( sOld.Text(), m_sFile.Text() );
		m_fMakeBackup = fMakingBackup;
	    }

	    return e;
	}

	m_fDirty = false;
	m_fUserDirty = false;

	// Delete the renamed original file, unless it became a backup.

	if( !fMakingBackup )
	    __DeleteFile( sOld.Text() );

	return SE_OK;
}

void
DbOpenHs2::Close()
{
	Assert( !IsDirty() );

	ClearUndo();

	delete m_pStores;
	m_pStores = 0;

	delete m_pCategories;
	m_pCategories = 0;

	if( m_pdb )
	{
	    ( (Hs2Database*)m_pdb )->SetPopen( 0 );

	    m_pdb->Close();

	    delete m_pdb;
	    m_pdb = 0;
	}

	//$ todo: (chrisant) Zero out everything else as appropriate.

	m_sLookup.Clear();
}

LPCTSTR
DbOpenHs2::Name() const
{
	Assert( m_pdb );

	if( !m_sName.Length() )
	{
	    LPCSTR psz = m_pdb->GetDbName();
	    SkipPrefix( &psz );
	    m_sName.SetA( psz );
	}

	return m_sName.Text();
}

bool
DbOpenHs2::HasPrefix() const
{
	LPCSTR const pszDbName = m_pdb->GetDbName();
	LPCSTR psz = pszDbName;
	SkipPrefix( &psz );
	return psz != pszDbName;
}

LPCTSTR
DbOpenHs2::FileName() const
{
	Assert( m_pdb );

	if( !m_sFile.Length() )
	{
	    m_sFile.Set( Name() );
	    m_sFile.Append( TEXT( ".pdb" ) );
	}

	return m_sFile.Text();
}

unsigned
DbOpenHs2::Count() const
{
	Assert( m_pdb );

	return m_pdb->DmNumRecords() - m_cMetadataRecords;
}

void
DbOpenHs2::OmitPrefix()
{
	m_fOmitPrefix = true;
}

SysError
DbOpenHs2::SetName( LPCTSTR pszName )
{
	Assert( m_pdb );
	Assert( pszName );
	Assert( CchLen( pszName ) < dmDBNameLength );

	StrT tmp;
	StrA s;

	bool const fOmitPrefix = m_fOmitPrefix;
	m_fOmitPrefix = false;

	// Prepend the HS2_ name prefix for consistency and Palm
	// compatibility.

	if( !fOmitPrefix )
	    tmp.Append( TEXT( "HS2_" ) );
	tmp.Append( pszName );

	s.SetT( tmp );

	if( s.Length() >= dmDBNameLength )
	    s.SetLength( dmDBNameLength - 1 );

	// Open the file and write the name.

	DmSetDatabaseInfo( m_pdb, s.Text(), 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 );
	SetDirty();

	return SE_OK;
}

void
DbOpenHs2::RefreshMetadataCounter()
{
#ifdef DEBUG
	unsigned cSeenStores = 0;
	unsigned cSeenCategories = 0;
#endif

	m_cMetadataRecords = 0;

	while( m_cMetadataRecords < Count() &&
	       ( GetItemFlags( m_cMetadataRecords, 0 ) & flagMetadataRecord ) )
	{
#ifdef DEBUG
	    MemHandle const h = DmQueryRecord( m_pdb, m_cMetadataRecords );
	    UInt8 const* const p = (UInt8 const*)MemHandleLock( h );

	    if( *p == tagMasterStoreList )
	    {
		if( cSeenStores == 1 )
		    PanicSz( "found a second master store list metadata record!" );
		else if( cSeenCategories == 1 && !cSeenStores )
		    PanicSz( "master store list is not first?!" );
		cSeenStores++;
	    }
	    else if( *p == tagCategoryList )
	    {
		if( cSeenCategories == 1 )
		    PanicSz( "found a second category list metadata record!" );
		cSeenCategories++;
	    }
	    else
	    {
		Panic1( "unexpected metadata record tag %u", *p );
	    }

	    MemHandleUnlock( h );
#endif

	    m_cMetadataRecords++;
	}

	// We will have asserted above immediately on detecting a second
	// master store list or category list, but here we will assert with an
	// exact count of superfluous metadata records.

	Assert1( cSeenStores <= 1, "found %u master store list records", cSeenStores );
	Assert1( cSeenCategories <= 1, "found %u category list records", cSeenCategories );
}

/*
 * Filtering.
 */

bool
DbOpenHs2::SetView( WORD view )
{
	CASSERT( viewAll == VIEW_ALL );
	CASSERT( viewNeed == VIEW_NEED );
	CASSERT( viewNotNeeded == VIEW_NOTNEEDED );
	CASSERT( viewAllNotPurchased == VIEW_ALLNOTPURCHASED );

	bool const fWasAltSort = FAltSort();

	if( view >= viewMAX )
	    view = VIEW_ALL;

	if( m_prefs.get_view() != view )
	{
	    m_prefs.set_view( view );
	    m_pbits->Clear();

	    SetDirty();
	}

	bool fHideChecked = !!m_prefs.fHideCompleted;

	if( VIEW_NEED == view )
	    fHideChecked = true;
	else if( VIEW_ALLNOTPURCHASED == view )
	    fHideChecked = false;

	if( !!m_prefs.fHideCompleted != fHideChecked )
	{
	    m_prefs.fHideCompleted = fHideChecked;
	    m_pbits->Clear();

	    SetDirty();
	}

	return fWasAltSort != FAltSort();
}

void
DbOpenHs2::SetCategory( WORD _category )
{
	//$ todo: (chrisant) Validate.

	UInt16 const dmCategory = ToDmCategory( _category );

	if( m_prefs.get_category() != dmCategory ||
	    m_prefs.m_fMultipleCategories )
	{
	    m_multiCategories.SetAll( false );

	    m_prefs.set_category( dmCategory );
	    m_prefs.m_fMultipleCategories = false;
	    m_pbits->Clear();

	    SetDirty();
	}
}

void
DbOpenHs2::SetStore( BYTE idStore )
{
	if( m_prefs.idStore != idStore ||
	    m_prefs.m_fMultipleStores )
	{
	    m_multiStores.SetAll( false );

	    m_prefs.idStore = idStore;
	    m_prefs.m_fMultipleStores = false;
	    m_pbits->Clear();

	    SetDirty();
	}
}

void
DbOpenHs2::SetHideChecked( bool fHideChecked )
{
	if( !!m_prefs.fHideCompleted != fHideChecked )
	{
	    m_prefs.fHideCompleted = fHideChecked;
	    m_pbits->Clear();

	    SetDirty();
	}

	if( FViewNeed( GetView() ) )
	{
	    m_prefs.set_view( UInt16( fHideChecked ? VIEW_NEED : VIEW_ALLNOTPURCHASED ) );

	    SetDirty();
	}
}

void
DbOpenHs2::SetCouponsOnly( bool fCouponsOnly )
{
	if( !!m_prefs.fCouponsOnly != fCouponsOnly )
	{
	    m_prefs.fCouponsOnly = fCouponsOnly;
	    m_pbits->Clear();

	    SetDirty();
	}
}

WORD
DbOpenHs2::GetView() const
{
	return m_prefs.get_view();
}

WORD
DbOpenHs2::GetCategory() const
{
	if( m_prefs.m_fMultipleCategories )
	    return dmAllCategories;

	return FromDmCategory( m_prefs.get_category() );
}

BYTE
DbOpenHs2::GetStore() const
{
	if( m_prefs.m_fMultipleStores )
	    return allStores;

	return m_prefs.idStore;
}

bool
DbOpenHs2::GetHideChecked() const
{
	return !!m_prefs.fHideCompleted;
}

bool
DbOpenHs2::GetCouponsOnly() const
{
	return !!m_prefs.fCouponsOnly;
}

void
DbOpenHs2::SetMultipleCategories( MultipleCategories const* pmulti )
{
	Assert( pmulti );

	unsigned const cCategoriesPlusUnfiled = GetCategoriesCount() + 1;
	unsigned const cMarked = pmulti->CountMarked( this );

	if( !cMarked )
	{
	    SetCategory( catUnfiledCategory );
	}
	else if( cMarked == cCategoriesPlusUnfiled )
	{
	    SetCategory( catAllCategories );
	}
	else
	{
	    if( !m_prefs.m_fMultipleCategories ||
		memcmp( pmulti, &m_multiCategories, sizeof( m_multiCategories ) ) != 0 )
	    {
		m_pbits->Clear();

		m_multiCategories = *pmulti;
		m_prefs.SetMultipleCategories( pmulti );

		SetDirty();
	    }
	}
}

bool
DbOpenHs2::GetMultipleCategories( MultipleCategories* pmulti ) const
{
	if( pmulti )
	{
	    if( m_prefs.m_fMultipleCategories )
		*pmulti = m_multiCategories;
	    else
		pmulti->SetAll( false );
	}

	return !!m_prefs.m_fMultipleCategories;
}
 
void
DbOpenHs2::SetMultipleStores( MultipleStores const* pmulti )
{
	Assert( pmulti );

	unsigned const cStoresPlusUnfiled = GetStoresCount() + 1;
	unsigned const cMarked = pmulti->CountMarked( this );

	if( !cMarked )
	{
	    SetStore( noStore );
	}
	else if( cMarked == cStoresPlusUnfiled )
	{
	    SetStore( allStores );
	}
	else
	{
	    if( !m_prefs.m_fMultipleStores ||
		memcmp( pmulti, &m_multiStores, sizeof( m_multiStores ) ) != 0 )
	    {
		m_pbits->Clear();

		m_multiStores = *pmulti;
		m_prefs.SetMultipleStores( pmulti );

		SetDirty();
	    }
	}
}

bool
DbOpenHs2::GetMultipleStores( MultipleStores* pmulti ) const
{
	if( pmulti )
	{
	    if( m_prefs.m_fMultipleStores )
		*pmulti = m_multiStores;
	    else
		pmulti->SetAll( false );
	}

	return !!m_prefs.m_fMultipleStores;
}

inline bool
DbOpenHs2::IsCategoryFiltered( WORD const category ) const
{
	Assert( category < GetCategoriesMaximum() ||
		category == catUnfiledCategory ||
		category == catAllCategories );

	if( m_fFilteringDisabled )
	    return false;

	if( m_prefs.m_fMultipleCategories )
	{
	    if( category == catUnfiledCategory )
		return !m_multiCategories.m_Unfiled;

	    if( category > GetCategoriesMaximum() )
		return true;

	    return !m_multiCategories.m_rg[category];
	}

	if( m_prefs.get_category() == dmAllCategories )
	    return false;

	return ( m_prefs.get_category() != ToDmCategory( category ) );
}

#if 0
#define DbgFiltered(x) DbgMsg(x)
#else
#define DbgFiltered(x) syntaxsafe
#endif

CASSERT( VIEW_ALL == 0 );
CASSERT( VIEW_NEED == 1 );
CASSERT( VIEW_NOTNEEDED == 2 );
CASSERT( VIEW_ALLNOTPURCHASED == 3 );
static bool const c_rgrgViewFilter[][VIEW_MAX] =
{
	{ true, false, true, false },	// ITEMSTATE_NONE
	{ true, true, false, true },	// ITEMSTATE_NEED
	{ true, false, true, true },	// ITEMSTATE_CHECKED
	{ true, false, false, true },	// ITEMSTATE_DEFERRED
};
CASSERT( DimensionOf( c_rgrgViewFilter ) == ITEMSTATE_MAX );

FilterReason
DbOpenHs2::IsRecordFiltered( unsigned record ) const
{
	MemHandle h = 0;
	FilterReason reason = notFiltered;
	UInt8* pRec;
	UInt8* pRecBegin;
	UInt16 flags;
	UInt16 attr;
	unsigned dmRecord = record;

	if( dmRecord != noRecordSelected )
	    dmRecord += m_cMetadataRecords;

	m_pdb->DmRecordInfo( dmRecord, &attr, 0, 0 );

	if( attr & dmRecAttrDelete )
	{
DbgFiltered(( 5, "filtered %u -- deleted", record ));
	    reason = reason | filteredDeleted;
	    if( !m_fCompleteReasons )
		goto LOut;
	}

	Assert( m_prefs.fCustomCategories );

	h = m_pdb->DmQueryRecord( dmRecord );
	pRec = ( UInt8* )( MemHandleLock( h ) );
	pRecBegin = pRec;

#if 0
	if( tagMasterStoreList == *pRec )
	{
	    reason = filteredCorrupt;
	    goto LOut;
	}
#endif

	if( tagFlags != *pRec )
	{
	    Panic2( "*pRec == %d, but should be tagFlags(%d)",
		    UInt16( UInt8( *pRec ) ), tagFlags );
	    reason = filteredCorrupt;
	    goto LOut;
	}

	++pRec;
	flags = ( UInt16( UInt8( pRec[0] ) ) << 8 ) | pRec[1];

	if( ( flags & flagPrivate ) && privHideRecords == GetPrivateRecordVisualMode() )
	{
DbgFiltered(( 5, "filtered %u -- private", record ));
	    reason = reason | filteredPrivate;
	    if( !m_fCompleteReasons )
		goto LOut;
	}

	if( m_prefs.fCouponsOnly && !( flags & flagCoupon ) )
	{
DbgFiltered(( 5, "filtered %u -- coupons only", record ));
	    reason = reason | filteredCoupon;
	    if( !m_fCompleteReasons )
		goto LOut;
	}

	if( ( m_prefs.idStore != allStores || m_prefs.m_fMultipleStores ) && !m_fFilteringDisabled )
	{
	    UInt8 cStores = 0;

	    // This is very tricky code, and was changed many times to get it
	    // finally working right.  Hopefully it is relatively well
	    // optimized.  Be very cautious changing the logic here!

	    if( !( flags & flagHasStore ) )
	    {
		if( m_prefs.m_fMultipleStores )
		{
		    if( m_multiStores.m_Unfiled )
			goto LStoreMatch;
		}
		else
		{
		    if( noStore == m_prefs.idStore )
			goto LStoreMatch;
		}
	    }
	    else
	    {
		Assert( tagStores == *( pRec + 2 ) );

		// Skip the tagStores as well.

		pRec += 3;

		// Notice decrement only happens if store doesn't match!

		if( m_prefs.m_fMultipleStores )
		{
		    for( cStores = *( pRec++ ); cStores; cStores--, pRec++ )
			if( *pRec < DimensionOf( m_multiStores.m_rg ) && m_multiStores.m_rg[*pRec] )
			    break;
		}
		else
		{
		    for( cStores = *( pRec++ ); cStores; cStores-- )
			if( m_prefs.idStore == *( pRec++ ) )
			    break;
		}
	    }

	    if( !cStores )
	    {
		// Didn't match one of the stores.

DbgFiltered(( 5, "filtered %u -- store", record ));
		reason = reason | filteredStore;
		if( !m_fCompleteReasons )
		    goto LOut;
	    }

LStoreMatch:
	    ;
	}

	//$ todo: (chrisant) This could be optimized a little because we know
	// where the tagCategory will be if it exists in the record...
	if( ( m_prefs.get_category() != dmAllCategories || m_prefs.m_fMultipleCategories ) && m_prefs.fCustomCategories )
	{
	    UInt16 recordCategory = catUnfiledCategory;
	    UInt8 const* const p = reinterpret_cast<UInt8 const*>( GetMemberPtr( pRecBegin, tagCategory ) );
	    if( p )
		recordCategory = *p;
	    if( IsCategoryFiltered( recordCategory ) )
	    {
DbgFiltered(( 5, "filtered %u -- category", record ));
		reason = reason | filteredCategory;
		goto LOut;
	    }
	}

	// Show items appropriate to the current view.  But always show All
	// items when a lookup string is active, to make it easy to find items
	// that already exist without creating duplicates.

	if( m_sLookup.Length() )
	{
	    // EFIGS Palm devices essentially use 1252, so use this higher
	    // performance lookup algorithm when possible.

#ifdef OS_WINDOWS
	    if( 1252 == m_codepage )
	    {
		UInt8* pRecBegin = ( UInt8* )( MemHandleLock( h ) );
		MemHandleUnlock( h );	// drop the lock count back to 1

		int n = -1;
		MemPtr p;

		p = GetMemberPtr( pRecBegin, tagDescription );
		if( p )
		    n = StrLookup( &m_sLookupA, reinterpret_cast<char const*>( p ) );

		if( n < 0 )
		{
		    p = GetMemberPtr( pRecBegin, tagNote );
		    if( p )
			n = StrLookup( &m_sLookupA, reinterpret_cast<char const*>( p ) );

		    if( n < 0 )
		    {
			p = GetMemberPtr( pRecBegin, tagCustomText );
			if( p )
			    n = StrLookup( &m_sLookupA, reinterpret_cast<char const*>( p ) );
		    }
		}

		if( n < 0 )
		{
DbgFiltered(( 5, "filtered %u -- lookup", record ));
		    reason = reason | filteredLookup;
		    if( !m_fCompleteReasons )
			goto LOut;
		}
	    }
	    else
#endif
	    {
		SPNew<Item> spitem = UnPackRecord( record );

		if( !spitem )
		{
DbgFiltered(( 5, "filtered %u -- error unpacking for lookup", record ));
		    reason = filteredCorrupt;	//$ review: (chrisant) Inaccurate filter reason...
		    goto LOut;
		}

		if( StrLookup( &m_sLookup, spitem->GetDescription() ) < 0 &&
		    StrLookup( &m_sLookup, spitem->GetNote() ) < 0 &&
		    StrLookup( &m_sLookup, spitem->GetCustomText() ) < 0 )
		{
DbgFiltered(( 5, "filtered %u -- lookup", record ));
		    reason = reason | filteredLookup;
		    if( !m_fCompleteReasons )
			goto LOut;
		}
	    }
	}
	else if( !m_fFilteringDisabled )
	{
	    // Test flags based on view.

	    if( !( c_rgrgViewFilter[ItemStateFromFlags( flags )][m_prefs.get_view()] ) )
	    {
		DbgFiltered(( 5, "filtered %u -- view", record ));
		reason = reason | filteredState;
		if( !m_fCompleteReasons )
		    goto LOut;
	    }
	}

LOut:
	__MemHandleUnlock( h );
	return reason;
}

/*
 * Columns.
 */

#ifdef DEBUG
#define HS2COLDEFENTRY( col, icx, bit ) { col, icx, bit },
#else
#define HS2COLDEFENTRY( col, icx, bit ) { icx, bit },
#endif

struct Hs2ColDefEntry
{
#ifdef DEBUG
	ColumnIdentifier	m_column;
#endif
	short			m_icxIndex;
	WORD			m_bitColumn;
};

static Hs2ColDefEntry const c_rgHs2ColDefs[] =
{
	HS2COLDEFENTRY( COLID_NEED, -1, colNeed )
	HS2COLDEFENTRY( COLID_PRIO, -1, colPrio )
	HS2COLDEFENTRY( COLID_ICON, -1, colIcon )
	HS2COLDEFENTRY( COLID_DESCRIPTION, -1, colName )
	HS2COLDEFENTRY( COLID_CATEGORY, icolcxCategory, colCategory )
	HS2COLDEFENTRY( COLID_CUSTOMTEXT, icolcxCustom, colCustomText )
	HS2COLDEFENTRY( COLID_QUANTITY, icolcxQuantity, colQuantity )
	HS2COLDEFENTRY( COLID_UNITS, icolcxUnits, colUnits )
	HS2COLDEFENTRY( COLID_AISLE, icolcxAisle, colAisle )
	HS2COLDEFENTRY( COLID_UNITPRICE, icolcxUnitPrice, colUnitPrice )
	HS2COLDEFENTRY( COLID_PRICE, icolcxPrice, colPrice )
	HS2COLDEFENTRY( COLID_COUPON, -1, colCoupon )
	HS2COLDEFENTRY( COLID_TAXABLE, -1, colTaxable )
	HS2COLDEFENTRY( COLID_TAXABLE2, -1, colTaxable2 )
	HS2COLDEFENTRY( COLID_DATE, icolcxDate, colDate )
};

CASSERT( DimensionOf( c_rgHs2ColDefs ) == COLID_MAX );

int
DbOpenHs2::GetColumnWidth( ColumnIdentifier column ) const
{
	Assert( m_pdb );

#ifdef DEBUG
	static bool s_fVerified = false;

	if( !s_fVerified )
	{
	    for( int ii = 0; ii < DimensionOf( c_rgHs2ColDefs ); ii++ )
		Assert2( c_rgHs2ColDefs[ii].m_column == ColumnIdentifier( ii ),
			 "Entry %d is for column id %d", ii, c_rgHs2ColDefs[ii].m_column );
	    s_fVerified = true;
	}

	Assert1( column >= 0 && column < COLID_MAX, "Unexpected column id: %d", column );
#endif

	// Bail early if the column is not enabled.

	if( !( m_prefs.get_columns() & c_rgHs2ColDefs[column].m_bitColumn ) )
	    return 0;

	// If the column is not resizable, use the default width.

	short index = c_rgHs2ColDefs[column].m_icxIndex;

	if( index < 0 )
	    return COLWIDTH_DEFAULT;

	// Some legacy versions didn't support drag resizing, so they allowed
	// choosing between Default, Narrow, Medium, and Wide.  On the PPC we
	// map all of these to the default width.

	int cx = m_prefs.rgColWidth[index];

	if( COLID_DATE == column )
	    return cx ? COLWIDTH_FULL : COLWIDTH_DEFAULT;

	if( cx < 5 )
	    return COLWIDTH_DEFAULT;

	return cx;
}

void
DbOpenHs2::SetColumnWidth( ColumnIdentifier column, int cx )
{
	Assert( m_pdb );

	Assert( column >= 0 && column < DimensionOf( c_rgHs2ColDefs ) );
	Assert( cx >= 0 );

	if( cx < 0 )
	    return;
	if( cx > 250 )
	    cx = 250;

	// Update the bit mask of enabled columns.

	Hs2ColDefEntry const* pCol = &c_rgHs2ColDefs[column];
	WORD columns = m_prefs.get_columns();

	if( cx )
	    columns |= pCol->m_bitColumn;
	else
	    columns &= ~pCol->m_bitColumn;

	if( columns != m_prefs.get_columns() )
	{
	    m_prefs.set_columns( columns );

	    SetDirty();
	}

	// For resizable columns, store the scaled column width.

	if( pCol->m_icxIndex >= 0 )
	{
	    UInt8 cxAdjusted = UInt8( cx );

	    if( cxAdjusted < 5 )
	    {
		if( COLID_DATE == column )
		    cxAdjusted = !!( COLWIDTH_FULL == cxAdjusted );
		else
		    cxAdjusted = 0;
	    }

	    if( cxAdjusted != m_prefs.rgColWidth[pCol->m_icxIndex] )
	    {
		m_prefs.rgColWidth[pCol->m_icxIndex] = cxAdjusted;

		SetDirty();
	    }
	}
}

/*
 * Read/write records.
 */

SysError
DbOpenHs2::DeleteRecord( unsigned record, bool fCanUndo )
{
	MemHandle h;
	UInt8* pRec;
	UInt8 c;
	UInt16 flags;
	UInt32 uid;
	SPNew<Item> spitemUndo;

	Assert( record != noRecordSelected );

	if( record != noRecordSelected )
	{
	    unsigned const dmRecord = record + m_cMetadataRecords;

#ifdef DEBUG
	    h = DmQueryRecord( m_pdb, dmRecord );
	    pRec = ( UInt8* )MemHandleLock( h );
	    AssertSz( tagMasterStoreList != *pRec, "bad remove on Stores list" );
	    AssertSz( tagCategoryList != *pRec, "bad remove on Categories list" );
	    MemHandleUnlock( h );
#endif

	    UInt32 const alarm = GetRecordAlarm( m_pdb, record, 0 );

	    if( fCanUndo )
	    {
		spitemUndo = UnPackRecord( record );//, maskALL, 0 );

		if( !spitemUndo )
		{
		    // The undo stack is very sensitive to missing actions, so
		    // we must clear the undo stack if we couldn't acquire the
		    // undo state for the item.

		    ClearUndo();
		}
	    }

	    h = DmQueryRecord( m_pdb, dmRecord );
	    pRec = ( UInt8* )MemHandleLock( h );

	    Assert( tagFlags == *pRec );

	    pRec++;

	    CASSERT( SameSize( flags, spitemUndo->m_flags ) );

	    flags = ( pRec[0] << 8 ) | pRec[1];

	    if( ( flags & ( flagNeed|flagPurchased ) ) == flagNeed )
	    {
		pRec += 2;
		c = 0;

		if( tagStores == *pRec )
		{
		    pRec++;
		    c = *pRec;
		    pRec++;
		}
		else
		{
		    pRec = 0;
		}

		StoreList stores;
		stores.Borrow( c, reinterpret_cast<Store*>( pRec ) );
		UpdateCounts( &stores, -1 );

		SetDirty();
	    }

	    MemHandleUnlock( h );

	    if( fCanUndo )
		SetupUndo( spitemUndo.UnsafeAddress(), record, true, false );

	    DmRecordInfo( m_pdb, dmRecord, 0, &uid, 0 );

#if 0
	    //$ review: (chrisant) DmDeleteRecord or DmRemoveRecord?  Undo
	    // doesn't care which; does anything else?
	    DmDeleteRecord( m_pdb, dmRecord );
	    // Move the deleted record to the end, so DmFindSortPosition is
	    // happy.
	    DmMoveRecord( m_pdb, dmRecord, DmNumRecords( m_pdb ) );
#else
	    DmRemoveRecord( m_pdb, dmRecord );
	    m_pbits->RemoveRecord( record );
#endif

	    SetDirty();

	    m_recAfterPack = noRecordSelected;

	    WriteAppInfo();	//$ review: (chrisant) Seems superfluous.

	    if( alarm )
	    {
		UpdateAlarm( m_pdb, alarm, false );
#if 0
		DeleteAlarmIfPosted( m_pdb, uid );
#endif
	    }
	}

	return SE_OK;
}

int
DbOpenHs2::GetChangeCounter() const
{
	return m_nChangeCounter;
}

unsigned
DbOpenHs2::GetRecordsMaximum() const
{
	return 65534 - m_cMetadataRecords;
}

SysError
DbOpenHs2::GetRawRecordBytes( unsigned record, BYTE** ppbOut, unsigned* pcb ) const
{
	Assert( record != noRecordSelected );
	Assert( pcb );

	unsigned const dmRecord = record + m_cMetadataRecords;
	MemHandle const hRec = DmQueryRecord( m_pdb, dmRecord );

	if( !hRec )
	    return SE_UNKNOWN;

	*pcb = MemHandleSize( hRec );

	if( ppbOut )
	{
	    *ppbOut = static_cast<BYTE*>( MemAlloc( *pcb ) );

	    if( !*ppbOut )
		return SE_OUTOFMEMORY;

	    memmove( *ppbOut, MemHandleLock( hRec ), *pcb );
	    MemHandleUnlock( hRec );
	}

	return SE_OK;
}

/*
 * Sorting.
 */

bool
DbOpenHs2::SetSortOneOrder( bool fSortOneOrder )
{
	UInt16 flags = m_prefs.get_flags();

	flags &= ~dpfSeparateSorts;

	if( !fSortOneOrder )
	    flags |= dpfSeparateSorts;

	if( m_prefs.get_flags() != flags )
	{
	    m_prefs.set_flags( flags );
	    m_pbits->Clear();

	    SetDirty();

	    return true;
	}

	return false;
}

bool
DbOpenHs2::SetSortOrderInfo( int index, const DbSortOrder* pso )
{
	Assert( 0 == ( index & ~1 ) );
	Assert( pso );

	SortOrderInfo soi;
	CASSERT( sizeof( soi ) == sizeof( m_prefs.rgOrder[index] ) );

	memcpy( &soi, &m_prefs.rgOrder[index], sizeof( soi ) );

	for( Int16 ii = DimensionOf( soi.m_rgfDescend ); ii--; )
	{
	    soi.m_rgfDescend[ii] = !!pso->m_rgfDescend[ii];
	    soi.set_m_sort( ii, pso->m_rgiSort[ii] );
	}

	soi.m_fCompletedAtBottom = !!pso->m_fCheckedLast;

	if( memcmp( &soi, &m_prefs.rgOrder[index], sizeof( soi ) ) != 0 )
	{
	    //$ todo: (chrisant) Stronger assertions/validations.

	    memcpy( &m_prefs.rgOrder[index], &soi, sizeof( soi ) );
	    m_pbits->Clear();

	    SetDirty();

	    return true;
	}

	return false;
}

bool
DbOpenHs2::IsSortedByColumn( ColumnIdentifier column ) const
{
	struct SortColumnInfo
	{
	    ColumnIdentifier	m_column;
	    SortType		m_sort;
	};

	static SortColumnInfo const c_rgSortColumnInfo[] =
	{
	    { COLID_PRIO,		SORT_PRIORITY },
	    { COLID_DATE,		SORT_DATE },
	    { COLID_AISLE,		SORT_AISLE },
	    { COLID_CATEGORY,		SORT_CATEGORY },
	    { COLID_DESCRIPTION,	SORT_DESCRIPTION },
	    { COLID_QUANTITY,		SORT_QUANTITY },
	    { COLID_UNITS,		SORT_UNITS },
	    { COLID_PRICE,		SORT_PRICE },
	    { COLID_CUSTOMTEXT,		SORT_CUSTOMTEXT },
	    { COLID_ICON,		SORT_ICON },
	};
	CASSERT( DimensionOf( c_rgSortColumnInfo ) == COLID_MAX - 5 /* less 5 for Need, UnitPrice, Coupon, Taxable, Taxable2 */ );
	CASSERT( DimensionOf( c_rgSortColumnInfo ) == SORT_MAX - 3 /* less 3 for None, EntryOrder, Alarm */ );

	if( CanLearnOrder( FAltSort() ) )
	    return false;

	unsigned ii;
	SortType sort = SORT_NONE;

	for( ii = DimensionOf( c_rgSortColumnInfo ); ii--; )
	    if( c_rgSortColumnInfo[ii].m_column == column )
	    {
		sort = c_rgSortColumnInfo[ii].m_sort;
		break;
	    }

	if( !sort )
	    return false;

	DbSortOrder const* const p = GetSortOrderInfo( FAltSort() );

	for( ii = DimensionOf( p->m_rgiSort ); ii--; )
	    if( p->m_rgiSort[ii] == sort )
		return true;

	return false;
}

bool
DbOpenHs2::GetSortOneOrder() const
{
	return !( m_prefs.get_flags() & dpfSeparateSorts );
}

DbSortOrder const*
DbOpenHs2::GetSortOrderInfo( int index ) const
{
	Assert( !( index & ~1 ) );

	static DbSortOrder s_rgOrder[NUM_SORT_ORDERS];
	DbSortOrder* pso = &s_rgOrder[index];

	ZeroMemory( pso, sizeof( *pso ) );

	for( Int16 ii = DimensionOf( pso->m_rgfDescend ); ii--; )
	{
	    pso->m_rgfDescend[ii] = !!m_prefs.rgOrder[index].m_rgfDescend[ii];
	    pso->m_rgiSort[ii] = m_prefs.rgOrder[index].get_m_sort( ii );
	}

	pso->m_fCheckedLast = !!m_prefs.rgOrder[index].m_fCompletedAtBottom;

	return pso;
}

SysError
DbOpenHs2::SortRecord( unsigned const recordFrom, unsigned* const precordTo )
{
	Assert( m_pdb );
	Assert( m_pbits );
	Assert( recordFrom != noRecordSelected );

	SysError e = SE_OK;
	SortContext sc( this );
	unsigned dmRecord = recordFrom + m_cMetadataRecords;
	unsigned recordTo = recordFrom;

#ifdef DEBUG
	AssertNoRecordLocks( TEXT( "before sort record" ) );
#endif

	if( GetUnpackSortMask() )
	{
	    UInt16		attr;
	    UInt32		uid;
	    MemHandle		h;

	    DmRecordInfo( m_pdb, dmRecord, &attr, &uid, 0 );

	    if( !FAlreadySorted( dmRecord ) )
	    {
		if( DmDetachRecord( m_pdb, dmRecord, &h ) )
		    goto LError;

		dmRecord = DmFindSortPosition( m_pdb, MemHandleLock( h ), CmpFindSort, 0 );
		MemHandleUnlock( h );

		Assert( dmRecord >= m_cMetadataRecords );
		recordTo = dmRecord - m_cMetadataRecords;

		DbgMsg(( 5, TEXT( "sorting from %u to %u" ), recordFrom, recordTo ));

		if( m_pbits )
		    m_pbits->MoveRecord( recordFrom, recordTo );

		if( DmAttachRecord( m_pdb, &dmRecord, h, 0 ) )
		    goto LError;

		Assert( dmRecord >= m_cMetadataRecords );
		recordTo = dmRecord - m_cMetadataRecords;

		if( DmSetRecordInfo( m_pdb, dmRecord, &attr, &uid ) )
		    goto LError;

		SetDirty();
	    }
	}

LOut:
	//m_itemCache.UpdateRecord( recordFrom, record );

#ifdef DEBUG
	AssertNoRecordLocks( TEXT( "after sort record" ) );
#endif

	if( precordTo )
	    *precordTo = recordTo;

	return e;

LError:
	PanicSz( "error while trying to sort record" );
	// Might as well have a reasonable exit path in case I hit "Continue"
	// in the debugger...
	recordTo = noRecordSelected;
	e = SE_UNKNOWN;
	goto LOut;
}

SysError
DbOpenHs2::SortDatabaseInternal( bool fDirty )
{
	Assert( m_pdb );

	Err err = 0;
	bool fSorted = false;
	SortContext sc(this);

	if( GetUnpackSortMask() )
	{
#if defined( DEBUG ) && defined( _WIN32_WCE )
	    AssertNoRecordLocks( TEXT( "before sort database" ) );
	    DWORD const tickBegin = GetTickCount();
#endif

	    ScopedProgressHolder sph;
	    if( Count() > 2000 )
		sph.Create();
	    sph.Init( 0, PROGRESS_NOWAIT|PROGRESS_NOMETER|PROGRESS_FORCEVISIBLE, CAPTION_SORTING );

	    CE( DmQuickSort( m_pdb, CmpFindSort, 0 ) );

	    sph.End();

	    RefreshMetadataCounter();

	    m_pbits->Clear();

	    if( fDirty )
		SetDirty();

	    fSorted = true;	// true == not cancelled, false == cancelled.

#if defined( DEBUG ) && defined( _WIN32_WCE )
	    DWORD const tickEnd = GetTickCount();
	    AssertNoRecordLocks( TEXT( "after sort database" ) );
	    DbgMsg(( 20, TEXT( "Sorted %u records in %u milliseconds." ), Count(), tickEnd - tickBegin ));
#endif
	}

LError:
	return err ? SE_UNKNOWN : SE_OK;
}

SysError
DbOpenHs2::SortDatabase()
{
	return SortDatabaseInternal( true );
}

bool
DbOpenHs2::SetCanLearnOrder( int const index, bool const f )
{
	Assert( !( index & ~1 ) );

	if( !m_prefs.m_rgfAutoOrder[index] != !f )
	{
	    m_prefs.m_rgfAutoOrder[index] = f;
	    SetDirty();
	    return true;
	}

	return false;
}

bool
DbOpenHs2::CanLearnOrder( int const index ) const
{
	Assert( !( index & ~1 ) );

	return !!m_prefs.m_rgfAutoOrder[index];
}

bool
DbOpenHs2::IsLearningOrder() const
{
	return CanLearnOrder( FAltSort() ) && m_prefs.m_fLearning;
}

bool
DbOpenHs2::ActivateLearningOrder( bool const fActivate )
{
	if( !ActivateLearningOrderHelper( fActivate ) )
	    return false;

	m_prefs.m_fLearning = fActivate;
	SetDirty();

	return true;
}

/*
 * Undo.
 */

void
DbOpenHs2::SetupUndo( Item** ppitemUndo, unsigned record, bool fDeleted, bool fNew )
{
	unsigned recordModified = noRecordSelected;
	unsigned recordDeleted = noRecordSelected;
	UInt16 c;

	Assert( ppitemUndo );
	Assert( noRecordSelected != record );
	Assert( FImplies( !*ppitemUndo, fNew ) );
	Assert( FImplies( *ppitemUndo, !fNew ) );

	if( *ppitemUndo || fNew )
	{
	    c = m_cUndo;

	    if( c >= DimensionOf( m_rgUndo ) )
	    {
		if( m_rgUndo[0].m_pitem )
		{
		    delete m_rgUndo[0].m_pitem;
		    m_rgUndo[0].m_pitem = 0;
		}
		c--;
		MemMove( &m_rgUndo[0], &m_rgUndo[1], c * sizeof( m_rgUndo[0] ) );
	    }

	    if( fDeleted )
		recordDeleted = record;
	    else
		recordModified = record;

	    if( *ppitemUndo )
		( *ppitemUndo )->m_storesORIG.Free();

	    m_rgUndo[c].m_pitem = *ppitemUndo;
	    m_rgUndo[c].m_recordModified = recordModified;
	    m_rgUndo[c].m_recordDeleted = recordDeleted;
	    DmRecordInfo( m_pdb, record + m_cMetadataRecords, 0, &m_rgUndo[c].m_uid, 0 );

	    m_cUndo = c + 1;

	    *ppitemUndo = 0;
	}
}

bool
DbOpenHs2::CanUndo()
{
	return !!m_cUndo;
}

SysError
DbOpenHs2::Undo( unsigned* precord, bool* pfDidUndo )
{
	SysError	e = SE_OK;
	UInt32		uid;
	unsigned	record;
	SPNew<Item>	spitemUndo = 0;
	UndoInfo*	pundo;

	if( precord )
	    *precord = 0;
	if( pfDidUndo )
	    *pfDidUndo = false;

	if( m_cUndo )
	{
	    m_cUndo--;

	    pundo = &m_rgUndo[m_cUndo];
	    spitemUndo = pundo->m_pitem;
	    pundo->m_pitem = 0;

	    if( spitemUndo )
		spitemUndo->m_flags &= ~flagWasNeeded;

	    record = pundo->m_recordModified;

	    if( noRecordSelected != record )
	    {
		unsigned dmRecord = record + m_cMetadataRecords;

		// See if the record index we cached is still correct.

		if( dmRecord < DmNumRecords( m_pdb ) )
		{
		    DmRecordInfo( m_pdb, dmRecord, 0, &uid, 0 );

		    if( uid != pundo->m_uid )
		    {
			record = noRecordSelected;
			dmRecord = noRecordSelected;
		    }
		}
		else
		{
		    record = noRecordSelected;
		    dmRecord = noRecordSelected;
		}

		// Look up the cached lid if we don't know the correct record
		// index.

		if( noRecordSelected == record )
		{
		    if( DmFindRecordByID( m_pdb, pundo->m_uid, &dmRecord ) )
		    {
			e = SE_UNKNOWN;
			goto LError;
		    }

		    Assert( dmRecord >= m_cMetadataRecords );

		    record = dmRecord - m_cMetadataRecords;
		}
	    }

	    // Get current set of stores, so we can update bold counts
	    // correctly.

	    if( noRecordSelected != record && spitemUndo )
	    {
		SPNew<Item> spitem = UnPackRecord( record );//, maskStores|maskFlags, 0 );
		if( !spitem )
		{
		    e = SE_UNKNOWN;
		    goto LError;
		}

		// Save off the item's current store list as the undo's
		// original store list.  This ensures that writing the undo
		// back into the database will undo the current bold counts.

		Assert( !spitemUndo->m_storesORIG.Count() );

		spitemUndo->m_storesORIG.Swap( spitem->m_stores );
		spitemUndo->m_fStoresDIRTY = true;

		if( spitem->m_flags & flagWasNeeded )
		    spitemUndo->m_flags |= flagWasNeeded;
	    }

	    // Try to undo.

	    if( spitemUndo )
	    {
		UInt16 flags = packNoUndo|packUndoing;

		if( GetUnpackSortMask() )
		    flags |= packForceSort;

		e = PackRecord( record, spitemUndo.UnsafeAddress(), flags, 0 );

		if( e )
		    goto LError;
	    }
	    else
	    {
		e = DeleteRecord( record, false );

		if( e )
		    goto LError;
	    }

	    record = pundo->m_recordDeleted;

	    if( noRecordSelected != record )
	    {
		const unsigned dmRecAfterPack = GetRecordAfterPack() + m_cMetadataRecords;
		const unsigned dmRecord = record + m_cMetadataRecords;

		DmSetRecordInfo( m_pdb, dmRecAfterPack, 0, &pundo->m_uid );
		DmMoveRecord( m_pdb, dmRecAfterPack, dmRecord );

		m_pbits->MoveRecord( GetRecordAfterPack(), record );
		m_recAfterPack = record;

		SetDirty();
	    }

	    // Return info.

	    if( precord )
		*precord = GetRecordAfterPack();
	    if( pfDidUndo )
		*pfDidUndo = true;
	}

LError:
	if( e )
	    ClearUndo();

	return e;
}

void
DbOpenHs2::ClearUndo()
{
	while( m_cUndo )
	{
	    m_cUndo--;

	    if( m_rgUndo[m_cUndo].m_pitem )
	    {
		delete m_rgUndo[m_cUndo].m_pitem;
		m_rgUndo[m_cUndo].m_pitem = 0;
	    }
	}
}

/*
 * Categories.
 */

unsigned
DbOpenHs2::GetCategoriesCount() const
{
	Assert( m_pCategories );
	return m_pCategories->Count();
}

unsigned
DbOpenHs2::GetCategoriesMaximum() const
{
	Assert( m_pCategories );
	Assert( MAX_CATEGORIES == m_pCategories->m_cMax );
	return m_pCategories->m_cMax;
}

LPCTSTR
DbOpenHs2::GetCategoryNameById( WORD const category ) const
{
	Assert( m_pdb );
	Assert( m_pCategories );

	if( category == catAllCategories )
	    return AllString();
	if( category == catUnfiledCategory )
	    return UnfiledString();

	m_iCat = !m_iCat;

	if( category >= GetCategoriesMaximum() )
	{
	    m_rgCat[m_iCat].Set( TEXT( "(Unknown)" ) );
	}
	else
	{
	    char const* const psz = m_pCategories->CategoryFromId( category )->m_szCategory;

	    m_rgCat[m_iCat].SetA( psz );
	}

	return m_rgCat[m_iCat].Text();
}

LPCTSTR
DbOpenHs2::GetCategoryNameByIndex( int index ) const
{
	Assert( m_pdb );
	Assert( m_pCategories );
	Assert( index >= 0 );
	Assert( unsigned( index ) < GetCategoriesMaximum() );

	WORD const category = m_pCategories->IdFromIndex( UInt8( index ) );

	return GetCategoryNameById( category );
}

WORD
DbOpenHs2::GetCategoryIdByName( LPCTSTR pszCategory ) const
{
	Assert( m_pdb );
	Assert( m_pCategories );
	Assert( pszCategory );
	Assert( *pszCategory );

	char sz[dmCategoryLength];

	TCharToNarrowString( pszCategory, RgC( sz ) );

	return m_pCategories->IdFromName( sz );
}

WORD
DbOpenHs2::GetCategoryIdByIndex( int index ) const
{
	Assert( m_pdb );
	Assert( m_pCategories );
	Assert( index >= 0 );
	Assert( unsigned( index ) < GetCategoriesMaximum() );

	Assert( *m_pCategories->CategoryFromIndex( UInt8( index ) )->m_szCategory );

	return m_pCategories->IdFromIndex( UInt8( index ) );
}

int
DbOpenHs2::GetCategoryIndexById( WORD const category ) const
{
	Assert( m_pdb );
	Assert( m_pCategories );
	Assert( category < GetCategoriesMaximum() );

	UInt8 const index = m_pCategories->IndexFromId_Slow( category );

	if( index == UInt8( -1 ) )
	    return -1;

	return index;
}

WORD
DbOpenHs2::GetCategoryIcon( WORD const category ) const
{
	Assert( m_pdb );
	Assert( m_pCategories );

	if( category < GetCategoriesMaximum() )
	    return m_pCategories->CategoryFromId( category )->GetIcon();

	return 0;
}

bool
DbOpenHs2::CategoryExistsById( WORD const category )
{
	Assert( m_pdb );
	Assert( m_pCategories );
	Assert( category < GetCategoriesMaximum() );

	return !!m_pCategories->CategoryFromId( category )->m_szCategory;
}

bool
DbOpenHs2::CategoryHasRecords( WORD const category )
{
	Assert( m_pdb );
	Assert( m_pCategories );

	MemHandle h;
	UInt8* pRec;
	unsigned const cMax = DmNumRecords( m_pdb );
	bool fHasRecords = false;
	ScopedProgressHolder sph;

	sph.Create();
	sph.Init( cMax, PROGRESS_PREDICTIVEWAIT );
	for( UInt16 record = 0; !fHasRecords && record < cMax; record++ )
	{
	    sph.Update( record );
	    h = DmGetRecord( m_pdb, record );
	    pRec = ( UInt8* )MemHandleLock( h );
	    if( tagMasterStoreList != *pRec && tagCategoryList != *pRec )
	    {
		if( category == dmAllCategories )
		{
		    fHasRecords = true;
		}
		else
		{
		    UInt16 categoryRecord = catUnfiledCategory;
		    UInt8 const* const pCategory = reinterpret_cast<UInt8 const*>( GetMemberPtr( pRec, tagCategory ) );
		    if( pCategory )
			categoryRecord = *pCategory;
		    fHasRecords = ( categoryRecord == category );
		}
	    }
	    MemHandleUnlock( h );
	    DmReleaseRecord( m_pdb, record, false );
	}

	return fHasRecords;
}

bool
DbOpenHs2::CategoryNew( LPCTSTR pszCategory, WORD icon )
{
	Assert( m_pdb );
	Assert( m_pCategories );

	// Can't create a new category if there aren't any more available.

	if( GetCategoriesCount() >= GetCategoriesMaximum() )
	    return false;

	if( !pszCategory )
	    return true;

	// Can't create a new category if one already exists by the same name.

	StrA s1;
	StrA s2;

	s1.SetT( pszCategory );
	s2.SetT( UnfiledString() );

	if( !strcmp( s1.Text(), s2.Text() ) || GetCategoryIdByName( pszCategory ) != catUnfiledCategory )
	    return false;

	// Create a new category.

	ClearUndo();

	StrA name;
	name.SetT( pszCategory );

	m_pCategories->New( name.Text(), icon );
	SetCategoriesDirty();

	return true;
}

bool
DbOpenHs2::CategoryMerge( WORD const oldCategory, WORD const newCategory )
{
	Assert( m_pdb );
	Assert( m_pCategories );
	Assert( oldCategory < GetCategoriesMaximum() );
	Assert( newCategory < GetCategoriesMaximum() || newCategory == catUnfiledCategory );

	bool fMerged = false;

	ClearUndo();

	Assert( m_pCategories->CategoryFromId( oldCategory )->m_szCategory[0] );
	Assert( m_pCategories->CategoryFromId( newCategory )->m_szCategory[0] );

	if( m_pCategories->CategoryFromId( oldCategory )->m_szCategory[0] &&
	    m_pCategories->CategoryFromId( newCategory )->m_szCategory[0] )
	{
	    // Recategorize records.

	    MoveCategory( newCategory, oldCategory );

	    // Remove the category name.

	    if( oldCategory == GetCategory() )
		SetCategory( newCategory );

	    CategoryDelete( oldCategory );

	    fMerged = true;
	}

	return fMerged;
}

bool
DbOpenHs2::CategoryRename( WORD const category, LPCTSTR const pszCategory )
{
	Assert( m_pdb );
	Assert( m_pCategories );
	Assert( category < GetCategoriesMaximum() );
	Assert( pszCategory );
	Assert( *pszCategory );

	bool fRenamed = false;

	Assert( category < GetCategoriesMaximum() );

	ClearUndo();

	Assert( m_pCategories->CategoryFromId( category )->m_szCategory[0] );

	if( m_pCategories->CategoryFromId( category )->m_szCategory[0] )
	{
	    StrA name;
	    name.SetT( pszCategory );

	    m_pCategories->SetName( category, name.Text() );
	    SetCategoriesDirty();

	    fRenamed = true;
	}

	return fRenamed;
}

bool
DbOpenHs2::CategoryDelete( WORD const category )
{
	Assert( m_pdb );
	Assert( m_pCategories );
	Assert( category < GetCategoriesMaximum() );

	ClearUndo();

	Assert( m_pCategories->CategoryFromId( category )->m_szCategory[0] );

	// Recategorize records as unfiled.

	MoveCategory( catUnfiledCategory, category );

	// Remove the category name.

	m_pCategories->Remove( category );
	SetCategoriesDirty();

	// Update selected category.

	if( category == GetCategory() )
	    SetCategory( catUnfiledCategory );

	return true;
}

bool
DbOpenHs2::SetCategoryIcon( WORD const category, WORD const icon )
{
	Assert( m_pdb );
	Assert( m_pCategories );
	Assert( category < GetCategoriesMaximum() );

	if( category < GetCategoriesMaximum() )
	{
	    if( m_pCategories->SetIcon( category, icon ) )
	    {
		SetCategoriesDirty();
		return true;
	    }
	}

	return false;
}

/*
 * Stores.
 */

unsigned
DbOpenHs2::GetStoresCount() const
{
	Assert( m_pStores );

	return m_pStores->m_cStores;
}

unsigned
DbOpenHs2::GetStoresMaximum() const
{
	Assert( m_pStores );
	Assert( MAX_STORES == m_pStores->m_cMax );

	return m_pStores->m_cMax;
}

LPCTSTR
DbOpenHs2::GetStoreNameByIndex( int index ) const
{
	Assert( m_pStores );

	LPCTSTR pszStore = TEXT( "" );

	if( index >= 0 && index < m_pStores->m_cStores )
	    pszStore = GetStoreNameById( m_pStores->IdFromIndex( UInt8( index ) ) );

	return pszStore;
}

LPCTSTR
DbOpenHs2::GetStoreNameById( BYTE const idStore ) const
{
	Assert( m_pdb );
	Assert( m_pCategories );

	if( idStore == storeAllStores )
	    return AllStoresString();
	if( idStore == storeUnfiledStore )
	    return UnfiledString();

	m_iStore = !m_iStore;

	if( idStore >= GetStoresMaximum() )
	{
	    m_rgStore[m_iStore].Set( TEXT( "(Unknown)" ) );
	}
	else
	{
	    char const* const pszStore = m_pStores->m_rgStore[idStore].m_szStore;

	    m_rgStore[m_iStore].SetA( pszStore );
	}

	return m_rgStore[m_iStore].Text();
}

static int
CmpStoreName( const void* pvFind, UInt16 index, void* pvCookie )
{
	MasterStoreList* pmsl = static_cast<MasterStoreList*>( pvCookie );
	return StrCompare( LPCSTR( pvFind ), pmsl->m_rgStore[pmsl->m_rgIndexToId[index]].m_szStore );
}

BYTE
DbOpenHs2::GetStoreIdByName( LPCTSTR pszStore ) const
{
	Assert( pszStore );
	Assert( m_pStores );

	BYTE idStore = noStore;
	StrA s;
	UInt16 index;

	s.SetT( pszStore );

	if( FBinarySearch( s.Text(), m_pStores, m_pStores->m_cStores, &index, CmpStoreName ) )
	    idStore = m_pStores->m_rgIndexToId[index];

	return idStore;
}

BYTE
DbOpenHs2::GetStoreIdByIndex( int index ) const
{
	Assert( m_pStores );

	BYTE idStore = noStore;

	if( index >= 0 && index < m_pStores->m_cStores )
	    idStore = m_pStores->IdFromIndex( UInt8( index ) );

	return idStore;
}

int
DbOpenHs2::GetStoreIndexById( BYTE idStore ) const
{
	Assert( m_pStores );

	return static_cast<int>( static_cast<signed char>( m_pStores->IndexFromId_Slow( idStore ) ) );
}

unsigned
DbOpenHs2::GetStoreNeededItemsCount( BYTE idStore ) const
{
	Assert( m_pStores );

	unsigned cNeeded = 0;

	if( allStores == idStore )
	    cNeeded = m_prefs.get_cBoldAll();
	else if( noStore == idStore )
	    cNeeded = m_prefs.get_cBoldNo();
	else if( idStore >= 0 && idStore < DimensionOf( m_pStores->m_rgStore ) && idStore < m_pStores->m_cMax )
	    cNeeded = m_pStores->m_rgStore[idStore].get_m_cBold();

	return cNeeded;
}

bool
DbOpenHs2::StoreExistsById( BYTE idStore )
{
	Assert( idStore >= 0 && idStore < GetStoresMaximum() );
	Assert( m_pStores );

	return !!m_pStores->m_rgStore[idStore].m_szStore[0];
}

bool
DbOpenHs2::StoreHasRecords( BYTE idStore )
{
	for( unsigned record = Count(); record--; )
	{
	    SPNew<Item> spitem = UnPackRecord( record /*, maskStores*/ );

	    if( !spitem )
		continue;

	    if( spitem->m_stores.Find( idStore ) )
		return true;
	}

	return false;
}

bool
DbOpenHs2::StoreNew( LPCTSTR pszStore )
{
	if( GetStoresCount() >= GetStoresMaximum() )
	    return false;

	if( pszStore )
	{
	    if( GetStoreIdByName( pszStore ) != noStore )
		return false;

	    ClearUndo();

	    StrA s;
	    s.SetT( pszStore );

	    StoreNewRaw( s.Text() );

	    SetDirty();
	}

	return true;
}

bool
DbOpenHs2::StoreNewWithId( LPCTSTR pszStore, BYTE id )
{
	if( GetStoresCount() >= GetStoresMaximum() )
	    return false;

	Assert( m_pStores );

	if( m_pStores->m_cStores >= m_pStores->m_cMax )
	    return false;

	if( pszStore )
	{
	    if( GetStoreIdByName( pszStore ) != noStore )
		return false;

	    ClearUndo();

	    if( id >= m_pStores->m_cMax )
		return false;

	    StrA s;
	    s.SetT( pszStore );

	    FillInStore( id, s.Text(), true );

	    SetDirty();
	}

	return true;
}

bool
DbOpenHs2::StoreMerge( BYTE oldStore, BYTE newStore )
{
	Assert( oldStore <= GetStoresMaximum() );
	Assert( newStore <= GetStoresMaximum() );
	Assert( m_pStores );
	Assert( m_pStores->m_rgStore[oldStore].m_szStore[0] );
	Assert( m_pStores->m_rgStore[newStore].m_szStore[0] );

	ClearUndo();

	// Update store associations.  This could be done more efficiently,
	// but it doesn't seem worth it just now since the HS2 database format
	// is supported mainly just for prototyping.

	for( unsigned record = Count(); record--; )
	{
	    WORD ces;
	    SPNew<Item> spitem = UnPackRecord( record );

	    // Copy old store associations.

	    Assert( spitem->m_storesORIG.FPristine() );

	    if( spitem->m_stores.Count() )
	    {
		//$ review: (chrisant) Error handling...
		spitem->m_fStoresDIRTY = !!spitem->m_storesORIG.Copy( spitem->m_stores );
	    }

	    // Rebuild list of stores, minus the old store.

	    spitem->m_stores.Clear();

	    for( WORD ii = spitem->m_storesORIG.Count(); ii--; )
		if( spitem->m_storesORIG[ii] != oldStore )
		    spitem->m_stores.Insert( spitem->m_storesORIG[ii] );

		// Insert the new store (a no-op if it already exists).

	    spitem->m_stores.Insert( newStore );

	    // Pack the updated item.

	    PackRecord( record, spitem.UnsafeAddress(), packNoSort|packNoUndo|packNoDelete, &ces );
	}

	// Clear the entry in the master store list.

	ZeroMemory( &m_pStores->m_rgStore[oldStore], sizeof( m_pStores->m_rgStore[oldStore] ) );

	// Shift other store info.

	int const index = GetStoreIndexById( oldStore );

	if( index >= 0 && index < int( m_pStores->m_cStores ) )
	{
	    CASSERT( sizeof( UInt8 ) == sizeof( m_pStores->m_rgIndexToId[0] ) );

	    memcpy( &m_pStores->m_rgIndexToId[index],
		    &m_pStores->m_rgIndexToId[index + 1],
		    m_pStores->m_cStores - index );
	}

	SetStoresDirty();

	// Write app info, to save the updated bold counts.

	WriteAppInfo();

	// Update selected store.

	if( oldStore == GetStore() )
	    SetStore( newStore );

	return true;
}

bool
DbOpenHs2::StoreRename( BYTE idStore, LPCTSTR pszStore )
{
	Assert( idStore < GetStoresMaximum() );
	Assert( pszStore );
	Assert( m_pStores );
	Assert( m_pStores->m_rgStore[idStore].m_szStore[0] );

	ClearUndo();

	StrA s;
	s.SetT( pszStore );

	Assert( s.Length() < DimensionOf( m_pStores->m_rgStore[idStore].m_szStore ) );

	FillInStore( idStore, s.Text(), false );

	SetDirty();

	return true;
}

void
DbOpenHs2::StoreDeleteIndexEntryRaw( UInt8 index )
{
	Assert( m_pStores );

	// Decrement number of stores.

	IncStore( -1 );

	// Shift other store info.

	if( index < m_pStores->m_cStores )
	{
	    CASSERT( sizeof( UInt8 ) == sizeof( m_pStores->m_rgIndexToId[0] ) );

	    memcpy( &m_pStores->m_rgIndexToId[index],
		    &m_pStores->m_rgIndexToId[index + 1],
		    m_pStores->m_cStores - index );
	}

	SetStoresDirty();
}

bool
DbOpenHs2::StoreDelete( BYTE idStore )
{
	Assert( idStore < GetStoresMaximum() );
	Assert( m_pStores );

	ClearUndo();

	// Not the items in the store, just the store itself.  Remove this
	// store from all items.

	ZapStore( idStore );

	// Clear the entry in the master store list.

	ZeroMemory( &m_pStores->m_rgStore[idStore], sizeof( m_pStores->m_rgStore[idStore] ) );

	// Delete the store index entry.

	UInt8 const index = m_pStores->IndexFromId_Slow( idStore );

	StoreDeleteIndexEntryRaw( index );

	// Write app info, to save the updated bold counts.

	WriteAppInfo();

	// Update selected store.

	if( idStore == GetStore() )
	    SetStore( noStore );

	SetDirty();

	return true;
}

bool
DbOpenHs2::RepairDuplicateStoreEntries()
{
	unsigned cRepairs = 0;
	unsigned const cStores = GetStoresCount();
	Store idPrev = noStore;

	for( unsigned ii = 0; ii < cStores; ii++ )
	{
	    Store const id = GetStoreIdByIndex( ii );

	    if( id != noStore && id == idPrev )
	    {
		StoreDeleteIndexEntryRaw( UInt8( ii ) );
		cRepairs++;

		ii--;	// Counteract the loop.
	    }

	    idPrev = id;
	}

	return !!cRepairs;
}

/*
 * Settings.
 */

void
DbOpenHs2::SetPrefFlag( UInt16 flag, bool f )
{
	if( !f != !( m_prefs.get_flags() & flag ) )
	{
	    m_prefs.and_flags( ~flag );

	    if( f )
		m_prefs.or_flags( flag );

	    SetDirty();
	}
}

void
DbOpenHs2::SetRoundToWholeNumbers( bool fRoundWholeNumbers )
{
	SetPrefFlag( dpfDecimals, !fRoundWholeNumbers );
}

void
DbOpenHs2::SetCurrency( LPCTSTR pszCurrency )
{
	Assert( pszCurrency );
	Assert( m_pdb );

	StrA s;

	m_sCurrency.Set( pszCurrency );
	s.SetT( pszCurrency );

	if( s.Length() >= DimensionOf( m_prefs.szCurrency ) )
	    s.SetLength( DimensionOf( m_prefs.szCurrency ) - 1 );

	if( StrCompare( m_prefs.szCurrency, s.Text() ) )
	{
	    StrCopy( m_prefs.szCurrency, s.Text() );

	    SetDirty();
	}
}

void
DbOpenHs2::SetTaxRate( WORD wTaxRate )
{
	//$ todo: (chrisant) Validate.

	Assert( wTaxRate >= 0 && wTaxRate <= 10000 );

	if( wTaxRate != m_prefs.get_wTaxRate() )
	{
	    m_prefs.set_wTaxRate( wTaxRate );

	    SetDirty();
	}
}

void
DbOpenHs2::SetTaxRate2( WORD wTaxRate2 )
{
	//$ todo: (chrisant) Validate.

	Assert( wTaxRate2 >= 0 && wTaxRate2 <= 10000 );

	if( wTaxRate2 != m_prefs.get_wTaxRate2() )
	{
	    m_prefs.set_wTaxRate2( wTaxRate2 );

	    SetDirty();
	}
}

void
DbOpenHs2::SetTwoTaxes( bool fTwoTaxes )
{
	SetPrefFlag( dpfTwoTaxes, fTwoTaxes );
}

void
DbOpenHs2::SetPerStorePrices( bool fPerStorePrices )
{
	SetPrefFlag( dpfPerStorePrices, fPerStorePrices );
}

void
DbOpenHs2::SetAutoQuantity( bool fAutoQuantity )
{
	SetPrefFlag( dpfNoAutoQty, !fAutoQuantity );
}

void
DbOpenHs2::SetAutoDelete( bool fAutoDelete )
{
	SetPrefFlag( dpfAutoDelete, fAutoDelete );
}

void
DbOpenHs2::SetRecordCreationDate( bool f )
{
	SetPrefFlag( dpfStampToday, f );
}

void
DbOpenHs2::SetRecordCompletionDate( bool f )
{
	SetPrefFlag( dpfNoCompDate, !f );
}

void
DbOpenHs2::SetTaxable( bool f )
{
	SetPrefFlag( dpfDefTaxable, f );
}

void
DbOpenHs2::SetDefaultPriority( BYTE prio )
{
	if( prio < 0 || prio > MAX_PRIO )
	    prio = 0;

	if( prio != m_prefs.prioDefault )
	{
	    m_prefs.prioDefault = prio;

	    SetDirty();
	}
}

void
DbOpenHs2::SetRowBackgroundColor( COLORREF crBackground )
{
	RGBColorType rgb = {};
	rgb.r = GetRValue( crBackground );
	rgb.g = GetGValue( crBackground );
	rgb.b = GetBValue( crBackground );

	if( memcmp( &m_prefs.rgbRowColor, &rgb, sizeof( rgb ) ) != 0 )
	{
	    m_prefs.rgbRowColor = rgb;
	    SetDirty();
	}
}

void
DbOpenHs2::SetFont( FontId id )
{
	UInt8 const font = PalmFontFromFontId( id ) + 1;

	if( m_prefs.fontPlusOne != font )
	{
	    m_prefs.fontPlusOne = font;
	    SetDirty();
	}
}

void
DbOpenHs2::SetMaxLinesPerItem( int cLines )
{
	Assert( cLines >= 0 && cLines <= 3 );

	if( m_prefs.cMaxLinesPerItem != cLines )
	{
	    // DbOpenHs2::GetMaxLinesPerItem is the primary location where we
	    // force the value into bounds, but we also need to force it into
	    // bounds here just for backward compatibility with HS2, since HS2
	    // isn't as robust about bounds checking.

	    if( cLines < 0 )
		cLines = 0;

	    if( cLines > 3 )
		cLines = 3;

	    m_prefs.cMaxLinesPerItem = UInt8( cLines );
	    SetDirty();
	}
}

void
DbOpenHs2::SetIcon( WORD icon )
{
	// Get AppInfo pointer.

	UInt16 version;
	LocalID idAppInfo;

	DmDatabaseInfo( m_pdb, 0, 0, &version, 0, 0, 0, 0, &idAppInfo, 0, 0, 0 );

	Assert( version >= DbVersion_alarmcache );

	ThisAppInfoPtr const pAppInfo = ThisAppInfoPtr( MemLocalIDToLockedPtr( idAppInfo, CARDNO ) );

	Assert( pAppInfo );

	pAppInfo->alarmCache.set_m_icon( icon );

	__MemPtrUnlock( pAppInfo );
}

bool
DbOpenHs2::GetRoundToWholeNumbers() const
{
	return !( m_prefs.get_flags() & dpfDecimals );
}

LPCTSTR
DbOpenHs2::GetCurrency() const
{
	Assert( m_pdb );

	if( !m_sCurrency.Length() )
	    m_sCurrency.SetA( m_prefs.szCurrency );

	return m_sCurrency.Text();
}

WORD
DbOpenHs2::GetTaxRate() const
{
	return m_prefs.get_wTaxRate();
}

WORD
DbOpenHs2::GetTaxRate2() const
{
	return m_prefs.get_wTaxRate2();
}

bool
DbOpenHs2::GetTwoTaxes() const
{
	return !!( m_prefs.get_flags() & dpfTwoTaxes );
}

bool
DbOpenHs2::GetPerStorePrices() const
{
	return !!( m_prefs.get_flags() & dpfPerStorePrices );
}

bool
DbOpenHs2::GetAutoQuantity() const
{
	return !( m_prefs.get_flags() & dpfNoAutoQty );
}

bool
DbOpenHs2::GetAutoDelete() const
{
	return !!( m_prefs.get_flags() & dpfAutoDelete );
}

bool
DbOpenHs2::GetRecordCreationDate() const
{
	return !!( m_prefs.get_flags() & dpfStampToday );
}

bool
DbOpenHs2::GetRecordCompletionDate() const
{
	return !( m_prefs.get_flags() & dpfNoCompDate );
}

bool
DbOpenHs2::GetTaxable() const
{
	return !!( m_prefs.get_flags() & dpfDefTaxable );
}

BYTE
DbOpenHs2::GetDefaultPriority() const
{
	Assert( m_prefs.prioDefault >= 0 && m_prefs.prioDefault < MAX_PRIO );

	return m_prefs.prioDefault;
}

COLORREF
DbOpenHs2::GetRowBackgroundColor() const
{
	return RGB( m_prefs.rgbRowColor.r,
		    m_prefs.rgbRowColor.g,
		    m_prefs.rgbRowColor.b );
}

FontId
DbOpenHs2::GetFont() const
{
	return FontIdFromPalmFont( m_prefs.fontPlusOne - 1 );
}

int
DbOpenHs2::GetMaxLinesPerItem() const
{
	int const cLines = m_prefs.cMaxLinesPerItem;

	if( cLines < 0 )
	    return 0;

	if( cLines > 3 )
	    return 3;

	return cLines;
}

WORD
DbOpenHs2::GetIcon() const
{
	// Get AppInfo pointer.

	UInt16 version;
	LocalID idAppInfo;

	DmDatabaseInfo( m_pdb, 0, 0, &version, 0, 0, 0, 0, &idAppInfo, 0, 0, 0 );

	Assert( version >= DbVersion_alarmcache );

	ThisAppInfoPtr const pAppInfo = ThisAppInfoPtr( MemLocalIDToLockedPtr( idAppInfo, CARDNO ) );

	Assert( pAppInfo );

	WORD const icon = pAppInfo->alarmCache.get_m_icon();

	__MemPtrUnlock( pAppInfo );

	return icon;
}

DWORD
DbOpenHs2::GetNextAlarm() const
{
	// Get AppInfo pointer.

	UInt16 version;
	LocalID idAppInfo;

	DmDatabaseInfo( m_pdb, 0, 0, &version, 0, 0, 0, 0, &idAppInfo, 0, 0, 0 );

	Assert( version >= DbVersion_alarmcache );

	ThisAppInfoPtr const pAppInfo = ThisAppInfoPtr( MemLocalIDToLockedPtr( idAppInfo, CARDNO ) );

	Assert( pAppInfo );

	DWORD const alarm = pAppInfo->alarmCache.get_m_seconds();

	__MemPtrUnlock( pAppInfo );

	return alarm;
}

/*
 * Export.
 */

void
DbOpenHs2::SetExportFlags( WORD flags )
{
	if( flags != m_prefs.get_wExportFlags() )
	{
	    m_prefs.set_wExportFlags( flags );
	    SetDirty();
	}
}

WORD
DbOpenHs2::GetExportFlags() const
{
	return m_prefs.get_wExportFlags();
}

/*
 * Maintenance.
 */

bool
DbOpenHs2::IsCauterized() const
{
	return false;
}

/*
 * Transactions.
 */

class Hs2DbGroupTransaction : public DbGroupTransaction
{
	typedef DbGroupTransaction base;

    public:
	Override( base, SysError, End, () );
};

SysError
Hs2DbGroupTransaction::End()
{
	if( IsActive() )
	    MarkEnded();

	return SE_OK;
}

SysError
DbOpenHs2::NewDbGroupTransaction( DbGroupTransaction** ppOut )
{
	*ppOut = new Hs2DbGroupTransaction;

	if( !*ppOut )
	    return SE_OUTOFMEMORY;

	return SE_OK;
}

/*
 * Debugging helpers.
 */

#ifdef DEBUG
void
DbOpenHs2::AssertNoRecordLocks( LPCTSTR pszContext )
{
	SUPPRESS_UNUSED_VAR( pszContext );

	for( unsigned ii = 0; ii < DmNumRecords( m_pdb ); ii++ )
	{
	    MemHandle const h = DmQueryRecord( m_pdb, ii );

	    if( h && MemHandleLockCount( h ) > 0 )
	    {
		DbgMsg(( 20, TEXT( "%s:\nrecord %u has %u locks" ), pszContext, ii, MemHandleLockCount( h ) ));
		DbgDump( MemHandleLock( h ), MemHandleSize( h ) );
		MemHandleUnlock( h );
	    }
	}
}

WORD
DbOpenHs2::GetDbVersion()
{
	UInt16 version;

	DmDatabaseInfo( m_pdb, 0, 0, &version, 0, 0, 0, 0, 0, 0, 0, 0 );

	return version;
}

bool
DbOpenHs2::HasRollbackFlag()
{
	return false;
}
#endif

/*
 * Create and open databases.
 */

void
MakeNewHs2DatabaseFileName( LPCTSTR pszDir, LPCTSTR const pszName, StrT* const psFile )
{
	Assert( pszDir );
#ifdef OS_WINDOWSMOBILE
	Assert( *pszDir == '\\' );
#endif
	Assert( pszName );
	Assert( *pszName );
	Assert( psFile );
	Assert( !psFile->Length() );

	StrT sTmp;
	StrT sName;

#ifndef _WIN32_WCE
	if( !*pszDir )
	    pszDir = TEXT( "." );
#endif

	// If the name ends with '.pdb' strip the extension so we don't end up
	// with a redundant '.pdb' at the end.

	sName.Set( pszName );

	TCHAR const* const pszExt = GetExtension( sName.Text() );

	if( pszExt && !LStrCmpI( pszExt, TEXT( ".pdb" ) ) )
	    sName.SetLength( sName.Length() - 4 );

	// Sanitize the display name.

	sName.SanitizeFileName();

	// Build the file name.

	psFile->Printf( TEXT( "%s\\HS2_%s.pdb" ), pszDir, sName.Text() );
}

SysError
NewHs2Database( LPCTSTR const pszFile, LPCTSTR const pszDisplayName, DbOpen** const ppdb )
{
	return DbOpenHs2::NewDatabase( pszFile, pszDisplayName, ppdb );
}

SysError
OpenHs2Database( LPCTSTR const pszFile, DbOpen** const ppdb )
{
	return DbOpenHs2::OpenDatabase( pszFile, ppdb );
}

