/*
 * IoCsv.cpp - CSV output interface for databases.
 */

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#include "precomp.h"

#ifdef OS_PALM
#error The converter is not designed for the Palm OS.
#endif

#include "std.h"
#include "str.h"
#include "sys.h"
#include "converter.h"
#include "storelist.h"
#include "number.h"
#include "item.h"
#include "dbopen.h"

static TCHAR const c_szExtItems[] = TEXT( ".csv" );
static TCHAR const c_szExtMetadata[] = TEXT( ".txt" );
static DateFormatType s_dateFormat = DATEFORMAT_AUTO;
static StrT s_sCurrentStore;

void SkipPrefix( char const** ppsz );
void SkipPrefix( TCHAR const** ppsz );

extern void DumpBytes( FILE* const pout, char const* const pszTag, void const* const pv, unsigned cb );

static bool
FixStoreName( StrA& s )
{
	bool fFixed = false;

	for( char const* psz = s.Text(); *psz; psz++ )
	    if( *psz == ';' || *psz == '=' )
	    {
		s.SetAt( psz, '_' );
		fFixed = true;
	    }

	return fFixed;
}

inline bool
IsDigit( TCHAR const ch )
{
	return ( ch >= '0' && ch <= '9' );
}

#ifdef UNICODE
inline bool
IsDigit( char const ch )
{
	return IsDigit( TCHAR( ch ) );
}
#endif

void
SetDateFormat( DateFormatType const format )
{
	s_dateFormat = format;
}

DateFormatType
GetDateFormat()
{
	return s_dateFormat;
}

static void
OutputErrorAndString( char const* const pszError, char const* const pszString )
{
	fprintf( stderr, "%s '%s'.\n", pszError, pszString );
}

static void
OutputErrorAndString( char const* const pszError, WCHAR const* const pszString )
{
	fprintf( stderr, "%s '%S'.\n", pszError, pszString );
}

template<class T> int
ParseUnsigned( T const** ppsz, unsigned* pn, T const chTerminate = ',' )
{
	T const* pszOriginal = *ppsz;

	*pn = 0;

	while( **ppsz >= '0' && **ppsz <= '9' )
	{
	    if( ( *pn ) * 10 < *pn )
	    {
		fprintf( stderr, "Numeric value out of range.\n" );
		return -1;
	    }

	    ( *pn ) *= 10;
	    ( *pn ) += ( **ppsz ) - '0';

	    ( *ppsz )++;
	}

	if( !**ppsz || !chTerminate )
	    return 0;

	if( **ppsz == chTerminate )
	{
	    ( *ppsz )++;
	    return 0;
	}

	OutputErrorAndString( "Invalid number", pszOriginal );
	return -1;
}

/*
 * __mklocaltime - Palm OS alarm times are local times in a variation on the
 * time_t format, which makes it difficult to convert them using standard CRT
 * calls, much less using Win32 calls.
 *
 * This implementation is dangerous and changes the _timezone and _daylight
 * CRT variables.  It restores them afterwards, but is not threadsafe with
 * respect to the variables and other code that might use them.
 */

static time_t
__mklocaltime( struct tm* pTm )
{
#if defined( _MSC_VER )
#pragma warning( push )
#pragma warning( disable: 4996 )
#endif

	//$ WARNING: (chrisant) DANGER, ALTERING SYSTEM GLOBALS!

	long const oldTimezone = _timezone;
	int const oldDaylight = _daylight;

	_timezone = 0;
	_daylight = 0;

	time_t const time = mktime( pTm );

	_timezone = oldTimezone;
	_daylight = oldDaylight;

#if defined( _MSC_VER )
#pragma warning( pop )
#endif

	return time;
}

/*
 * Schema (xfer functions).
 */

typedef int ( *XferColumnFn )( DbOpen* pdb, StrA& s, Item* pitem, bool fOut, bool fUseFriendlyNames, AppendQuoteFlags flags );

int
XferFlag( StrA& s, Item* pitem, bool fOut, WORD flag )
{
	if( fOut )
	{
	    if( pitem->m_flags & flag )
		s.Append( "x" );
	}
	else
	{
	    pitem->m_flags &= ~flag;

	    if( s.Length() )
		pitem->m_flags |= flag;
	}

	return 0;
}

int
XferNeed( DbOpen* pdb, StrA& s, Item* pitem, bool fOut, bool fUseFriendlyNames, AppendQuoteFlags flags )
{
	SUPPRESS_UNUSED_VAR( pdb );
	SUPPRESS_UNUSED_VAR( fUseFriendlyNames );
	SUPPRESS_UNUSED_VAR( flags );

	if( fOut )
	{
	    switch( pitem->m_flags & ( flagNeed|flagPurchased ) )
	    {
	    case flagNeed:
		s.Append( "x" );
		break;
	    case flagPurchased:
		s.Append( "have" );
		break;
	    case flagNeed|flagPurchased:
		s.Append( "deferred" );
		break;
	    case 0:
		break;
	    default:
		Panic();
		break;
	    }
	}
	else
	{
	    pitem->m_flags &= ~( flagNeed|flagPurchased );

	    if( s.Length() )
	    {
		if( LStrCmpI( s.Text(), "have" ) == 0 )
		    pitem->m_flags |= flagPurchased;
		else if( LStrCmpI( s.Text(), "deferred" ) == 0 )
		    pitem->m_flags |= flagNeed|flagPurchased;
		else
		    pitem->m_flags |= flagNeed;
	    }
	}

	return 0;
}

int
XferPriority( DbOpen* pdb, StrA& s, Item* pitem, bool fOut, bool fUseFriendlyNames, AppendQuoteFlags flags )
{
	SUPPRESS_UNUSED_VAR( pdb );
	SUPPRESS_UNUSED_VAR( fUseFriendlyNames );
	SUPPRESS_UNUSED_VAR( flags );

	if( fOut )
	{
	    s.Printf( "%d", pitem->GetPriority() + 1 );
	}
	else
	{
	    int n = atoi( s.Text() );

	    if( n > MAX_PRIO )
		n = MAX_PRIO;
	    n--;
	    if( n < 0 )
		n = 0;

	    pitem->SetPriority( BYTE( n ) );
	}

	return 0;
}

int
XferDescription( DbOpen* pdb, StrA& s, Item* pitem, bool fOut, bool fUseFriendlyNames, AppendQuoteFlags flags )
{
	SUPPRESS_UNUSED_VAR( pdb );
	SUPPRESS_UNUSED_VAR( fUseFriendlyNames );

	if( fOut )
	{
	    StrA tmp;
	    tmp.SetT( pitem->GetDescription()->Text() );

	    s.AppendQuote( tmp.Text(), flags );
	}
	else
	{
	    Assert( !pitem->GetDescription()->Length() );

	    if( s.Length() )
	    {
		StrT desc;
		desc.SetA( s.Text() );

		pitem->SetDescription( &desc );
	    }
	}

	return 0;
}

int
XferCustomText( DbOpen* pdb, StrA& s, Item* pitem, bool fOut, bool fUseFriendlyNames, AppendQuoteFlags flags )
{
	SUPPRESS_UNUSED_VAR( pdb );
	SUPPRESS_UNUSED_VAR( fUseFriendlyNames );

	if( fOut )
	{
	    StrA tmp;
	    tmp.SetT( pitem->GetCustomText()->Text() );

	    s.AppendQuote( tmp.Text(), flags );
	}
	else
	{
	    Assert( !pitem->GetCustomText()->Length() );

	    if( s.Length() )
	    {
		StrT customText;
		customText.SetA( s.Text() );

		pitem->SetCustomText( &customText );
	    }
	}

	return 0;
}

int
XferCategory( DbOpen* pdb, StrA& s, Item* pitem, bool fOut, bool fUseFriendlyNames, AppendQuoteFlags flags )
{
	SUPPRESS_UNUSED_VAR( pdb );
	SUPPRESS_UNUSED_VAR( fUseFriendlyNames );

	if( fOut )
	{
	    StrA tmp;
	    tmp.SetT( pdb->GetCategoryNameById( pitem->GetCategory() ) );

	    s.AppendQuote( tmp.Text(), flags );
	}
	else
	{
	    if( s.Length() )
	    {
		StrT categoryName;
		categoryName.SetA( s.Text() );

		WORD const category = pdb->GetCategoryIdByName( categoryName.Text() );
		TCHAR const* const pszCategory = pdb->GetCategoryNameById( category );

		if( !categoryName.Equal( pszCategory ) )
		{
		    fprintf( stderr, "Category '%s' does not exist.\n", s.Text() );
		    return -1;
		}

		pitem->SetCategory( category );
	    }
	    else
	    {
		pitem->SetCategory( catUnfiledCategory );
	    }
	}

	return 0;
}

int
XferQuantity( DbOpen* pdb, StrA& s, Item* pitem, bool fOut, bool fUseFriendlyNames, AppendQuoteFlags flags )
{
	SUPPRESS_UNUSED_VAR( pdb );
	SUPPRESS_UNUSED_VAR( fUseFriendlyNames );
	SUPPRESS_UNUSED_VAR( flags );

	if( fOut )
	{
	    if( pitem->GetQuantity() )
	    {
		StrA tmp;
		tmp.SetT( LocalizeNumber( pitem->GetQuantity(), lnTrimRightZeros, QTY_PRECISION, 0 ) );

		s.AppendQuote( tmp.Text() );
	    }
	}
	else
	{
	    Assert( !pitem->GetQuantity() );

	    StrT tmp;
	    tmp.SetA( s.Text() );

	    long const l = ParseNumber( tmp.Text(), QTY_PRECISION, 0, true );

	    if( l == parseError )
	    {
		OutputErrorAndString( "Invalid number", s.Text() );
		return -1;
	    }

	    pitem->SetQuantity( l, 0 );
	}

	return 0;
}

int
XferUnits( DbOpen* pdb, StrA& s, Item* pitem, bool fOut, bool fUseFriendlyNames, AppendQuoteFlags flags )
{
	SUPPRESS_UNUSED_VAR( pdb );
	SUPPRESS_UNUSED_VAR( fUseFriendlyNames );

	if( fOut )
	{
	    if( pitem->GetUnits()->Length() )
	    {
		StrA tmp;
		tmp.SetT( pitem->GetUnits()->Text() );

		s.AppendQuote( tmp.Text(), flags );
	    }
	}
	else
	{
	    StrT units;
	    units.SetA( s.Text() );

	    if( units.Length() > cchMaxUnit - 1 )
		units.SetLength( cchMaxUnit - 1 );

	    pitem->SetUnits( &units );
	}

	return 0;
}

int
XferPrice( DbOpen* pdb, StrA& s, Item* pitem, bool fOut, bool fUseFriendlyNames, AppendQuoteFlags flags )
{
	SUPPRESS_UNUSED_VAR( pdb );
	SUPPRESS_UNUSED_VAR( fUseFriendlyNames );
	SUPPRESS_UNUSED_VAR( flags );

	if( fOut )
	{
	    if( pitem->GetPrice() )
	    {
		StrA tmp;
		tmp.SetT( LocalizeNumber( pitem->GetPrice(), lnTrimRightZeros2, PRICE_PRECISION, 0 ) );

		s.AppendQuote( tmp.Text() );
	    }
	}
	else
	{
	    Assert( !pitem->GetPrice() );

	    StrT tmp;
	    tmp.SetA( s.Text() );

	    long const l = ParseNumber( tmp.Text(), PRICE_PRECISION, 0, true );

	    if( l == parseError )
	    {
		OutputErrorAndString( "Invalid number", s.Text() );
		return -1;
	    }

	    pitem->SetPrice( l );
	}

	return 0;
}

int
XferAisle( DbOpen* pdb, StrA& s, Item* pitem, bool fOut, bool fUseFriendlyNames, AppendQuoteFlags flags )
{
	SUPPRESS_UNUSED_VAR( pdb );
	SUPPRESS_UNUSED_VAR( fUseFriendlyNames );

	if( fOut )
	{
	    if( pitem->GetAisle()->Length() )
	    {
		StrA aisle;
		aisle.SetT( pitem->GetAisle()->Text() );

		if( strchr( aisle.Text(), '/' ) ||
		    strchr( aisle.Text(), ',' ) ||
		    strchr( aisle.Text(), '=' ) ||
		    strchr( aisle.Text(), '\"' ) )
		{
		    fprintf( stderr, "Aisles cannot contain / or , or = or \" characters.\n" );
		    return -1;
		}

		s.AppendQuote( aisle.Text(), flags );
	    }
	}
	else
	{
	    Assert( !pitem->GetAisle()->Length() );

	    StrT aisle;
	    aisle.SetA( s.Text() );

	    if( aisle.Length() > cchMaxAisle - 1 )
		aisle.SetLength( cchMaxAisle - 1 );

	    pitem->SetAisle( &aisle );
	}

	return 0;
}

static int
ParseDate( char const*& psz, int& nYear, int& nMonth, int& nDay, bool fEnd = true )
{
	char const* pszBegin = psz;
	int n1 = 0;
	int n2 = 0;
	int n3 = 0;

	while( IsDigit( *psz ) )
	    n1 = ( n1 * 10 ) + ( *( psz++ ) - '0' );
	if( *psz == '/' || *psz == '.' || *psz == '-' )
	{
	    psz++;
	    while( IsDigit( *psz ) )
		n2 = ( n2 * 10 ) + ( *( psz++ ) - '0' );
	}
	if( *psz == '/' || *psz == '.' || *psz == '-' )
	{
	    psz++;
	    while( IsDigit( *psz ) )
		n3 = ( n3 * 10 ) + ( *( psz++ ) - '0' );
	}

	while( isspace( *psz ) )
	    psz++;

	if( fEnd && *psz )
	{
	    OutputErrorAndString( "Invalid date", pszBegin );
	    return -1;
	}

	DateFormatType dateFormat = GetDateFormat();

	if( dateFormat == DATEFORMAT_AUTO )
	{
	    if( n1 >= 31 )
		dateFormat = DATEFORMAT_YMD;
	    else if( n3 >= 31 )
	    {
		if( n1 > 12 )
		    dateFormat = DATEFORMAT_DMY;
		else
		    dateFormat = DATEFORMAT_MDY;
	    }
	}

	switch( dateFormat )
	{
	case DATEFORMAT_DMY:
	    nDay = n1;
	    nMonth = n2;
	    nYear = n3;
	    break;
	case DATEFORMAT_YMD:
	    nYear = n1;
	    nMonth = n2;
	    nDay = n3;
	    break;
	case DATEFORMAT_MDY:
	    nMonth = n1;
	    nDay = n2;
	    nYear = n3;
	    break;
	case DATEFORMAT_AUTO:
	    // Couldn't figure it out, so we'll error out below.
	    break;
	default:
	    Panic1( "Unhandled date format %d", dateFormat );
	    break;
	}

	if( nMonth && nDay && nYear >= 0 && nYear < 100 )
	    nYear += 2000;

	int const firstYear	= 1904;
	int const numberOfYears	= 128;
	int const lastYear	= ( firstYear + numberOfYears - 1 );

	if( !nYear || !nMonth || !nDay || nYear < firstYear || nYear > lastYear )
	{
	    OutputErrorAndString( "Invalid date", pszBegin );
	    return -1;
	}

	struct tm test = { 0 };
	test.tm_year = nYear - 1900;
	test.tm_mon = nMonth - 1;
	test.tm_mday = nDay;
	if( mktime( &test ) == time_t( -1 ) )
	{
	    OutputErrorAndString( "Invalid date", pszBegin );
	    return -1;
	}

	return 0;
}

static int
ParseTime( char const*& psz, int& nHour, int& nMinute, int& nSecond, bool fEnd = true )
{
	char const* pszBegin = psz;
	int n1 = 0;
	int n2 = 0;
	int n3 = 0;

	if( *psz == ':' || *psz == '@' || *psz == '.' )
	    psz++;
	while( isspace( *psz ) )
	    psz++;

	while( IsDigit( *psz ) )
	    n1 = ( n1 * 10 ) + ( *( psz++ ) - '0' );
	if( *psz == ':' || *psz == '.' )
	{
	    psz++;
	    while( IsDigit( *psz ) )
		n2 = ( n2 * 10 ) + ( *( psz++ ) - '0' );
	}
	if( *psz == ':' || *psz == '.' )
	{
	    psz++;
	    while( IsDigit( *psz ) )
		n3 = ( n3 * 10 ) + ( *( psz++ ) - '0' );
	}

	while( isspace( *psz ) )
	    psz++;

	if( n1 >= 1 && n1 <= 12 )
	{
	    bool fSkip = true;

	    switch( *psz )
	    {
	    case 'a':
	    case 'A':
		if( n1 == 12 )
		    n1 = 0;
		break;
	    case 'p':
	    case 'P':
		if( n1 != 12 )
		    n1 += 12;
		break;
	    default:
		fSkip = false;
		break;
	    }

	    if( fSkip )
	    {
		psz++;
		if( *psz == 'm' || *psz == 'M' )
		    psz++;

		while( isspace( *psz ) )
		    psz++;
	    }
	}

	if( fEnd && *psz )
	{
	    OutputErrorAndString( "Invalid time", pszBegin );
	    return -1;
	}

	nHour = n1;
	nMinute = n2;
	nSecond = n3;

	if( nHour < 0 || nHour >= 24 ||
	    nMinute < 0 || nMinute >= 60 ||
	    nSecond < 0 || nSecond >= 60 )
	{
	    OutputErrorAndString( "Invalid time", pszBegin );
	    return -1;
	}

	return 0;
}

int
XferDate( DbOpen* pdb, StrA& s, Item* pitem, bool fOut, bool fUseFriendlyNames, AppendQuoteFlags flags )
{
	SUPPRESS_UNUSED_VAR( pdb );
	SUPPRESS_UNUSED_VAR( fUseFriendlyNames );
	SUPPRESS_UNUSED_VAR( flags );

	if( fOut )
	{
	    if( !pitem->FNoDate() )
	    {
		SYSTEMTIME const* const psystime = pitem->GetDate();

		s.Printf( "%04u/%02u/%02u", psystime->wYear, psystime->wMonth, psystime->wDay );
	    }
	}
	else
	{
	    char const* psz = s.Text();
	    int nYear = 0;
	    int nMonth = 0;
	    int nDay = 0;
	    int n;

	    Assert( pitem->FNoDate() );

	    if( !*psz )
		return 0;

	    n = ParseDate( psz, nYear, nMonth, nDay );
	    if( n < 0 )
		return n;

	    SYSTEMTIME systime = { 0 };
	    systime.wYear = WORD( nYear );
	    systime.wMonth = WORD( nMonth );
	    systime.wDay = WORD( nDay );

	    pitem->SetDate( &systime );
	}

	return 0;
}

int
XferAlarm( DbOpen* pdb, StrA& s, Item* pitem, bool fOut, bool fUseFriendlyNames, AppendQuoteFlags flags )
{
	SUPPRESS_UNUSED_VAR( pdb );
	SUPPRESS_UNUSED_VAR( fUseFriendlyNames );
	SUPPRESS_UNUSED_VAR( flags );

	DWORD const c_UnixToPalm = 2082844800;	// difference in seconds between 1970 and 1904

	if( fOut )
	{
	    if( pitem->GetAlarm() && pitem->GetAlarm() >= c_UnixToPalm )
	    {
		time_t time = pitem->GetAlarm() - c_UnixToPalm;
		struct tm tmAlarm;

		if( _gmtime32_s( &tmAlarm, &time ) )
		{
		    fprintf( stderr, "Invalid time.\n" );
		    return -1;
		}

		s.Printf( "%04u/%02u/%02u %02u:%02u:%02u",
			  tmAlarm.tm_year + 1900, tmAlarm.tm_mon + 1, tmAlarm.tm_mday,
			  tmAlarm.tm_hour, tmAlarm.tm_min, tmAlarm.tm_sec );
	    }
	}
	else
	{
	    char const* psz = s.Text();
	    int nYear = 0;
	    int nMonth = 0;
	    int nDay = 0;
	    int nHour = 0;
	    int nMinute = 0;
	    int nSecond = 0;
	    int n;

	    Assert( !pitem->GetAlarm() );

	    if( !*psz )
		return 0;

	    n = ParseDate( psz, nYear, nMonth, nDay, false );
	    if( n < 0 )
		return -1;

	    n = ParseTime( psz, nHour, nMinute, nSecond );
	    if( n < 0 )
		return -1;

	    struct tm tmAlarm;
	    tmAlarm.tm_year = nYear - 1900;
	    tmAlarm.tm_mon = nMonth - 1;
	    tmAlarm.tm_mday = nDay;
	    tmAlarm.tm_hour = nHour;
	    tmAlarm.tm_min = nMinute;
	    tmAlarm.tm_sec = nSecond;
	    tmAlarm.tm_isdst = 0;

	    time_t const time = __mklocaltime( &tmAlarm );

	    if( time == -1 )
	    {
		OutputErrorAndString( "Invalid alarm", s.Text() );
		return -1;
	    }

	    //$ review: (chrisant) wacky alarms too far in the future will
	    // freak it out.
	    pitem->SetAlarm( time + c_UnixToPalm );
	}

	return 0;
}

int
XferAlarmMidi( DbOpen* pdb, StrA& s, Item* pitem, bool fOut, bool fUseFriendlyNames, AppendQuoteFlags flags )
{
	SUPPRESS_UNUSED_VAR( pdb );
	SUPPRESS_UNUSED_VAR( fUseFriendlyNames );
	SUPPRESS_UNUSED_VAR( flags );

	if( fOut )
	    s.Printf( "%u", pitem->GetAlarmMidi() );
	else
	    pitem->SetAlarmMidi( atoi( s.Text() ) );

	return 0;
}

int
XferIcon( DbOpen* pdb, StrA& s, Item* pitem, bool fOut, bool fUseFriendlyNames, AppendQuoteFlags flags )
{
	SUPPRESS_UNUSED_VAR( pdb );
	SUPPRESS_UNUSED_VAR( fUseFriendlyNames );
	SUPPRESS_UNUSED_VAR( flags );

	if( fOut )
	    s.Printf( "%u", pitem->GetRawIcon() );
	else
	    pitem->SetRawIcon( WORD( atoi( s.Text() ) ) );

	return 0;
}

int
XferEntryOrder( DbOpen* pdb, StrA& s, Item* pitem, bool fOut, bool fUseFriendlyNames, AppendQuoteFlags flags )
{
	SUPPRESS_UNUSED_VAR( pdb );
	SUPPRESS_UNUSED_VAR( fUseFriendlyNames );
	SUPPRESS_UNUSED_VAR( flags );

	if( fOut )
	    s.Printf( "%u", pitem->GetEntryOrder() );
	else
	    pitem->SetEntryOrder( DWORD( atoi( s.Text() ) ) );

	return 0;
}

int
XferCoupon( DbOpen* pdb, StrA& s, Item* pitem, bool fOut, bool fUseFriendlyNames, AppendQuoteFlags flags )
{
	SUPPRESS_UNUSED_VAR( pdb );
	SUPPRESS_UNUSED_VAR( fUseFriendlyNames );
	SUPPRESS_UNUSED_VAR( flags );

	return XferFlag( s, pitem, fOut, flagCoupon );
}

int
XferTax( DbOpen* pdb, StrA& s, Item* pitem, bool fOut, bool fUseFriendlyNames, AppendQuoteFlags flags )
{
	SUPPRESS_UNUSED_VAR( pdb );
	SUPPRESS_UNUSED_VAR( fUseFriendlyNames );
	SUPPRESS_UNUSED_VAR( flags );

	return XferFlag( s, pitem, fOut, flagTaxable );
}

int
XferTax2( DbOpen* pdb, StrA& s, Item* pitem, bool fOut, bool fUseFriendlyNames, AppendQuoteFlags flags )
{
	SUPPRESS_UNUSED_VAR( pdb );
	SUPPRESS_UNUSED_VAR( fUseFriendlyNames );
	SUPPRESS_UNUSED_VAR( flags );

	return XferFlag( s, pitem, fOut, flagTaxable2 );
}

int
XferAutoDelete( DbOpen* pdb, StrA& s, Item* pitem, bool fOut, bool fUseFriendlyNames, AppendQuoteFlags flags )
{
	SUPPRESS_UNUSED_VAR( pdb );
	SUPPRESS_UNUSED_VAR( fUseFriendlyNames );
	SUPPRESS_UNUSED_VAR( flags );

	return XferFlag( s, pitem, fOut, flagAutoDelete );
}

int
XferPrivate( DbOpen* pdb, StrA& s, Item* pitem, bool fOut, bool fUseFriendlyNames, AppendQuoteFlags flags )
{
	SUPPRESS_UNUSED_VAR( pdb );
	SUPPRESS_UNUSED_VAR( fUseFriendlyNames );
	SUPPRESS_UNUSED_VAR( flags );

	return XferFlag( s, pitem, fOut, flagPrivate );
}

int
XferNote( DbOpen* pdb, StrA& s, Item* pitem, bool fOut, bool fUseFriendlyNames, AppendQuoteFlags flags )
{
	SUPPRESS_UNUSED_VAR( pdb );
	SUPPRESS_UNUSED_VAR( fUseFriendlyNames );

	if( fOut )
	{
	    StrA tmp;
	    tmp.SetT( pitem->GetNote()->Text() );

	    s.AppendQuote( tmp.Text(), flags );
	}
	else
	{
	    Assert( !pitem->GetNote()->Length() );

	    if( s.Length() )
	    {
		StrT note;
		note.SetA( s.Text() );

		pitem->SetNote( &note );
	    }
	}

	return 0;
}

int
XferStores( DbOpen* pdb, StrA& s, Item* pitem, bool fOut, bool fUseFriendlyNames, AppendQuoteFlags flags )
{
	unsigned short ii;

	if( fOut )
	{
	    StrA sStores;

	    for( ii = 0; ii < pitem->m_stores.Count(); ii++ )
	    {
		if( sStores.Length() )
		    sStores.Append( ";" );

		if( fUseFriendlyNames )
		{
		    StrA tmp;

		    tmp.SetT( pdb->GetStoreNameById( pitem->m_stores[ii] ) );
		    FixStoreName( tmp );
		    sStores.Printf( "%s", tmp.Text() );
		}
		else
		{
		    sStores.Printf( "%d", pitem->m_stores[ii] );
		}
	    }

	    s.AppendQuote( sStores.Text(), flags );
	}
	else
	{
	    Store id;
	    char const* psz = s.Text();

	    Assert( pitem->m_stores.FPristine() );

	    while( *psz )
	    {
		if( fUseFriendlyNames )
		{
		    char const* pszWalk = strpbrk( psz, ";" );
		    StrA tmp;
		    StrT sStoreName;

		    if( pszWalk )
		    {
			tmp.Set( psz, pszWalk - psz );
			psz = pszWalk + 1;
		    }
		    else
		    {
			tmp.Set( psz );
			psz += tmp.Length();
		    }

		    sStoreName.SetA( tmp.Text() );
		    id = pdb->GetStoreIdByName( sStoreName.Text() );

		    if( id == storeUnfiledStore )
		    {
			fprintf( stderr, "Store '%s' does not exist.\n", tmp.Text() );

			if( IsDigit( *tmp.Text() ) )
			    fprintf( stderr, "Did you forget to use the -i flag?\n" );

			return -1;
		    }
		}
		else
		{
		    if( !IsDigit( *psz ) )
		    {
LInvalidStoresListString:
			fprintf( stderr,
				 "The Stores field must be a list of store id's, "
				 "separated by semicolons (;).\n"
				 "For example \"33;67;82\".\n" );
			return -1;
		    }

		    unsigned const nStore = atoi( psz );

		    if( int( nStore ) < 0 || nStore >= pdb->GetStoresMaximum() )
		    {
			fprintf( stderr, "Store id must be between 0 and %u.\n", pdb->GetStoresMaximum() - 1 );
			return -1;
		    }

		    id = Store( nStore );

		    if( !pdb->StoreExistsById( id ) )
		    {
			fprintf( stderr, "Store id %d does not exist.\n", id );
			//$ todo: (chrisant) should this return -1?
		    }

		    while( IsDigit( *psz ) )
			psz++;

		    if( *psz == ';' )
			psz++;
		    else if( *psz )
			goto LInvalidStoresListString;
		}

		// check for duplicate association

		for( ii = 0; ii < pitem->m_stores.Count(); ii++ )
		    if( pitem->m_stores[ii] == id )
			break;

		// associate item with the store

		if( ii >= pitem->m_stores.Count() )
		    pitem->m_stores.Insert( id );
	    }
	}

	return 0;
}

int
XferPerStoreInfo( DbOpen* pdb, StrA& s, Item* pitem, bool fOut, bool fUseFriendlyNames, AppendQuoteFlags flags )
{
	unsigned short ii;

	if( fOut )
	{
	    StrA sStores;
	    StrA sAisle;

	    for( ii = 0; ii < pitem->m_perstore.Count(); ii++ )
	    {
		PerStore const* p = &pitem->m_perstore[ii];

		if( sStores.Length() )
		    sStores.Append( ";" );

		if( fUseFriendlyNames )
		{
		    StrA tmp;

		    tmp.SetT( pdb->GetStoreNameById( p->m_idStore ) );
		    FixStoreName( tmp );
		    sStores.Printf( "%s", tmp.Text() );
		}
		else
		{
		    sStores.Printf( "%d", p->m_idStore );
		}

		sAisle.SetT( p->m_szAisle );

		if( strchr( sAisle.Text(), '/' ) ||
		    strchr( sAisle.Text(), ',' ) ||
		    strchr( sAisle.Text(), '=' ) ||
		    strchr( sAisle.Text(), '\"' ) )
		{
		    fprintf( stderr, "Aisles cannot contain / or , or = or \" characters.\n" );
		    return -1;
		}

		sStores.Printf( "=%s/%ls", sAisle.Text(), LocalizeNumber( p->m_lPrice, lnTrimRightZeros2, PRICE_PRECISION, 0 ) );
	    }

	    s.AppendQuote( sStores.Text(), flags );
	}
	else
	{
	    BYTE x;
	    StrT sField;
	    StrT tmp;

	    sField.SetA( s.Text() );

	    TCHAR const* psz = sField.Text();

	    Assert( pitem->m_perstore.FPristine() );

	    while( *psz )
	    {
		PerStore* p;

		if( fUseFriendlyNames )
		{
		    TCHAR const* pszWalk = _tcschr( psz, '=' );

		    if( pszWalk )
		    {
			tmp.Set( psz, pszWalk - psz );
			psz = pszWalk;
		    }
		    else
		    {
			tmp.Set( psz );
			psz += tmp.Length();
		    }

		    x = pdb->GetStoreIdByName( tmp.Text() );

		    if( x == storeUnfiledStore )
		    {
			fprintf( stderr, "Store '%ls' does not exist.\n", tmp.Text() );

			if( IsDigit( *tmp.Text() ) )
			    fprintf( stderr, "Did you forget to use the -i flag?\n" );

			return -1;
		    }
		}
		else
		{
		    if( !IsDigit( *psz ) )
			goto LInvalidPerStoreInfoString;

		    unsigned const nStore = _ttoi( psz );

		    if( int( nStore ) < 0 || nStore >= pdb->GetStoresMaximum() )
		    {
			fprintf( stderr, "Store id must be between 0 and %u.\n", pdb->GetStoresMaximum() - 1 );
			return -1;
		    }

		    x = BYTE( nStore );

		    if( !pdb->StoreExistsById( x ) )
		    {
			fprintf( stderr, "Store id %d does not exist.\n", x );
		    }

		    while( IsDigit( *psz ) )
			psz++;
		}

		// check for duplicates

		for( ii = 0; ii < pitem->m_perstore.Count(); ii++ )
		    if( pitem->m_perstore[ii].m_idStore == x )
		    {
			if( fUseFriendlyNames )
			    fprintf( stderr, "Duplicate per-store info for '%ls'.\n", pdb->GetStoreNameById( x ) );
			else
			    fprintf( stderr, "Duplicate per-store info for %d.\n", x );
			return -1;
		    }

		// add entry

		p = pitem->m_perstore.Insert( x );

		if( *psz == '=' )
		    psz++;
		else if( *psz )
		    goto LInvalidPerStoreInfoString;

		TCHAR const* pszSlash;
		pszSlash = _tcschr( psz, '/' );

		if( !pszSlash )
		    goto LInvalidPerStoreInfoString;

		tmp.Set( psz, pszSlash - psz );
		if( tmp.Length() > DimensionOf( p->m_szAisle ) - 1 )
		    tmp.SetLength( DimensionOf( p->m_szAisle ) - 1 );

		ZeroMemory( p->m_szAisle, sizeof( p->m_szAisle ) );
		CchCopy( tmp.Text(), RgC( p->m_szAisle ) );
		psz = pszSlash + 1;

		if( _tcschr( tmp.Text() , '/' ) ||
		    _tcschr( tmp.Text(), ',' ) ||
		    _tcschr( tmp.Text(), '=' ) ||
		    _tcschr( tmp.Text(), '\"' ) )
		{
		    fprintf( stderr, "Aisles cannot contain / or , or = or \" characters.\n" );
		    return -1;
		}

		long const l = ParseNumber( psz, PRICE_PRECISION, &psz, true );
		if( l == parseError )
		{
		    fprintf( stderr, "Invalid price.\n" );
		    return -1;
		}

		p->m_lPrice = l;

		if( *psz == ';' )
		    psz++;
		else if( *psz )
		{
LInvalidPerStoreInfoString:
		    if( fUseFriendlyNames )
		    {
			fprintf( stderr,
				 "The PerStoreInfo field must be a list of data, separated by semicolons (;).\n"
				 "The data must consist of the store name, an equals sign (=), the aisle, a\n"
				 "forward slash (/), and the price.\n"
				 "For example \"K-Mart=A/3.99;Seven Eleven=15/3.59\".\n" );
		    }
		    else
		    {
			fprintf( stderr,
				 "The PerStoreInfo field must be a list of data, separated by semicolons (;).\n"
				 "The data must consist of the store id, an equals sign (=), the aisle, a\n"
				 "forward slash (/), and the price.\n"
				 "For example \"33=A/3.99;67=15/3.59;82=0/3.79\".\n" );
		    }
		    return -1;
		}
	    }
	}

	return 0;
}

int
XferAutoOrder( DbOpen* pdb, StrA& s, Item* pitem, bool fOut, bool fUseFriendlyNames, AppendQuoteFlags flags )
{
	unsigned short ii;

	if( fOut )
	{
	    StrA sList;

	    for( ii = 0; ii < pitem->CountAutoOrders(); ii++ )
	    {
		AutoOrder const* p = pitem->GetAutoOrderByIndex( ii );

		if( !p->m_ulOrder )	// Skip entries that don't have any order information (for example the phantom store entry if it has never been filled in).
		    continue;

		if( sList.Length() )
		    sList.Append( ";" );

		if( fUseFriendlyNames )
		{
		    StrA tmp;

		    tmp.SetT( pdb->GetStoreNameById( p->m_idStore ) );
		    FixStoreName( tmp );
		    sList.Printf( "%s", tmp.Text() );
		}
		else
		{
		    sList.Printf( "%d", p->m_idStore );
		}

		sList.Printf( "=%u", p->m_ulOrder );
	    }

	    s.AppendQuote( sList.Text(), flags );
	}
	else
	{
	    BYTE x;
	    StrT sField;
	    StrT tmp;

	    sField.SetA( s.Text() );

	    TCHAR const* psz = sField.Text();

	    Assert( !pitem->CountAutoOrders() );

	    while( *psz )
	    {
		if( fUseFriendlyNames )
		{
		    TCHAR const* pszWalk = _tcschr( psz, '=' );

		    if( pszWalk )
		    {
			tmp.Set( psz, pszWalk - psz );
			psz = pszWalk;
		    }
		    else
		    {
			tmp.Set( psz );
			psz += tmp.Length();
		    }

		    x = pdb->GetStoreIdByName( tmp.Text() );

		    if( x == storeUnfiledStore )
		    {
			fprintf( stderr, "Store '%ls' does not exist.\n", tmp.Text() );

			if( IsDigit( *tmp.Text() ) )
			    fprintf( stderr, "Did you forget to use the -i flag?\n" );

			return -1;
		    }
		}
		else
		{
		    if( !IsDigit( *psz ) )
			goto LInvalidString;

		    unsigned const nStore = _ttoi( psz );

		    if( int( nStore ) < 0 || nStore >= pdb->GetStoresMaximum() )
		    {
			fprintf( stderr, "Store id must be between 0 and %u.\n", pdb->GetStoresMaximum() - 1 );
			return -1;
		    }

		    x = BYTE( nStore );

		    if( !pdb->StoreExistsById( x ) )
		    {
			fprintf( stderr, "Store id %d does not exist.\n", x );
		    }

		    while( IsDigit( *psz ) )
			psz++;
		}

		// check for duplicates

		for( ii = 1; ii < pitem->CountAutoOrders(); ii++ )
		    if( pitem->GetAutoOrderByIndex( ii )->m_idStore == x )
		    {
			if( fUseFriendlyNames )
			    fprintf( stderr, "Duplicate auto order info for '%ls'.\n", pdb->GetStoreNameById( x ) );
			else
			    fprintf( stderr, "Duplicate auto order info for %d.\n", x );
			return -1;
		    }

		// add entry

		if( *psz == '=' )
		    psz++;
		else if( *psz )
		    goto LInvalidString;

		unsigned ul;
		if( ParseUnsigned( &psz, &ul, TCHAR( 0 ) ) )
		    return -1;
		pitem->SetAutoOrder( ul, true, x );

		if( *psz == ';' )
		    psz++;
		else if( *psz )
		{
LInvalidString:
		    if( fUseFriendlyNames )
		    {
			fprintf( stderr,
				 "The AutoOrder field must be a list of data, separated by semicolons (;).\n"
				 "The data must consist of the store name, an equals sign (=), and the order\n"
				 "number.\n"
				 "For example \"K-Mart=52;Seven Eleven=17\".\n" );
		    }
		    else
		    {
			fprintf( stderr,
				 "The AutoOrder field must be a list of data, separated by semicolons (;).\n"
				 "The data must consist of the store id, an equals sign (=), and the order\n"
				 "number.\n"
				 "For example \"33=52;67=17\".\n" );
		    }
		    return -1;
		}
	    }
	}

	return 0;
}

struct SchemaEntry
{
	char const*		m_pszColumnName;
	XferColumnFn		m_pfnXfer;
};

#define SCHEMAENTRY( name ) { #name, Xfer##name },
static const SchemaEntry c_rgSchema[] =
{
	SCHEMAENTRY( Need )
	SCHEMAENTRY( Priority )
	SCHEMAENTRY( Description )
	SCHEMAENTRY( CustomText )
	SCHEMAENTRY( Quantity )
	SCHEMAENTRY( Units )
	SCHEMAENTRY( Price )
	SCHEMAENTRY( Aisle )
	SCHEMAENTRY( Date )
	SCHEMAENTRY( Category )
	SCHEMAENTRY( Stores )
	SCHEMAENTRY( PerStoreInfo )
	SCHEMAENTRY( EntryOrder )
	SCHEMAENTRY( Coupon )
	SCHEMAENTRY( Tax )
	SCHEMAENTRY( Tax2 )
	SCHEMAENTRY( AutoDelete )
	SCHEMAENTRY( Private )
	SCHEMAENTRY( Note )
	SCHEMAENTRY( Alarm )
	SCHEMAENTRY( AlarmMidi )
	SCHEMAENTRY( Icon )
	SCHEMAENTRY( AutoOrder )
	{ 0, 0 }
};

void
GetHeaderString( StrA& s )
{
	s.Clear();

	for( const SchemaEntry* pse = c_rgSchema; pse->m_pszColumnName; pse++ )
	{
	    if( s.Length() )
		s.Append( "," );
	    s.Append( pse->m_pszColumnName );
	}
}

int
XferItem( DbOpen* pdb, FILE* pfile, Item* pitem, bool fOut, int* pcRecord, int* pcLine, bool fUseFriendlyNames, AppendQuoteFlags flags = AQF_ESCAPEQUOTES )
{
	int n = 0;
	int const cNumFields = DimensionOf( c_rgSchema ) - 1;
	bool fEnd = false;
	StrA s;

	for( int c = 0; !fEnd && c < cNumFields; ++c )
	{
	    s.Clear();

	    if( !fOut )
	    {
		char ch;
		bool fQuote = false;

		// read field content

		while( !fEnd )
		{
		    ch = char( fgetc( pfile ) );

		    if( feof( pfile ) )
		    {
			if( !c && !fQuote && !s.Length() )
			{
			    // haven't read anything yet
			    return 0;
			}

LEof:
			if( fQuote || c + 1 < cNumFields )
			{
			    fprintf( stderr, "Unexpected end of file (line %d, record %d, field '%s').\n", *pcLine, *pcRecord, c_rgSchema[c].m_pszColumnName );
			    return -1;
			}

			break;
		    }

		    if( ch == '\"' )
		    {
			if( !fQuote && !s.Length() )
			{
			    fQuote = true;
			    continue;
			}
			else if( fQuote )
			{
			    fQuote = false;

			    ch = char( fgetc( pfile ) );
			    if( feof( pfile ) )
				goto LEof;

			    switch( ch )
			    {
			    case ',':
			    case '\n':
				// nothing to do
				break;
			    case '\"':
				// two double quotes in a row are interpreted
				// as one double quote and do not leave quote
				// mode.
				fQuote = true;
				s.Append( &ch, 1 );
				continue;
			    default:
				fprintf( stderr, "Unexpected character (line %d, record %d, field '%s').\n", *pcLine, *pcRecord, c_rgSchema[c].m_pszColumnName );
				return -1;
			    }
			}
		    }

		    if( fQuote )
		    {
			if( ch == '\n' )
			    ( *pcLine )++;
		    }
		    else
		    {
			if( ch == '\n' )
			{
#if 0
			    if( c + 1 < cNumFields )
			    {
				fprintf( stderr, "Unexpected end of line (line %d, record %d, field '%s').\n", *pcLine, *pcRecord, c_rgSchema[c].m_pszColumnName );
				return -1;
			    }
#endif

			    ( *pcLine )++;
			    fEnd = true;
			    break;
			}
			if( ch == ',' )
			{
			    if( c + 1 < cNumFields )
				break;

			    fprintf( stderr, "Too many fields (line %d, record %d).\n", *pcLine, *pcRecord );
			    return -1;
			}
		    }

		    s.Append( &ch, 1 );
		}
	    }

	    // xfer field content to/from Str/ItemPtr

	    if( !fOut )
		s.ConvertToNativeLineEndings();

	    n = c_rgSchema[c].m_pfnXfer( pdb, s, pitem, fOut, fUseFriendlyNames, flags );
	    if( n < 0 )
	    {
		if( fOut )
		    fprintf( stderr, "Error writing field '%s'.\n", c_rgSchema[c].m_pszColumnName );
		else
		    fprintf( stderr, "Error reading field '%s' (line %d, record %d).\n", c_rgSchema[c].m_pszColumnName, *pcLine, *pcRecord );
		return n;
	    }

	    // output field content

	    if( fOut )
	    {
		//DumpBytes( stdout, "Before", s.Text(), s.Length() );
		s.ConvertToCrtLineEndings();
		//DumpBytes( stdout, "After", s.Text(), s.Length() );
		fprintf( pfile, c ? ",%s" : "%s", s.Text() );
	    }
	}

	if( fOut )
	{
	    fputs( "\n", pfile );
	}
	else
	{
	    ( *pcRecord )++;
	}

	return n;
}

/*
 * OutCSV.
 */

class OutCSV : public OutputInterface
{
    public:
			OutCSV( bool fSpreadsheetCompatibility );
			~OutCSV();

	DeclareOutputInterfaceMembers( IMPL );

    private:
	void		Close();

    private:
	FILE*		m_pfileItems;
	FILE*		m_pfileMetadata;
	int		m_cCategories;
	int		m_cStores;
	int		m_cItems;

	StrT		m_sNameItems;
	StrT		m_sNameMetadata;
	bool		m_fForceOverwrite;
	bool		m_fUseFriendlyNames;
	bool		m_fFixedStoreNames;

	bool const	m_fSpreadsheetCompatibility;
};

OutCSV::OutCSV( bool fSpreadsheetCompatibility )
	: m_fSpreadsheetCompatibility( fSpreadsheetCompatibility )
{
	m_pfileItems = 0;
	m_pfileMetadata = 0;

	m_cCategories = 0;
	m_cStores = 0;
	m_cItems = 0;

	m_fUseFriendlyNames = true;
	m_fFixedStoreNames = false;
}

OutCSV::~OutCSV()
{
	Close();
}

void
OutCSV::Close()
{
	if( m_pfileItems )
	    fclose( m_pfileItems );
	m_pfileItems = 0;

	if( m_pfileMetadata )
	    fclose( m_pfileMetadata );
	m_pfileMetadata = 0;
}

int
OutCSV::SetFileName( TCHAR const* const pszFile, bool const fForceOverwrite )
{
	TCHAR const* pszExt;

	m_sNameItems.Set( pszFile );
	pszExt = GetExtension( m_sNameItems.Text() );
	if( LStrCmpI( pszExt, c_szExtItems ) )
	    m_sNameItems.Append( c_szExtItems );

	m_sNameMetadata.Set( m_sNameItems );
	pszExt = GetExtension( m_sNameMetadata.Text() );
	if( LStrCmpI( pszExt, c_szExtMetadata ) )
	{
	    m_sNameMetadata.SetEnd( pszExt );
	    m_sNameMetadata.Append( c_szExtMetadata );
	}

	m_fForceOverwrite = fForceOverwrite;
	return 0;
}

int
OutCSV::SetFriendlyNames( bool const fUseFriendlyNames )
{
	m_fUseFriendlyNames = fUseFriendlyNames;
	return 0;
}

int
OutCSV::Begin( DbOpen const* const pdb )
{
	if( m_sNameItems.Length() )
	{
	    StrA nameItems;
	    StrA nameMetadata;
	    nameItems.SetT( m_sNameItems );
	    nameMetadata.SetT( m_sNameMetadata );

	    bool const fExistsItems = FileExists( m_sNameItems.Text() );
	    bool const fExistsMetadata = FileExists( m_sNameMetadata.Text() );

	    if( fExistsItems || fExistsMetadata )
	    {
		if( m_fForceOverwrite )
		{
		    __DeleteFile( m_sNameItems.Text() );
		    __DeleteFile( m_sNameMetadata.Text() );
		}
		else
		{
		    fprintf( stderr, "Output file '%s' already exists.\n", fExistsItems ? nameItems.Text() : nameMetadata.Text() );
		    return -1;
		}
	    }

	    if( fopen_s( &m_pfileItems, nameItems.Text(), "w+" ) ||
		fopen_s( &m_pfileMetadata, nameMetadata.Text(), "w+" ) ||
		!m_pfileItems || !m_pfileMetadata )
	    {
		fprintf( stderr, "Unable to open output file.\n" );
		return -1;
	    }

	    StrA s;
	    StrA tmp;

	    tmp.SetT( pdb->Name() );
	    s.AppendQuote( tmp.Text(), AQF_QUOTESPACES );

	    fprintf( m_pfileMetadata, "DatabaseName: %s\n", s.Text() );

	    if( !pdb->HasPrefix() )
		fprintf( m_pfileMetadata, "OmitPrefix\n" );

	    fprintf( m_pfileMetadata, "\n" );
	}

	return 0;
}

int
OutCSV::OutputCategory( DbOpen const* const pdb, TCHAR const* const pszCategory, WORD id )
{
	StrA s;
	StrA tmp;

	if( pszCategory )
	{
	    WORD const icon = pdb->GetCategoryIcon( id );

	    if( !m_cCategories )
		s.Append( "Categories:\n" );

	    tmp.SetT( pszCategory );

	    s.Append( "\t" );
	    s.AppendQuote( tmp.Text() );
	    if( icon )
		s.Printf( ", icon=%u", icon );
	    s.Append( "\n" );

	    m_cCategories++;
	}
	else
	{
	    s.Append( "\n" );
	}

	fputs( s.Text(), m_pfileMetadata );

	return 0;
}

static char const*
SzYesNo( bool const f )
{
	return f ? "Yes" : "No";
}

static char const*
GetViewStr( WORD const view )
{
	switch( view )
	{
	case VIEW_ALL:
	    return "All";
	case VIEW_NEED:
	case VIEW_ALLNOTPURCHASED:
	    return "Needed";
	case VIEW_NOTNEEDED:
	    return "NotNeeded";
	default:
	    Panic1( "unexpected view %d", view );
	}
	return "";
}

struct ColumnInfo
{
	char const*		szName;
	ColumnIdentifier	col;
	bool			fWidths;
	bool			fDate;
};

static ColumnInfo const c_rgCols[] =
{
	{ "ColumnPrio",		COLID_PRIO, false, false },
	{ "ColumnQuantity",	COLID_QUANTITY, true, false },
	{ "ColumnUnits",	COLID_UNITS, true, false },
	{ "ColumnAisle",	COLID_AISLE, true, false },
	{ "ColumnCategory",	COLID_CATEGORY, true, false },
	{ "ColumnPrice",	COLID_PRICE, true, false },
	{ "ColumnTaxable",	COLID_TAXABLE, false, false },
	{ "ColumnDate",		COLID_DATE, true, true },
	{ "ColumnCoupon",	COLID_COUPON, false, false },
	{ "ColumnCustomText",	COLID_CUSTOMTEXT, true, false },
	{ "ColumnUnitPrice",	COLID_UNITPRICE, true, false },
	{ 0 }
};

static char const* const c_rgszWidth[] = { "Default", "Narrow", "Medium", "Wide" };

static char const*
SzWidth( int width, bool const fDate )
{
	Assert( width >= 0 && width < DimensionOf( c_rgszWidth ) );

	if( width && fDate )
	    width = 3;

	return c_rgszWidth[width];
}

static char const*
SzFontName( FontId const font )
{
	switch( font )
	{
	case FID_NORMAL:	return "Standard";
	case FID_BOLD:		return "Bold";
	case FID_LARGEBOLD:	return "LargeBold";
	case FID_NARROW:	return "Narrow";
	case FID_TINY:		return "Tiny";
	}

	return 0;
}

struct ExportInfo
{
	char const*	szName;
	int		flag;
};

#define EXPINFO( x ) { #x, export##x },
static ExportInfo const c_rgExp[] =
{
	EXPINFO( Quantity )
	EXPINFO( Price )
	EXPINFO( Note )
	EXPINFO( Total )
	EXPINFO( Units )
	EXPINFO( Aisle )
	{ "Dashes", exportDash },
	{ 0 }
};

static char const* const c_rgSortField[] =
{
	"None",
	"Priority",
	"Date",
	"Aisle",
	"Category",
	"Description",
	"Quantity",
	"Units",
	"Price",
	"EntryOrder",
	"CustomText",
	"Alarm",
	"Icon"
};
CASSERT( DimensionOf( c_rgSortField ) == SORT_MAX );

static void
OutputSortOrder( FILE* const pfile, DbSortOrder const* const pso, bool const fCanLearn, bool const fAll )
{
	char const* const pszPrefix = fAll ? "All" : "Need";
	char const* const pszYesNo = SzYesNo( pso->m_fCheckedLast );

	fprintf( pfile, "\tSort_%s_CompletedAtBottom=%s\n", pszPrefix, pszYesNo );

	for( unsigned short ii = 0; ii < NUM_SORT_FIELDS; ii++ )
	{
	    char const* const pszField = c_rgSortField[pso->m_rgiSort[ii]];
	    char const* const pszOrder = pso->m_rgfDescend[ii] ? "Descend" : "Ascend";

	    fprintf( pfile, "\tSort_%s_Field%d=%s,%s\n", pszPrefix, ii + 1, pszField, pszOrder );
	}

	fprintf( pfile, "\tSort_%s_AutoOrder=%s\n", pszPrefix, SzYesNo( fCanLearn ) );
}

int
OutCSV::OutputPreferences( DbOpen const* const pdb )
{
	StrA s;
	StrA tmp;

	fprintf( m_pfileMetadata, "Preferences:\n" );

	fprintf( m_pfileMetadata, "\tCurrentStore=" );
	switch( pdb->GetStore() )
	{
	case storeAllStores:
	    fprintf( m_pfileMetadata, "All Stores\n" );
	    break;
	case storeUnfiledStore:
	    fprintf( m_pfileMetadata, "Unfiled\n" );
	    break;
	default:
	    if( m_fUseFriendlyNames )
	    {
		tmp.SetT( pdb->GetStoreNameById( pdb->GetStore() ) );

		s.Clear();
		s.AppendQuote( tmp.Text(), AQF_ESCAPEQUOTES|AQF_QUOTESPACES );
		fprintf( m_pfileMetadata, "%s\n", s.Text() );
	    }
	    else
	    {
		fprintf( m_pfileMetadata, "%u\n", pdb->GetStore() );
	    }
	    break;
	}
	tmp.SetT( pdb->GetCategoryNameById( pdb->GetCategory() ) );
	fprintf( m_pfileMetadata, "\tCurrentCategory=%s\n", tmp.Text() );
	fprintf( m_pfileMetadata, "\tCurrentView=%s\n", GetViewStr( pdb->GetView() ) );

	fprintf( m_pfileMetadata, "\t%s=%s\n", "FlagPerStorePrices", SzYesNo( pdb->GetPerStorePrices() ) );
	fprintf( m_pfileMetadata, "\t%s=%s\n", "StampTodayOnNewItems", SzYesNo( pdb->GetRecordCreationDate() ) );
	fprintf( m_pfileMetadata, "\t%s=%s\n", "StampTodayWhenComplete", SzYesNo( pdb->GetRecordCompletionDate() ) );
	fprintf( m_pfileMetadata, "\t%s=%s\n", "AutoResetQuantities", SzYesNo( pdb->GetAutoQuantity() ) );
	fprintf( m_pfileMetadata, "\t%s=%s\n", "NewItemsAreAutoDelete", SzYesNo( pdb->GetAutoDelete() ) );
	fprintf( m_pfileMetadata, "\t%s=%s\n", "NewItemsAreTaxable", SzYesNo( pdb->GetTaxable() ) );
	fprintf( m_pfileMetadata, "\t%s=%s\n", "TwoTaxes", SzYesNo( pdb->GetTwoTaxes() ) );
	fprintf( m_pfileMetadata, "\t%s=%s\n", "RoundPricesToWholeNumbers", SzYesNo( pdb->GetRoundToWholeNumbers() ) );
	fprintf( m_pfileMetadata, "\t%s=%s\n", "SeparateSortOrders", SzYesNo( !pdb->GetSortOneOrder() ) );

	fprintf( m_pfileMetadata, "\tHideCompletedItems=%s\n", SzYesNo( pdb->GetHideChecked() ) );
	fprintf( m_pfileMetadata, "\tShowOnlyCoupons=%s\n", SzYesNo( pdb->GetCouponsOnly() ) );
	fprintf( m_pfileMetadata, "\tShowLinesPerItem=" );
	if( pdb->GetMaxLinesPerItem() <= 0 )
	    fprintf( m_pfileMetadata, "Auto\n" );
	else
	    fprintf( m_pfileMetadata, "%u\n", pdb->GetMaxLinesPerItem() );

	for( ColumnInfo const* pcol = c_rgCols; pcol->szName; pcol++ )
	{
	    int const cx = pdb->GetColumnWidth( pcol->col );
	    bool const fVisible = !!cx;

	    fprintf( m_pfileMetadata, "\t%s=%s", pcol->szName, SzYesNo( fVisible ) );

	    if( fVisible && pcol->fWidths )
	    {
		if( !pcol->fDate && cx >= 5 )
		    fprintf( m_pfileMetadata, ",%u", cx );
		else
		    fprintf( m_pfileMetadata, ",%s", SzWidth( cx, pcol->fDate ) );
	    }

	    fprintf( m_pfileMetadata, "\n" );
	}

	CASSERT( NUM_SORT_ORDERS == 2 );
	for( int ii = 0; ii < NUM_SORT_ORDERS; ii++ )
	    OutputSortOrder( m_pfileMetadata, pdb->GetSortOrderInfo( ii ), pdb->CanLearnOrder( ii ), ii > 0 );

	s.SetT( pdb->GetCurrency() );
	fprintf( m_pfileMetadata, "\tCurrency=%s\n", s.Text() );
	fprintf( m_pfileMetadata, "\tTaxRate=%ls\n", LocalizeNumber( pdb->GetTaxRate(), lnTrimRightZeros|lnShowZero, 100, 0 ) );
	fprintf( m_pfileMetadata, "\tTaxRate2=%ls\n", LocalizeNumber( pdb->GetTaxRate2(), lnTrimRightZeros|lnShowZero, 100, 0 ) );

	fprintf( m_pfileMetadata, "\tDefaultPriority=%u\n", pdb->GetDefaultPriority() + 1 );

	s.Clear();
	fprintf( m_pfileMetadata, "\tExportFlags=" );
	for( ExportInfo const* pexp = c_rgExp; pexp->szName; pexp++ )
	    if( pdb->GetExportFlags() & pexp->flag )
	    {
		if( s.Length() )
		    s.Append( "," );
		s.Append( pexp->szName );
	    }
	fprintf( m_pfileMetadata, "%s\n", s.Text() );

	COLORREF const rgbRowColor = pdb->GetRowBackgroundColor();
	fprintf( m_pfileMetadata, "\tRowColorRGB=%u,%u,%u\n", GetRValue( rgbRowColor ), GetGValue( rgbRowColor ), GetBValue( rgbRowColor ) );

	WORD const icon = pdb->GetIcon();
	if( icon )
	    fprintf( m_pfileMetadata, "\tIcon=%u\n", icon );

	fprintf( m_pfileMetadata, "\tFont=%s\n", SzFontName( pdb->GetFont() ) );

	fputs( "\n", m_pfileMetadata );

	return 0;
}

int
OutCSV::OutputStore( DbOpen const* const pdb, TCHAR const* const pszStore, Store const id )
{
	SUPPRESS_UNUSED_VAR( pdb );

	StrA s;
	StrA tmp;
	bool fFixed = false;

	if( pszStore )
	{
	    if( !m_cStores )
		s.Append( "Stores:\n" );

	    s.Append( "\t" );
	    if( m_fUseFriendlyNames )
	    {
		tmp.SetT( pszStore );

		if( FixStoreName( tmp ) )
		{
		    if( !m_fFixedStoreNames )
			fprintf( stderr,
				 "Store names cannot contain = or ; unless the -i flag is used.\n"
				 "The following store names have been changed:\n" );

		    if( !fFixed )
			fprintf( stderr, "\"%s\" -> \"%s\"\n", pszStore, tmp.Text() );

		    m_fFixedStoreNames = true;
		    fFixed = true;
		}

		s.AppendQuote( tmp.Text() );
	    }
	    else
	    {
		StrA tmp;
		tmp.SetT( pszStore );

		s.Printf( "%u,", id );
		s.AppendQuote( tmp.Text() );
	    }
	    s.Append( "\n" );

	    m_cStores++;
	}
	else
	{
	    s.Append( "\n" );
	}

	fputs( s.Text(), m_pfileMetadata );

	return 0;
}

int
OutCSV::OutputItem( DbOpen const* const pdb, Item const* const pitem )
{
	int n;

	if( pitem )
	{
	    // Header row.

	    if( !m_cItems )
	    {
		StrA s;

		GetHeaderString( s );
		fprintf( m_pfileItems, "%s\n", s.Text() );
	    }

	    // The item.

	    AppendQuoteFlags const aqf = m_fSpreadsheetCompatibility ? AQF_ESCAPEQUOTES|AQF_SPREADSHEETCOMPATIBILITY : AQF_ESCAPEQUOTES;
	    n = XferItem( const_cast<DbOpen*>( pdb ), m_pfileItems, const_cast<Item*>( pitem ), true, 0, 0, m_fUseFriendlyNames, aqf );
	    m_cItems++;

	    if( n < 0 )
		return n;
	}

	return 0;
}

int
OutCSV::Finished( DbOpen const* const pdb )
{
	SUPPRESS_UNUSED_VAR( pdb );

	Close();
	return 0;
}

OutputInterface*
NewOutputCSV( bool fSpreadsheetCompatibility )
{
	return new OutCSV( fSpreadsheetCompatibility );
}

/*
 * DoImportCSV.
 */

static int
ParseBoolean( char const* psz, bool* pf )
{
	if( _stricmp( psz, "yes" ) == 0 ||
	    _stricmp( psz, "true" ) == 0 ||
	    _stricmp( psz, "x" ) == 0 ||
	    _stricmp( psz, "1" ) == 0 )
	{
	    *pf = true;
	    return 0;
	}

	if( _stricmp( psz, "no" ) == 0 ||
	    _stricmp( psz, "false" ) == 0 ||
	    !*psz ||
	    _stricmp( psz, "0" ) == 0 )
	{
	    *pf = false;
	    return 0;
	}

	fprintf( stderr, "Expected boolean value (yes/no, true/false, 1/0).\n" );
	return -1;
}

typedef int ( *FnParsePref )( DbOpen* pdb, char const* pszValue );
struct PrefParseInfo
{
	char const*	szName;
	FnParsePref	pfnParse;
};

int
ParseCurrentStore( DbOpen* pdb, char const* psz )
{
	SUPPRESS_UNUSED_VAR( pdb );

	s_sCurrentStore.SetA( psz );
	return 0;
}

int
ParseCurrentCategory( DbOpen* pdb, char const* psz )
{
	WORD category;

	if( _stricmp( psz, "all" ) == 0 || _stricmp( psz, "all categories" ) == 0 )
	{
	    category = catAllCategories;
	}
	else
	{
	    StrT s;
	    s.SetA( psz );

	    category = pdb->GetCategoryIdByName( s.Text() );

	    if( catUnfiledCategory == category )
	    {
		fprintf( stderr, "Category '%s' not found.\n", psz );
		return -1;
	    }
	}

	pdb->SetCategory( category );
	return 0;
}

int
ParseCurrentView( DbOpen* pdb, char const* psz )
{
	if( _stricmp( psz, "all" ) == 0 )
	{
	    pdb->SetView( VIEW_ALL );
	    return 0;
	}
	if( _stricmp( psz, "need" ) == 0 || _stricmp( psz, "needed" ) == 0 )
	{
	    pdb->SetView( VIEW_NEED );
	    return 0;
	}
	if( _stricmp( psz, "notneed" ) == 0 || _stricmp( psz, "notneeded" ) == 0 )
	{
	    pdb->SetView( VIEW_NOTNEEDED );
	    return 0;
	}

	fprintf( stderr, "View must be All, Needed, or NotNeeded.\n" );
	return -1;
}

int
ParseHideCompletedItems( DbOpen* pdb, char const* psz )
{
	bool f;
	int const n = ParseBoolean( psz, &f );

	if( n < 0 )
	    return n;

	pdb->SetHideChecked( f );
	return 0;
}

int
ParseShowOnlyCoupons( DbOpen* pdb, char const* psz )
{
	bool f;
	int const n = ParseBoolean( psz, &f );

	if( n < 0 )
	    return n;

	pdb->SetCouponsOnly( f );
	return 0;
}

int
ParseShowLinesPerItem( DbOpen* pdb, char const* psz )
{
	if( _stricmp( psz, "auto" ) == 0 )
	{
	    pdb->SetMaxLinesPerItem( 0 );
	    return 0;
	}

	int const c = atoi( psz );

	if( c < 1 || c > 3 )
	{
	    fprintf( stderr, "ShowLinesPerItem must be Auto, 1, 2, or 3.\n" );
	    return -1;
	}

	pdb->SetMaxLinesPerItem( c );
	return 0;
}

int
ParseCurrency( DbOpen* pdb, char const* psz )
{
	char const* pszCurrency = psz;

	while( *psz && !IsDigit( *psz ) || *psz == ',' || *psz == '.' )
	    ++psz;

	if( strlen( pszCurrency ) < 1 || *psz )
	{
	    fprintf( stderr, "The currency symbol must not contain digits or the decimal separator.\n" );
	    return -1;
	}

	StrT s;
	s.SetA( pszCurrency );

	pdb->SetCurrency( s.Text() );
	return 0;
}

int
ParseTaxRateEx( DbOpen* pdb, char const* psz, bool fTwo )
{
	StrT s;
	s.SetA( psz );

	long const l = ParseNumber( s.Text(), WORD( -100 ), 0, true );

	if( parseError == l || l > 10000 )
	{
	    fprintf( stderr, "Tax rate must be between 0.00 and 100.00.\n" );
	    return -1;
	}

	if( fTwo )
	    pdb->SetTaxRate2( WORD( l ) );
	else
	    pdb->SetTaxRate( WORD( l ) );
	return 0;
}

int
ParseTaxRate( DbOpen* pdb, char const* psz )
{
	return ParseTaxRateEx( pdb, psz, false );
}

int
ParseTaxRate2( DbOpen* pdb, char const* psz )
{
	return ParseTaxRateEx( pdb, psz, true );
}

int
ParseDefaultPriority( DbOpen* pdb, char const* psz )
{
	int const prio = atoi( psz );

	if( prio < 1 || prio > 5 )
	{
	    fprintf( stderr, "Default priority must be 1, 2, 3, 4, or 5.\n" );
	    return -1;
	}

	pdb->SetDefaultPriority( BYTE( prio - 1 ) );
	return 0;
}

int
ParseExportFlags( DbOpen* pdb, char const* psz )
{
	int n = 0;
	char const* pszComma;
	WORD wExportFlags = 0;
	StrA s;

	while( psz && *psz )
	{
	    pszComma = strchr( psz, ',' );
	    if( pszComma )
	    {
		s.Set( psz, pszComma - psz );
		pszComma++;
	    }
	    else
		s.Set( psz );

	    n = -1;
	    for( ExportInfo const* pexp = c_rgExp; pexp->szName; pexp++ )
	    {
		if( _stricmp( pexp->szName, s.Text() ) == 0 )
		{
		    wExportFlags |= pexp->flag;
		    n = 0;
		    break;
		}
	    }

	    if( n < 0 )
	    {
		fprintf( stderr, "Unexpected '%s' in ExportFlags.\n", s.Text() );
		return n;
	    }

	    psz = pszComma;
	}

	pdb->SetExportFlags( wExportFlags );
	return 0;
}

int
ParseRowColor( DbOpen* pdb, char const* psz )
{
	unsigned r, g, b;
	int n;

	n = ParseUnsigned( &psz, &r );
	if( n < 0 )
	    return n;

	n = ParseUnsigned( &psz, &g );
	if( n < 0 )
	    return n;

	n = ParseUnsigned( &psz, &b );
	if( n < 0 )
	    return n;

	pdb->SetRowBackgroundColor( RGB( r, g, b ) );
	return 0;
}

int
ParseIcon( DbOpen* pdb, char const* psz )
{
	int const icon = atoi( psz );

	if( icon < 0 || icon > 0xffff )
	{
	    fprintf( stderr, "Icon index must be 0 through 65535.\n" );
	    return -1;
	}

	pdb->SetIcon( WORD( icon ) );
	return 0;
}

int
ParseFontName( DbOpen* pdb, char const* psz )
{
	FontId font = FID_MAX;

	if( _stricmp( psz, "std" ) == 0 || _stricmp( psz, "standard" ) == 0 )
	    font = FID_NORMAL;
	else if( _stricmp( psz, "bold" ) == 0 )
	    font = FID_BOLD;
	else if( _stricmp( psz, "large" ) == 0 || _stricmp( psz, "largebold" ) == 0 )
	    font = FID_LARGEBOLD;
	else if( _stricmp( psz, "nrw" ) == 0 || _stricmp( psz, "narrow" ) == 0 )
	    font = FID_NORMAL;
	else if( _stricmp( psz, "tiny" ) == 0 )
	    font = FID_TINY;

	if( font < FID_MAX )
	{
	    pdb->SetFont( font );
	    return 0;
	}

	fprintf( stderr, "Font must be Standard, Bold, LargeBold, Narrow, or Tiny.\n" );
	return -1;
}

int
ParseFlagPerStorePrices( DbOpen* pdb, char const* psz )
{
	bool f;
	int const n = ParseBoolean( psz, &f );

	if( n < 0 )
	    return n;

	pdb->SetPerStorePrices( f );
	return 0;
}

int
ParseStampTodayOnNewItems( DbOpen* pdb, char const* psz )
{
	bool f;
	int const n = ParseBoolean( psz, &f );

	if( n < 0 )
	    return n;

	pdb->SetRecordCreationDate( f );
	return 0;
}

int
ParseStampTodayWhenComplete( DbOpen* pdb, char const* psz )
{
	bool f;
	int const n = ParseBoolean( psz, &f );

	if( n < 0 )
	    return n;

	pdb->SetRecordCompletionDate( f );
	return 0;
}

int
ParseAutoResetQuantities( DbOpen* pdb, char const* psz )
{
	bool f;
	int const n = ParseBoolean( psz, &f );

	if( n < 0 )
	    return n;

	pdb->SetAutoQuantity( f );
	return 0;
}

int
ParseNewItemsAreAutoDelete( DbOpen* pdb, char const* psz )
{
	bool f;
	int const n = ParseBoolean( psz, &f );

	if( n < 0 )
	    return n;

	pdb->SetAutoDelete( f );
	return 0;
}

int
ParseNewItemsAreTaxable( DbOpen* pdb, char const* psz )
{
	bool f;
	int const n = ParseBoolean( psz, &f );

	if( n < 0 )
	    return n;

	pdb->SetTaxable( f );
	return 0;
}

int
ParseTwoTaxes( DbOpen* pdb, char const* psz )
{
	bool f;
	int const n = ParseBoolean( psz, &f );

	if( n < 0 )
	    return n;

	pdb->SetTwoTaxes( f );
	return 0;
}

int
ParseRoundPricesToWholeNumbers( DbOpen* pdb, char const* psz )
{
	bool f;
	int const n = ParseBoolean( psz, &f );

	if( n < 0 )
	    return n;

	pdb->SetRoundToWholeNumbers( f );
	return 0;
}

int
ParseSeparateSortOrders( DbOpen* pdb, char const* psz )
{
	bool f;
	int const n = ParseBoolean( psz, &f );

	if( n < 0 )
	    return n;

	pdb->SetSortOneOrder( !f );
	return 0;
}

int
ParseSortCompletedAtBottomEx( DbOpen* pdb, char const* psz, bool fAll )
{
	bool f;
	int const n = ParseBoolean( psz, &f );

	if( n < 0 )
	    return n;

	DbSortOrder sortOrder = *pdb->GetSortOrderInfo( !!fAll );
	sortOrder.m_fCheckedLast = f;
	pdb->SetSortOrderInfo( !!fAll, &sortOrder );
	return 0;
}

int
ParseSortCompletedAtBottomNeed( DbOpen* pdb, char const* psz )
{
	return ParseSortCompletedAtBottomEx( pdb, psz, false );
}

int
ParseSortCompletedAtBottomAll( DbOpen* pdb, char const* psz )
{
	return ParseSortCompletedAtBottomEx( pdb, psz, true );
}

int
ParseSortFieldEx( DbOpen* pdb, char const* psz, bool fAll, int iField )
{
	int n = 0;
	char const* pszComma;
	StrA s;

	Assert( iField >= 0 && iField < NUM_SORT_FIELDS );

	pszComma = strchr( psz, ',' );
	if( pszComma )
	{
	    s.Set( psz, pszComma - psz );
	    pszComma++;
	}
	else
	    s.Set( psz );

	if( s.Length() )
	{
	    n = -1;

	    for( WORD sort = 0; sort < DimensionOf( c_rgSortField ); sort++ )
	    {
		if( _stricmp( c_rgSortField[sort], s.Text() ) == 0 )
		{
		    DbSortOrder sortOrder = *pdb->GetSortOrderInfo( !!fAll );
		    sortOrder.m_rgiSort[iField] = sort;
		    pdb->SetSortOrderInfo( !!fAll, &sortOrder );
		    n = 0;
		    break;
		}
	    }

	    if( n < 0 )
	    {
		fprintf( stderr, "Unexpected '%s'.\n", s.Text() );
		return n;
	    }
	}

	if( pszComma && *pszComma )
	{
	    DbSortOrder sortOrder = *pdb->GetSortOrderInfo( !!fAll );

	    if( _stricmp( pszComma, "ascend" ) == 0 )
		sortOrder.m_rgfDescend[iField] = false;
	    else if( _stricmp( pszComma, "descend" ) == 0 )
		sortOrder.m_rgfDescend[iField] = true;
	    else
	    {
		fprintf( stderr, "Unexpected '%s' (must be Ascend or Descend).\n", pszComma );
		return -1;
	    }

	    pdb->SetSortOrderInfo( !!fAll, &sortOrder );
	}

	return 0;
}

int
ParseSortField1Need( DbOpen* pdb, char const* psz )
{
	return ParseSortFieldEx( pdb, psz, false, 0 );
}

int
ParseSortField2Need( DbOpen* pdb, char const* psz )
{
	return ParseSortFieldEx( pdb, psz, false, 1 );
}

int
ParseSortField3Need( DbOpen* pdb, char const* psz )
{
	return ParseSortFieldEx( pdb, psz, false, 2 );
}

int
ParseSortField1All( DbOpen* pdb, char const* psz )
{
	return ParseSortFieldEx( pdb, psz, true, 0 );
}

int
ParseSortField2All( DbOpen* pdb, char const* psz )
{
	return ParseSortFieldEx( pdb, psz, true, 1 );
}

int
ParseSortField3All( DbOpen* pdb, char const* psz )
{
	return ParseSortFieldEx( pdb, psz, true, 2 );
}

int
ParseSortAutoOrderEx( DbOpen* pdb, char const* psz, bool fAll )
{
	bool f;
	int const n = ParseBoolean( psz, &f );

	if( n < 0 )
	    return n;

	pdb->SetCanLearnOrder( !!fAll, f );
	return 0;
}

int
ParseSortAutoOrderNeed( DbOpen* pdb, char const* psz )
{
	return ParseSortAutoOrderEx( pdb, psz, false );
}

int
ParseSortAutoOrderAll( DbOpen* pdb, char const* psz )
{
	return ParseSortAutoOrderEx( pdb, psz, true );
}

static PrefParseInfo const c_rgPrefs[] =
{
	{ "CurrentStore",			ParseCurrentStore },
	{ "CurrentCategory",			ParseCurrentCategory },
	{ "CurrentView",			ParseCurrentView },
	{ "HideCompletedItems",			ParseHideCompletedItems },
	{ "ShowOnlyCoupons",			ParseShowOnlyCoupons },
	{ "ShowLinesPerItem",			ParseShowLinesPerItem },
	{ "Sort_Need_CompletedAtBottom",	ParseSortCompletedAtBottomNeed },
	{ "Sort_Need_Field1",			ParseSortField1Need },
	{ "Sort_Need_Field2",			ParseSortField2Need },
	{ "Sort_Need_Field3",			ParseSortField3Need },
	{ "Sort_Need_AutoOrder",		ParseSortAutoOrderNeed },
	{ "Sort_All_CompletedAtBottom",		ParseSortCompletedAtBottomAll },
	{ "Sort_All_Field1",			ParseSortField1All },
	{ "Sort_All_Field2",			ParseSortField2All },
	{ "Sort_All_Field3",			ParseSortField3All },
	{ "Sort_All_AutoOrder",			ParseSortAutoOrderAll },
	{ "Currency",				ParseCurrency },
	{ "TaxRate",				ParseTaxRate },
	{ "TaxRate2",				ParseTaxRate2 },
	{ "DefaultPriority",			ParseDefaultPriority },
	{ "ExportFlags",			ParseExportFlags },
	{ "RowColorRGB",			ParseRowColor },
	{ "Icon",				ParseIcon },
	{ "Font",				ParseFontName },
	{ "FlagPerStorePrices",			ParseFlagPerStorePrices },
	{ "StampTodayOnNewItems",		ParseStampTodayOnNewItems },
	{ "StampTodayWhenComplete",		ParseStampTodayWhenComplete },
	{ "AutoResetQuantities",		ParseAutoResetQuantities },
	{ "NewItemsAreAutoDelete",		ParseNewItemsAreAutoDelete },
	{ "NewItemsAreTaxable",			ParseNewItemsAreTaxable },
	{ "TwoTaxes",				ParseTwoTaxes },
	{ "RoundPricesToWholeNumbers",		ParseRoundPricesToWholeNumbers },
	{ "SeparateSortOrders",			ParseSeparateSortOrders },
	{ 0 }
};

static int
ReadPref( char const* psz, DbOpen* pdb )
{
	int n = 0;
	char const* pszValue;
	StrA s;

	pszValue = strchr( psz, '=' );
	if( !pszValue )
	    return -1;
	s.Set( psz, pszValue - psz );
	pszValue++;

	// Columns.

	for( ColumnInfo const* pcol = c_rgCols; pcol->szName; pcol++ )
	{
	    if( _stricmp( pcol->szName, s.Text() ) == 0 )
	    {
		char const* pszWidth = strchr( pszValue, ',' );
		StrA sValue;
		bool f;

		if( pszWidth )
		{
		    sValue.Set( pszValue, pszWidth - pszValue );
		    pszWidth++;
		}
		else
		    sValue.Set( pszValue );

		n = ParseBoolean( sValue.Text(), &f );
		if( n < 0 )
		    return n;

		int cx = f ? COLWIDTH_DEFAULT : 0;

		if( pcol->fWidths && f && pszWidth )
		{
		    n = -1;

		    for( int ii = 0; ii < DimensionOf( c_rgszWidth ); ii++ )
		    {
			if( _stricmp( pszWidth, c_rgszWidth[ii] ) == 0 )
			{
			    cx = ( ii == 3 ) ? COLWIDTH_FULL : COLWIDTH_DEFAULT;
			    n = 0;
			    break;
			}
		    }

		    if( n < 0 )
		    {
			cx = atoi( pszWidth );

			if( cx >= 5 && cx < 160 && !pcol->fDate )
			    n = 0;
		    }

		    if( n < 0 )
		    {
			if( pcol->fDate )
			    fprintf( stderr, "%s width must be Default or Wide.\n", s.Text() );
			else
			    fprintf( stderr, "%s width must be Default, Narrow, Medium, Wide, or 5 thru 160.\n", s.Text() );
			return n;
		    }
		}

		pdb->SetColumnWidth( pcol->col, cx );

		return 0;
	    }
	}

	// Other prefs.

	for( PrefParseInfo const* pParse = c_rgPrefs; pParse->szName; pParse++ )
	{
	    if( _stricmp( pParse->szName, s.Text() ) == 0 )
	    {
		if( pParse->pfnParse )
		{
		    n = pParse->pfnParse( pdb, pszValue );
		    if( n < 0 )
			return n;
		}

		return 0;
	    }
	}

	// Scream.

	fprintf( stderr, "Unrecognized pref '%s'.\n", s.Text() );
	return -1;
}

static int
ReadMetadata( FILE* const pfile, DbOpen* const pdb, bool const fUseFriendlyNames )
{
	enum MetaMode
	{
	    mmCategories,
	    mmStores,
	    mmPreferences,
	    mmNUMSECTIONS,
	    mmHunting = mmNUMSECTIONS
	};

	static char const* const c_rgSections[mmNUMSECTIONS] =
	{
	    "Categories:",
	    "Stores:",
	    "Preferences:",
	};

	bool rgfSections[mmNUMSECTIONS] = { false };
	MetaMode mm = mmHunting;
	StrT s;
	char sz[1024];
	int ii;
	int cLine = 1;
	int cCategories = 0;

	while( !feof( pfile ) && fgets( RgC( sz ), pfile ) )
	{
	    int cThisLine = cLine;
	    int cch = strlen( sz );
	    int n;
	    char const* psz;

	    if( !cch )
		continue;

	    // Check for trailing newline.

	    if( sz[cch - 1] == '\n' )
	    {
		cLine++;		// Increment line number.
		sz[--cch] = 0;		// Trim newline.
	    }

	    if( !cch )
		continue;

	    // Skip comment lines.

	    if( sz[0] == '#' || sz[0] == ';' )
		continue;

	    // Unless it's a section header, the line must begin with tab.

	    if( mm != mmHunting && sz[0] != '\t' )
		mm = mmHunting;

	    // Parser.

	    switch( mm )
	    {
	    case mmHunting:
		if( _strnicmp( "DatabaseName:", sz, 13 ) == 0 )
		{
		    StrA tmpA;
		    StrT tmpT;

		    psz = sz + 13;
		    while( *psz == ' ' || *psz == '\t' )
			++psz;

		    tmpA.Set( psz );
		    tmpA.Unquote();
		    psz = tmpA.Text();

		    SkipPrefix( &psz );
		    s.SetA( psz );

		    SysError const eSetName = pdb->SetName( s.Text() );

		    if( eSetName )
		    {
			fprintf( stderr, "Error while setting database name.\n" );
			return -1;
		    }

		    n = 0;
		    break;
		}
		if( _stricmp( "OmitPrefix", sz ) == 0 )
		{
		    // The original database didn't have the HS2_ prefix, so
		    // to avoid creating a second database on the handheld,
		    // we'll strip the prefix.

		    StrT tmpT;
		    tmpT.Set( pdb->Name() );

		    pdb->OmitPrefix();
		    SysError const eSetName = pdb->SetName( tmpT.Text() );

		    if( eSetName )
		    {
			fprintf( stderr, "Error while setting database name.\n" );
			return -1;
		    }

		    n = 0;
		    break;
		}
		for( ii = 0; ii < DimensionOf( c_rgSections ); ii++ )
		{
		    // Hunt for a matching section name.

		    if( _stricmp( c_rgSections[ii], sz ) == 0 )
		    {
			mm = MetaMode( ii );
			if( rgfSections[mm] )
			{
			    fprintf( stderr, "Encountered more than one '%s' section (line %d).\n", c_rgSections[mm], cThisLine );
			    return -1;
			}
			rgfSections[mm] = true;
			break;
		    }
		}
		n = ( mm == mmHunting ) ? -1 : 0;
		break;

	    case mmCategories:
		n = -1;
		{
		    StrT sCategory;
		    StrT sToken;
		    WORD icon = 0;

		    s.SetA( sz + 1 );
		    s.GetToken( sCategory );
		    sCategory.Unquote();

		    while( true )
		    {
			s.TrimLeft();

			if( !s.GetToken( sToken ) )
			    break;

			if( _tcsnicmp( TEXT( "icon=" ), sToken.Text(), 5 ) == 0 )
			{
			    long const val = _ttoi( sToken.Text() + 5 );

			    if( val & 0xffff0000 )
				icon = 0;
			    else
				icon = WORD( val );
			}
			else
			{
			    //$ todo: (chrisant) unexpected token
			}
		    }

		    if( cCategories )
		    {
			if( !sCategory.Length() )
			{
			    fprintf( stderr, "Missing category name.\n" );
			    break;
			}
			else if( sCategory.Length() >= c_cchMaxCategoryName )
			{
			    fprintf( stderr, "Category name '%ls' too long.\n", sCategory.Text() );
			    break;
			}
			else if( sCategory.Equal( TEXT( "Unfiled" ) ) ||
				 pdb->GetCategoryIdByName( sCategory.Text() ) != catUnfiledCategory )
			{
			    fprintf( stderr, "Category '%ls' already exists.\n", sCategory.Text() );
			    break;
			}
			else if( pdb->GetCategoriesCount() + 1 >= pdb->GetCategoriesMaximum() )
			{
			    fprintf( stderr, "Too many categories (maximum of %u).\n", pdb->GetCategoriesMaximum() );
			    break;
			}

			bool const fAdded = pdb->CategoryNew( sCategory.Text(), icon );

			if( !fAdded )
			{
			    fprintf( stderr, "Unexpected error while trying to add category '%ls'.\n", sCategory.Text() );
			    break;
			}
		    }
		    else
		    {
			if( !sCategory.Equal( TEXT( "Unfiled" ) ) )
			{
			    fprintf( stderr, "The first category must be 'Unfiled'.\n" );
			    break;
			}
		    }

		    cCategories++;
		    n = 0;
		}
		break;

	    case mmStores:
		n = -1;
		psz = sz + 1;
		{
		    Store id = storeUnfiledStore;

		    if( !fUseFriendlyNames )
		    {
			if( !IsDigit( *psz ) )
			{
			    fprintf( stderr, "Missing store id.\n" );
			    break;
			}

			unsigned const nStore = atoi( psz );

			if( int( nStore ) < 0 || nStore >= pdb->GetStoresMaximum() )
			{
			    fprintf( stderr, "Store id must be 0 through %u.\n", pdb->GetStoresMaximum() - 1 );
			    break;
			}

			id = Store( nStore );

			while( IsDigit( *psz ) )
			    psz++;

			if( *psz != ',' )
			{
			    fprintf( stderr, "Missing comma after store id.\n" );
			    break;
			}

			psz++;
		    }

		    s.SetA( psz );
		    s.Unquote();

		    if( !s.Length() )
		    {
			fprintf( stderr, "Missing store name.\n" );
			break;
		    }
		    else if( fUseFriendlyNames && _tcspbrk( s.Text(), TEXT( "=;" ) ) )
		    {
			fprintf( stderr, "Store names cannot contain = or ; unless the -i flag is used.\n" );
			break;
		    }
		    else if( s.Length() >= c_cchMaxStoreName )
		    {
			fprintf( stderr, "Store name '%ls' too long.\n", s.Text() );
			break;
		    }
		    else if( pdb->GetStoreIdByName( s.Text() ) != storeUnfiledStore )
		    {
			fprintf( stderr, "Store '%ls' already exists.\n", s.Text() );
			break;
		    }
		    else if( !fUseFriendlyNames && ( id < 0 || id >= pdb->GetStoresMaximum() ) )
		    {
			fprintf( stderr, "Store id must be 0 through %u.\n", pdb->GetStoresMaximum() - 1 );
			break;
		    }
		    else if( !fUseFriendlyNames && pdb->StoreExistsById( id ) )
		    {
			fprintf( stderr, "Store id %u already exists.\n", id );
			break;
		    }
		    else if( pdb->GetStoresCount() + 1 >= pdb->GetStoresMaximum() )
		    {
			fprintf( stderr, "Too many stores (maximum of %u).\n", pdb->GetStoresMaximum() );
			break;
		    }

		    bool const fAdded = fUseFriendlyNames ? pdb->StoreNew( s.Text() ) : pdb->StoreNewWithId( s.Text(), id );

		    if( !fAdded )
		    {
			fprintf( stderr, "Unexpected error while trying to add store '%ls'.\n", s.Text() );
			break;
		    }

		    n = 0;
		}
		break;

	    case mmPreferences:
		n = ReadPref( sz + 1, pdb );
		break;

	    default:
		Panic1( "Unhandled case mm == %d", mm );
		n = -1;
		break;
	    }

	    if( n < 0 )
	    {
		fprintf( stderr, "Unable to process line %d.\n", cThisLine );
		return n;
	    }
	}

	// Translate current store name.

	s.Set( s_sCurrentStore );
	s.Unquote();

	if( !s.Length() || !_tcsicmp( s.Text(), TEXT( "All Stores" ) ) )
	    pdb->SetStore( storeAllStores );
	else if( !_tcsicmp( s.Text(), TEXT( "Unfiled" ) ) )
	    pdb->SetStore( storeUnfiledStore );
	else
	{
	    unsigned id;

	    if( fUseFriendlyNames )
	    {
		id = pdb->GetStoreIdByName( s.Text() );

		if( id == storeUnfiledStore )
		{
		    fprintf( stderr, "CurrentStore name '%ls' unknown.\n", s.Text() );
		    return -1;
		}
	    }
	    else
	    {
		id = _ttoi( s.Text() );

		if( int( id ) < 0 || id >= pdb->GetStoresMaximum() )
		{
		    fprintf( stderr, "CurrentStore id must be between 0 and %u.\n", pdb->GetStoresMaximum() - 1 );
		    return -1;
		}
	    }

	    pdb->SetStore( BYTE( id ) );
	}

	return 0;
}

static int
ReadItems( FILE* pfile, DbOpen* pdb, bool fUseFriendlyNames )
{
	int n = 0;
	int cRecord = 1;
	int cLine = 2;
	StrA s;
	char sz[1024];

	// Verify header.

	GetHeaderString( s );
	s.Append( "\n" );

	if( !fgets( RgC( sz ), pfile ) )
	{
	    fprintf( stderr, "Error reading header line.\n" );
	    return -1;
	}

	if( _stricmp( s.Text(), sz ) != 0 )
	{
	    fprintf( stderr, "Header line is incorrect.\n" );
	    return -1;
	}

	// Read items.

	while( n >= 0 && !feof( pfile ) )
	{
	    SPNew<Item> spitem = new Item;

	    // Read from input file.

	    n = XferItem( pdb, pfile, spitem, false, &cRecord, &cLine, fUseFriendlyNames );

	    // Avoid overflow.

	    if( pdb->Count() >= pdb->GetRecordsMaximum() )
	    {
		fprintf( stderr, "Too many records.\n" );
		return -1;
	    }

	    // Add to database.

	    if( n >= 0 )
	    {
#ifdef DEBUG
		spitem->SetCanPack();
#endif

		WORD const flags = packNoSort|packNoUndo|packInsertAtIndex;
		SysError const ePack = pdb->PackRecord( pdb->Count(), spitem.UnsafeAddress(), flags, 0 );

		if( ePack )
		{
		    fprintf( stderr, "Error writing record %u.\n", pdb->Count() );
		    return -1;
		}
	    }
	}

	return n;
}

int
DoImportCSV( TCHAR const* const _pszFrom, TCHAR const* const pszTo, bool const fUseStoreIds, bool const fForceOverwrite )
{
	SPNew<DbOpen> spdb;

	// Check for an existing file, and either overwrite it or error out.

	if( FileExists( pszTo ) )
	{
	    if( fForceOverwrite )
	    {
		__DeleteFile( pszTo );
	    }
	    else
	    {
		fprintf( stderr, "Output file '%ls' already exists.\n", pszTo );
		return -1;
	    }
	}

	SysError const eNew = NewDatabase( pszTo, TEXT( "" ), &spdb );

	if( eNew )
	{
	    fprintf( stderr, "Unable to create new database '%ls'.\n", pszTo );
	    return -1;
	}

	// Determine input file names.

	StrT sItemsFile;
	StrT sMetadataFile;

	sItemsFile.Set( _pszFrom );
	if( GetExtension( sItemsFile.Text() ) )
	    sItemsFile.SetEnd( GetExtension( sItemsFile.Text() ) );
	sMetadataFile.Set( sItemsFile );

	sItemsFile.Append( c_szExtItems );
	sMetadataFile.Append( c_szExtMetadata );

	// Open and read metadata input file.

	int n = 0;
	FILE* pfile = 0;

	if( _tfopen_s( &pfile, sMetadataFile.Text(), TEXT( "r" ) ) || !pfile )
	{
	    fprintf( stderr, "Unable to open metadata input file '%ls'.\n", sMetadataFile.Text() );
	    return -1;
	}

	n = ReadMetadata( pfile, spdb, !fUseStoreIds );

	fclose( pfile );

	if( n < 0 )
	{
	    fprintf( stderr, "Error reading metadata file '%ls'.\n", sMetadataFile.Text() );
	    return n;
	}

	// Open and read items input file.

	if( _tfopen_s( &pfile, sItemsFile.Text(), TEXT( "r" ) ) || !pfile )
	{
	    fprintf( stderr, "Unable to open input file '%ls'.\n", sItemsFile.Text() );
	    return -1;
	}

	n = ReadItems( pfile, spdb, !fUseStoreIds );

	fclose( pfile );

	if( n < 0 )
	{
	    fprintf( stderr, "Error reading input file '%ls'.\n", sItemsFile.Text() );
	    return n;
	}

	// Validate the generated database.

	if( !spdb->Name() || !*spdb->Name() )
	{
	    fprintf( stderr, "Database name was not defined by the metadata file.\n" );
	    return -1;
	}

	// Post-processing.

	if( spdb->CanFlush( 0 ) )
	{
	    SysError const eFlush = spdb->Flush();

	    if( eFlush )
	    {
		fprintf( stderr, "Error saving the new database.\n" );
		return -1;
	    }
	}

	Assert( !n );

	return 0;
}

