/*
 * Number.h - abstraction around a floating point number.
 */

#pragma once

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#ifndef _WIN32_WCE
#include <float.h>
#endif

/*
 * NUMBER - abstraction for numbers, currently uses a combination of fixed
 * point and floating point arithmetic.
 */

#define MAX_PLACES	( 6 )
#define MAX_NUMBER	( 1000000 )	// should be "1" followed by MAX_PLACES zeroes

struct NUMBER
{
			NUMBER() { m_value = 0.0; }
			NUMBER( long l, WORD precision ) { FromLong( l, precision ); }

	long		ToLong( WORD precision ) const;
	long		ToLong_OverflowOk( WORD const precision ) const { return ToLong_Internal( precision ); }
	void		FromLong( long l, WORD precision ) { m_value = double( l ) / precision; }

	int		operator==( NUMBER const& num ) { return m_value == num.m_value; }
	int		operator<( NUMBER const& num ) { return m_value < num.m_value; }
	int		operator>( NUMBER const& num ) { return m_value > num.m_value; }

	void		operator+=( NUMBER const& num ) { m_value += num.m_value; }

	void		Add( NUMBER const& numX, NUMBER const& numY ) { m_value = numX.m_value + numY.m_value; }
	void		Subtract( NUMBER const& numX, NUMBER const& numY ) { m_value = numX.m_value - numY.m_value; }
	void		Multiply( NUMBER const& numX, NUMBER const& numY ) { m_value = numX.m_value * numY.m_value; }
	void		Divide( NUMBER const& numX, NUMBER const& numY ) { m_value = numX.m_value / numY.m_value; }

	void		Negate() { m_value = _chgsign( m_value ); }

	bool		IsNegative() const { return m_value < 0.0; }
	bool		IsOverflow() const;

	void		SetRawValue( double const value ) { m_value = value; }
	double		GetRawValue() const { return m_value; }

    protected:
	long		ToLong_Internal( WORD precision ) const;

    private:
	double		m_value;
};

inline long
NUMBER::ToLong( WORD const precision ) const
{
	Assert( !IsOverflow() );
	return ToLong_Internal( precision );
}

/*
 * Number and string helpers.
 */

void SkipSpaces( TCHAR const** ppsz );
void SkipDigits( TCHAR const** ppsz );
void SkipLinefeeds( TCHAR const** ppsz );
bool SkipCurrency( TCHAR const** ppsz, LPCTSTR pszCurrency = 0 );

#define QTY_PRECISION		( 1000 )
#define PRICE_PRECISION		( 100 )

#define cchMaxCurrency		( 7 )	// 6 chars, plus null terminator
#define cchMaxLocalizedNumber	( 13 + cchMaxCurrency )	// currency, plus "9,999,999.99", plus null terminator

enum
{
	lnCurrencySymbol	= 0x0001,
	lnAllowRounding		= 0x0002,
	lnTrimRightZeros	= 0x0004,
	lnShowZero		= 0x0008,
	lnTrimRightZeros2	= 0x0010,
	lnShowDecimals		= 0x0020,
};

#define parseError		( LONG_MAX )

LPCTSTR LocalizeNumber( long l, WORD flags, WORD precision, int* pcch = 0, LPCTSTR pszCurrency = 0 );
LPCTSTR LocalizeNumber( NUMBER num, WORD flags, WORD precision, int* pcch = 0, LPCTSTR pszCurrency = 0 );
long ParseNumber( LPCTSTR psz, WORD precision, TCHAR const** ppszOut, bool fSilent, LPCTSTR pszCurrency = 0 );

