/*
 * PalmLayer.h - Palm compatibility layer.
 */

#pragma once

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <ctype.h>

#include "str.h"
#include "sys.h"
#include "progress.h"
#include "_types.h"

/*
 * Time helpers.
 */

time_t __localtime();

/*
 * DataMgr.
 */

typedef Int16 DmComparF( void*, void*, Int32 other );

/*
 * Category equates.
 */

#define dmRecAttrCategoryMask		0x0F	// Mask for category number.
#define dmRecNumCategories		16	// Number of categories.
#define dmCategoryLength		16	// 15 chars + 1 null terminator.

#define dmAllCategories			0xff
#define dmUnfiledCategory  		0

#define dmMaxRecordIndex		0xffff

/*
 * Record Attributes.
 */

#define dmRecAttrDelete			0x80	// Delete this record next sync.
#define dmRecAttrDirty			0x40	// Archive this record next sync.
#define dmRecAttrBusy			0x20	// Record currently in use.
#define dmRecAttrSecret			0x10	// "Secret" record - password protected.

#define dmAllRecAttrs			( dmRecAttrDelete |		\
					  dmRecAttrDirty |		\
					  dmRecAttrBusy |		\
					  dmRecAttrSecret )

#define dmSysOnlyRecAttrs		( dmRecAttrBusy )

/*
 * Database Header equates.
 */

#define dmDBNameLength			32	// 31 chars + 1 null terminator.

/*
 * Attributes of a Database.
 */

#define dmHdrAttrResDB			0x0001	// Resource database.
#define dmHdrAttrReadOnly		0x0002	// Read Only database.
#define dmHdrAttrAppInfoDirty		0x0004	// Set if Application Info block is dirty.
#define dmHdrAttrBackup			0x0008	// Set if database should be backed up to PC if no app-specific synchronization conduit has been supplied.
#define dmHdrAttrOKToInstallNewer	0x0010
#define dmHdrAttrResetAfterInstall	0x0020	// Device requires a reset after this database is installed.
#define dmHdrAttrCopyPrevention		0x0040	// This database should not be copied.

#define dmHdrAttrStream			0x0080	// This database is used for file stream implementation.
#define dmHdrAttrHidden			0x0100	// This database should generally be hidden from view; for data (non-resource) databases, this hides the record count within the launcher info screen.
#define dmHdrAttrLaunchableData		0x0200	// This data database (not applicable for executables) can be "launched" by passing it's name to it's owner app ('appl' database with same creator) using the sysAppLaunchCmdOpenNamedDB action code.

#define dmHdrAttrOpen			0x8000	// Database not closed properly.

#define dmAllHdrAttrs			( dmHdrAttrResDB |		\
					  dmHdrAttrReadOnly |		\
					  dmHdrAttrAppInfoDirty |	\
					  dmHdrAttrBackup |		\
					  dmHdrAttrOKToInstallNewer |	\
					  dmHdrAttrResetAfterInstall |	\
					  dmHdrAttrCopyPrevention |	\
					  dmHdrAttrStream |		\
					  dmHdrAttrOpen )

#define dmSysOnlyHdrAttrs		( dmHdrAttrResDB |		\
					  dmHdrAttrOpen )

/*
 * Categories.
 */

#pragma pack( push, 2 )
struct AppInfoType
{
    private:
	UInt16		renamedCategories;	// Mac byte order...
    public:
	Char		categoryLabels [dmRecNumCategories] [dmCategoryLength];
	UInt8		categoryUniqIDs[dmRecNumCategories];
	UInt8		lastUniqID;		// Uniq IDs generated by the device are between
						// 0 - 127.  Those from the PC are 128 - 255.
	UInt8		padding;
};
#pragma pack( pop )

CASSERT( sizeof( AppInfoType ) == 276 );

typedef AppInfoType* AppInfoPtr;

/*
 * Colors; from Bitmap.h.
 */

// This is the structure of a color table. It maps pixel values into RGB
// colors. Each element in the table corresponds to the next index, starting
// at 0.

struct RGBColorType
{
	UInt8		index;			// Index of color or best match to cur CLUT or unused.
	UInt8		r;			// Amount of red, 0->255.
	UInt8		g;			// Amount of green, 0->255.
	UInt8		b;			// Amount of blue, 0->255.
};

/*
 * Palm OS layers.
 */

#include "_palm.h"
#include "_mem.h"
#include "_pdb.h"
#include "_dm.h"

/*
 * Palm OS utility helpers.
 */

#define DmWriteElement( p, elm ) \
	do { \
	    /* elm must be both member of p, and a variable */ \
	    AssertSameSize( p->elm, elm ); \
	    DmWrite( p, OffsetOf( *p, elm ), &elm, sizeof( elm ) ); \
	} while( 0 )

#define DmWriteElementEx( p, elm, var ) \
	do { \
	    AssertSameSize( p->elm, var ); \
	    DmWrite( p, OffsetOf( *p, elm ), &var, sizeof( var ) ); \
	} while( 0 )

inline Boolean FMultipleBits( UInt16 x ) { return !!( x & ( x-1 ) ); }
inline Int16 __Abs( Int16 x ) { return ( x >= 0 ) ? x : -x; }

Err __MemHandleUnlock( MemHandle h );
Err __MemHandleFree( MemHandle h );

Err __MemPtrUnlock( MemPtr pv );
#define __MemPtrFree __MemChunkFree
Err __MemChunkFree( MemPtr pv );
MemHandle __MemPtrRecoverHandle( MemPtr pv );

