/*
 * categories.c - Custom category support routines.
 */

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#include "precomp.h"

#include "storelist.h"
#include "item.h"
#include "palmlayer.h"

#include "globalhs2.h"
#include "columns.h"
#include "data1x.h"
#include "storeshs2.h"
#include "db.h"
#include "pdbhs2.h"

extern void* GetMemberPtr( void* pvRecord, BYTE tag );

/*
 * GetCategories - get the categories from the specified database.
 */

static AppInfoType*
GetCategories( DmOpenRef pdb )
{
	LocalID appInfoID;

	// Get the categories from the database.

	if( DmDatabaseInfo( pdb, 0, 0, 0, 0, 0, 0, 0, &appInfoID, 0, 0, 0 ) )
	{
	    PanicSz( "Bad database ID" );
	    return 0;
	}

	return ( AppInfoType* )MemLocalIDToLockedPtr( appInfoID, 0 );
}

void
DbOpenHs2::MoveCategory( WORD const toCategory, WORD const fromCategory )
{
	Assert( m_pdb );
	Assert( m_pCategories );

	Assert( fromCategory != catUnfiledCategory && fromCategory != catAllCategories );
	Assert( toCategory != catAllCategories );
	Assert( fromCategory < GetCategoriesMaximum() );
	Assert( FImplies( toCategory != catUnfiledCategory, toCategory < GetCategoriesMaximum() ) );
	Assert( FImplies( toCategory != catUnfiledCategory, m_pCategories->IndexFromId_Slow( toCategory ) < GetCategoriesMaximum() ) );

	MemHandle h;
	UInt8 *pRec;
	unsigned const cMax = DmNumRecords( m_pdb );
	UInt8 const categoryToWrite = UInt8( toCategory );
	ScopedProgressHolder sph;

	sph.Create();
	sph.Init( cMax, PROGRESS_PREDICTIVEWAIT );
	for( UInt16 record = 0; record < cMax; record++ )
	{
	    Boolean fDirty = false;

	    sph.Update( record );
	    h = DmGetRecord( m_pdb, record );
	    pRec = ( UInt8* )MemHandleLock( h );
	    if( tagMasterStoreList != *pRec && tagCategoryList != *pRec )
	    {
		UInt16 categoryRecord = catUnfiledCategory;
		UInt8 const* const pCategory = ( UInt8 const* )GetMemberPtr( pRec, tagCategory );
		if( pCategory )
		    categoryRecord = *pCategory;
		if( categoryRecord == fromCategory )
		{
		    Assert( pCategory );
		    if( toCategory == catUnfiledCategory )
		    {
			UInt32 const ofs = pCategory - 1 - pRec;
			DmWrite( pRec, ofs, pRec + ofs + 1 + sizeof( categoryToWrite ), MemHandleSize( h ) - ( ofs + 1 + sizeof( categoryToWrite ) ) );
		    }
		    else
		    {
			CASSERT( SameSize( categoryToWrite, UInt8 ) );
			DmWrite( pRec, pCategory - pRec, &categoryToWrite, sizeof( categoryToWrite ) );
		    }
		    fDirty = true;
		}
	    }
	    MemHandleUnlock( h );
	    DmReleaseRecord( m_pdb, record, fDirty );
	}
}

bool
DbOpenHs2::CategoriesConvertFromNative()
{
#ifdef DEBUG
	Assert( !m_prefs.fCustomCategories );
	Assert( !m_pCategories );
	UInt16 cConverted = 0;
#endif

	unsigned const cRecords = DmNumRecords( m_pdb );
	UInt16 rgNew[dmRecNumCategories] = {};
	AppInfoType* const appCategories = GetCategories( m_pdb );
	UInt16 const dmCategoryFilter = m_prefs.get_category();
	ScopedProgressHolder sph;

	m_prefs.fCustomCategories = true;
	m_pCategories = new CustomCategoryList;
	SetCategoriesDirty();

	// Build mappings from old categories to new categories.
	for( UInt8 index = 1; index < dmRecNumCategories; index++ )	// Start at 1 to skip Unfiled.
	    if( *appCategories->categoryLabels[index] )
		rgNew[index] = m_pCategories->New( appCategories->categoryLabels[index], m_prefs.get_CategoryIcon_deprecated( index ) );

	// Convert the records to use the new categories.
	ClearUndo();
	sph.Create();
	sph.Init( cRecords, PROGRESS_PREDICTIVEWAIT );
	for( unsigned dmRecord = 0; dmRecord < cRecords; dmRecord++ )
	{
	    sph.Update( dmRecord );

	    MemHandle const h = DmGetRecord( m_pdb, dmRecord );
	    UInt8 const tagRecord = *( UInt8* )MemHandleLock( h );
	    MemHandleUnlock( h );
	    DmReleaseRecord( m_pdb, dmRecord, false );

	    if( tagRecord != tagMasterStoreList && tagRecord != tagCategoryList )
	    {
		UInt16 attr;

		DmRecordInfo( m_pdb, dmRecord, &attr, 0, 0 );

		UInt16 const categoryOld = attr & dmRecAttrCategoryMask;

		if( categoryOld != dmUnfiledCategory )
		{
		    Assert( dmRecord >= m_cMetadataRecords );
		    ItemPtr pitem = UnPackRecord( dmRecord - m_cMetadataRecords );
		    pitem->m_category = rgNew[categoryOld];
		    pitem->m_flags |= flagCATEGORY;
		    PackRecord( dmRecord - m_cMetadataRecords, &pitem, packNoSort|packNoUndo|packNoDelete|packNoCounts, 0 );
#ifdef DEBUG
		    cConverted++;
#endif
		}
	    }
	}

	// Convert prefs.
	WORD categoryFilter = catAllCategories;
	if( dmCategoryFilter == dmAllCategories )
	    categoryFilter = catAllCategories;
	else if( dmCategoryFilter == dmUnfiledCategory )
	    categoryFilter = catUnfiledCategory;
	else if( dmCategoryFilter < DimensionOf( rgNew ) )
	    categoryFilter = rgNew[dmCategoryFilter];
	// SetCategory will no-op if the category id is the same, but since
	// we're changing all the id's any equality is accidental.  So first
	// force the category to a value that hasn't changed, and then set it
	// to the new value.  This ensures the bit table gets refreshed, etc.
	SetCategory( catAllCategories );
	SetCategory( categoryFilter );

	// Zap the native categories structure.
	DmSet( appCategories, OffsetOf( *appCategories, categoryLabels ), sizeof( appCategories->categoryLabels ), 0 );
	DmStrCopy( appCategories, OffsetOf( *appCategories, categoryLabels ), "WrongCategories" );

	MemPtrUnlock( appCategories );
	WriteAppInfo();

#ifdef DEBUG
	DbgMsg(( 20, "converted %u records", cConverted ));
#endif

	return true;
}

/*
 * CustomCategoryList.
 */

CustomCategoryList::CustomCategoryList()
	: m_tag( tagCategoryList )
	, m_cMax( MAX_CATEGORIES )
	, m_reserved( 0 )
{
#ifdef DEBUG
	BYTE const* pb = reinterpret_cast<BYTE const*>( this );
	for( unsigned ii = 2; ii < sizeof( *this ); ii++ )
	    Assert( !pb[ii] );
#endif
}

CategoryDef const*
CustomCategoryList::CategoryFromIndex( UInt8 index ) const
{
	Assert( index < m_cCategories );
	return &m_rgCategory[IdFromIndex( index )];
}

static int
CmpCategoryFind( const void *pvFind, WORD index, void* pvCookie )
{
	CustomCategoryList const *pCategories = const_cast<CustomCategoryList const*>( static_cast<CustomCategoryList*>( pvCookie ) );
	return StrCaselessCompareSort( ( PCSZ )pvFind, pCategories->CategoryFromIndex( UInt8( index ) )->m_szCategory );
}

UInt16
CustomCategoryList::IdFromName( PCSZ const psz ) const
{
	UInt16 indexFound;

	if( FBinarySearch( psz, const_cast<void*>( static_cast<void const*>( this ) ), Count(), &indexFound, CmpCategoryFind ) )
	    return IdFromIndex( UInt8( indexFound ) );

	return catUnfiledCategory;
}

UInt8
CustomCategoryList::IndexFromId_Slow( UInt16 const category ) const
{
	Assert( category != catUnfiledCategory && category != catAllCategories );
	Assert( category < m_cMax );
	UInt8 c = m_cCategories;
	UInt8 const *pid = m_rgIndexToId;
	UInt8 const idCompare = UInt8( category );
	while( c-- )
	{
	    if( *pid == idCompare )
		return ( UInt8 )( pid - m_rgIndexToId );
	    pid++;
	}

	Panic1( "BAD: category %d not found!", category );
	return UInt8( -1 );
}

UInt16
CustomCategoryList::HashCategoryId( PCSZ psz ) const
{
	UInt16 categoryNew;

	Assert( Count() < m_cMax );

	if( *psz >= 'a' && *psz <= 'z' )
	    categoryNew = 18 + 3 * ( *psz - 'a' );
	else if( *psz >= 'A' && *psz <= 'Z' )
	    categoryNew = 18 + 3 * ( *psz - 'A' );
	else
	    categoryNew = 0;

	UInt8 c = m_cMax;
	Assert( MAX_CATEGORIES == c );
	CASSERT( MAX_CATEGORIES == DimensionOf( m_rgCategory ) );

	while( 1 )
	{
	    AssertSz( c, "category count is incorrect" );
	    c--;

	    if( !*( m_rgCategory[categoryNew].m_szCategory ) )
		break;

	    categoryNew++;
	    categoryNew %= MAX_CATEGORIES;
	}

	return categoryNew;
}

void
CustomCategoryList::AddToEndOfIndex( UInt16 const category )
{
	Assert( m_cCategories < m_cMax );
	Assert( category < m_cMax );

	m_rgIndexToId[m_cCategories] = UInt8( category );
	m_cCategories++;
}

UInt16
CustomCategoryList::New( PCSZ pszName, UInt16 icon )
{
	Assert( pszName );
	Assert( *pszName );

	WORD const category = HashCategoryId( pszName );

	SetName( category, pszName );
	SetIcon( category, icon );

	AddToEndOfIndex( category );
	Sort();

	return category;
}

void
CustomCategoryList::Remove( UInt16 const category )
{
	UInt8 const index = IndexFromId_Slow( category );

	// Decrement the number of categories.
	Assert( m_cCategories > 0 );
	m_cCategories--;

	// Clear the category's entry.
	ZeroMemory( m_rgCategory + category, sizeof( m_rgCategory[0] ) );

	// Shift the other categories in the sorted list.
	if( index < m_cCategories )
	    MemMove( m_rgIndexToId + index, m_rgIndexToId + index + 1, ( m_cCategories - index ) * sizeof( m_rgIndexToId[0] ) );
}

int
CustomCategoryList::CompareCategories( MemPtr pA, MemPtr pB, Int32 other )
{
#define pThis ( ( CustomCategoryList* )other )
	Assert( other );
	return StrCaselessCompare( pThis->m_rgCategory[*( UInt8* )pA].m_szCategory,
				   pThis->m_rgCategory[*( UInt8* )pB].m_szCategory );
#undef pThis
}

void
CustomCategoryList::Sort()
{
	UInt8 rgid[MAX_CATEGORIES];

	Assert( DimensionOf( rgid ) == m_cMax );
	AssertSameSize( m_rgIndexToId[0], rgid[0] );
	AssertSameSize( m_rgIndexToId, rgid );

	memcpy( rgid, m_rgIndexToId, sizeof( rgid ) );

	if( m_cCategories )
	    SysInsertionSort( rgid, m_cCategories, sizeof( rgid[0] ), CompareCategories, ( Int32 )this );

	memcpy( m_rgIndexToId, rgid, sizeof( rgid ) );
}

