/*
 * pdbcat.h - Custom category support routines.
 */

#pragma once

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

/*
 * Category List Metadata Record
 *
 * The category list is a record in the database, at index 0 (if there is no
 * master stores list) or 1 (if there is a master stores list).  The record is
 * a fixed size, which eliminates complicated code for resizing the record and
 * the lists it contains, making the app run faster as well as saving battery
 * power.  The record is not created until the custom categories are activated
 * (when the user adds a new category); if the user deletes all defined
 * categories, we also delete the category list record.
 *
 * The format of the record is:
 *
 *	UInt8 tag - tagCategoryList, identifies this as the record containing
 *	the category list.
 *
 *	UInt8 cMax - maximum number of categories this database supports (cMax
 *	must be an even number, to ensure UInt16 alignment for rgCategory
 *	array).  This is for future compatability and is otherwise unused.
 *
 *	UInt8 cCategories - number of categories (not including the magic
 *	"Unfiled" and "All" categories).
 *
 *	UInt8 reserved - ensure correct alignment.
 *
 *	UInt8 rgIndexToId[cMax] - sorted index of categories, maps sorted
 *	indices to actual category entries.
 *
 *	CategoryDef rgCategory[cMax] - info for each category, hashes based on
 *	first character of name.
 *
 * For a 100-category list, the total data size of the record is around 2kb.
 */

#define MAX_CATEGORIES		( 100 )		// Maximum number of categories.

struct CategoryDef
{
    public:
	UInt16			GetIcon() const { return MacWord( m_icon ); }
	void			SetIcon( UInt16 const icon ) { m_icon = MacWord( icon ); }
    private:
	UInt16			m_icon;
    public:
	Char			m_szCategory[dmCategoryLength];
};
CASSERT( ( sizeof( CategoryDef ) & 1 ) == 0 );	// Verify alignment.
CASSERT( ( MAX_CATEGORIES & 1 ) == 0 );		// MAX_CATEGORIES must be even, for alignment.

class CustomCategoryList
{
    public:
	CategoryDef const*	CategoryFromId( UInt16 category ) const;
	CategoryDef const*	CategoryFromIndex( UInt8 index ) const;
	UInt16			IdFromName( PCSZ psz ) const;
	UInt16			IdFromIndex( UInt8 index ) const;
	UInt8			IndexFromId_Slow( UInt16 category ) const;
	UInt16			HashCategoryId( PCSZ psz ) const;
	UInt8			Count() const { return m_cCategories; }

				CustomCategoryList();

	UInt16			New( PCSZ pszName, UInt16 icon );
	void			Remove( UInt16 category );
	void			Sort();

	bool			SetName( UInt16 category, PCSZ psz );
	bool			SetIcon( UInt16 category, UInt16 icon );

    public:
	UInt8 const		m_tag;		// Must be tagCategoryList.
	UInt8 const		m_cMax;		// Must be MAX_CATEGORIES.
    private:
	UInt8			m_cCategories;	// Number of categories.
	UInt8 const		m_reserved;	// Alignment.
	UInt8			m_rgIndexToId[MAX_CATEGORIES];
	CategoryDef		m_rgCategory[MAX_CATEGORIES];

    private:
	void			AddToEndOfIndex( UInt16 category );
	static int		CompareCategories( MemPtr pA, MemPtr pB, Int32 other );
};

inline CategoryDef const*
CustomCategoryList::CategoryFromId( UInt16 category ) const
{
	Assert( category < m_cMax );
	Assert( category < MAX_CATEGORIES );
	return m_rgCategory + category;
}

inline UInt16
CustomCategoryList::IdFromIndex( UInt8 index ) const
{
	Assert( index < m_cCategories );
	CASSERT( SameSize( m_rgIndexToId[index], UInt8 ) );
	return m_rgIndexToId[index];
}

inline bool
CustomCategoryList::SetName( UInt16 category, PCSZ psz )
{
	Assert( category < m_cMax );
	Assert( category < MAX_CATEGORIES );
	Assert( psz );
	Assert( *psz );
	if( StrCompare( m_rgCategory[category].m_szCategory, psz ) == 0 )
	    return false;
	unsigned const cch = Min( unsigned( StrLen( psz ) ), DimensionOf( m_rgCategory[0].m_szCategory ) - 1 );
	memcpy( m_rgCategory[category].m_szCategory, psz, cch * sizeof( *psz ) );
	m_rgCategory[category].m_szCategory[cch] = 0;
	return true;
}

inline bool
CustomCategoryList::SetIcon( UInt16 category, UInt16 icon )
{
	Assert( category <= m_cMax );
	Assert( category <= MAX_CATEGORIES );
	if( m_rgCategory[category].GetIcon() == icon )
	    return false;
	m_rgCategory[category].SetIcon( icon );
	return true;
}

/*
 * Category APIs.
 */

void __CategorySetName( DbOpenHs2* popen, UInt16 category, Char const* psz, UInt16 icon );
void __CategoryRemove( DbOpenHs2* popen, UInt16 category );

