/*
 * pdbhs2.cpp - HS2 database class.
 */

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#include "precomp.h"

#include "str.h"
#include "sys.h"

#include "storelist.h"
#include "item.h"
#include "palmlayer.h"

#include "globalhs2.h"
#include "columns.h"
#include "data1x.h"
#include "storeshs2.h"
#include "db.h"
#include "pdbhs2.h"

extern void* GetMemberPtr( void* pvRecord, BYTE tag );

/*
 * Hs2Database.
 */

Hs2Database::Hs2Database()
{
	m_popen = 0;
}

void
Hs2Database::SetPopen( DbOpenHs2* popen )
{
	if( popen )
	    Assert( !m_popen );
	else
	    Assert( m_popen );

	m_popen = popen;
}

/*
 * DbOpenHs2 functions to open a database.
 */

DbOpenHs2*
DbOpenHs2::DmNewDatabase( LPCTSTR pszFile, LPCSTR pszName )
{
	Assert( pszFile );
#ifdef OS_WINDOWSMOBILE
	Assert( *pszFile == '\\' );
#endif
	Assert( pszName );

	Err err = 0;
	Hs2Database* pdb = 0;
	UInt16 dbAttrs;
	UInt16 version = DBVersion;
	DbOpenHs2* popen;

	popen = new DbOpenHs2( pszFile );
	CPE( popen );

	pdb = new Hs2Database;
	pdb->Initialize( pszName, AppType, DBType, false );

	// Ensures the database will always get backed up during HotSync, even
	// if there is no conduit for the application.

	CE( pdb->DmDatabaseInfo( 0, &dbAttrs, 0, 0, 0, 0, 0, 0, 0, 0, 0 ) );
	dbAttrs = dbAttrs | dmHdrAttrBackup;
	CE( pdb->DmSetDatabaseInfo( 0, &dbAttrs, &version, 0, 0, 0, 0, 0, 0, 0, 0 ) );

	// Initialize the AppInfo block.

	popen->m_pdb = pdb;

	CE( popen->ReadAppInfo( newlistDefault ) );

	pdb->SetPopen( popen );

	popen->RefreshMetadataCounter();

Out:
	if( err )
	{
	    if( pdb )
	    {
		DmCloseDatabase( pdb );
		popen->m_pdb = 0;

		delete pdb;
		pdb = 0;
	    }

	    if( popen )
	    {
		delete popen;
		popen = 0;
	    }
	}

	return popen;

LError:
	goto Out;
}

DbOpenHs2*
DbOpenHs2::DmOpenDatabase( LPCTSTR pszFile, SysError& e, IProgress* pProgress )
{
	Err		err = 0;
	DbOpenHs2*	popen;
	DbOpenHs2*	popenRet = 0;

	Assert( pszFile );

	e = SE_UNKNOWN;

	popen = new DbOpenHs2( pszFile );
	if( !popen )
	{
	    e = SE_OUTOFMEMORY;
	    goto LError;
	}

	popen->m_pdb = new Hs2Database;
	if( !popen->m_pdb )
	{
	    e = SE_OUTOFMEMORY;
	    goto LError;
	}

	popen->m_fReadonly = true;
	( ( Hs2Database* )popen->m_pdb )->SetPopen( popen );

	if( popen->m_pdb->Read( pszFile, pProgress ) < 0 )
	    goto LError;

#ifdef DEBUG
	UInt16 versionBefore;
	CE( DmDatabaseInfo( popen->m_pdb, 0, 0, &versionBefore, 0, 0, 0, 0, 0, 0, 0, 0 ) );
#endif

	UInt32 type;
	UInt32 creator;

	CE( DmDatabaseInfo( popen->m_pdb, 0, 0, 0, 0, 0, 0, 0, 0, 0, &type, &creator ) );

	if( DBType != type || AppType != creator )
	{
	    e = SE_UNRECOGNIZEDDATABASE;
	    goto LError;
	}

	popen->m_fReadonly = false;

	err = popen->ReadAppInfo( newlistDefault );
	if( err == dmErrDatabaseProtected )
	    e = SE_OLDFORMATDATABASE;
	if( err )
	    goto LError;

#ifdef DEBUG
	UInt16 versionAfter;
	CE( DmDatabaseInfo( popen->m_pdb, 0, 0, &versionAfter, 0, 0, 0, 0, 0, 0, 0, 0 ) );

	if( popen->m_prefs.fCustomCategories )
	    DbgMsg(( 5, TEXT( "database uses CUSTOM categories" ) ));
	else
	    DbgMsg(( 5, TEXT( "database uses normal categories" ) ));
#endif

	popen->RefreshMetadataCounter();

	Assert( !popen->IsDirty() || versionBefore != versionAfter );

	popenRet = popen;
	popen = 0;

	e = SE_OK;

LError:

	delete popen;
	return popenRet;
}

/*
 * Hs2DbQuickOpen.
 */

DbQuickOpen*
QuickOpenHs2Database( LPCTSTR pszFile )
{
	Assert( pszFile );
#ifdef OS_WINDOWSMOBILE
	Assert( *pszFile == '\\' );
#endif

	SysError e = SE_OK;
	Hs2DbQuickOpen* pqo = 0;
	SPNew<FileIO> spFile = FileIO::Create();

	if( !spFile )
	    return 0;

	spFile->Set( pszFile );
	if( spFile->Open( FM_READ, e ) )
	{
	    PdbHeader hdr;

	    if( spFile->Read( &hdr, sizeof( hdr ), e ) &&
		hdr.GetCreator() == AppType &&
		hdr.GetType() == DBType )
	    {
		char sz[dmDBNameLength];
		const char* psz = sz;

		CchCopy( hdr.GetName(), RgC( sz ) );
		SkipPrefix( &psz );

		pqo = new Hs2DbQuickOpen;
		pqo->m_sName.SetA( psz );
		pqo->m_cItems = hdr.NumRecords();
		pqo->m_cNeeded = unsigned( -1 );

#ifdef DEBUG
		StripFilename( pszFile, pqo->m_sPath, &pqo->m_sFile );
		pqo->m_version = hdr.GetVersion();
#endif

		if( hdr.NumRecords() )
		{
		    BYTE bTag0;
		    BYTE bTag1 = tagEND;
		    RawRecordEntry raw0;
		    RawRecordEntry raw1;

		    if( ( spFile->Read( &raw0, sizeof( raw0 ), e ) ) &&
			( hdr.NumRecords() <= 1 || ( spFile->Read( &raw1, sizeof( raw1 ), e ) ) ) &&
			( spFile->Seek( raw0.GetOfsData(), e ) && spFile->Read( &bTag0, sizeof( bTag0 ), e ) ) &&
			( hdr.NumRecords() <= 1 || ( spFile->Seek( raw1.GetOfsData(), e ) && spFile->Read( &bTag1, sizeof( bTag1 ), e ) ) ) )
		    {
			if( bTag0 == tagCategoryList )
			{
			    pqo->m_cItems--;
			}
			else if( bTag0 == tagMasterStoreList )
			{
			    pqo->m_cItems--;

			    if( hdr.NumRecords() > 1 && bTag1 == tagCategoryList )
				pqo->m_cItems--;
			}
		    }
		    else
		    {
			pqo->m_cItems = unsigned( -1 );
		    }
		}

		if( pqo->m_cItems != -1 &&
		    hdr.GetVersion() >= DBVersion_RECALCITEMCOUNTS &&
		    hdr.GetOfsAppInfo() )
		{
		    UInt16 cBoldAll;
		    const UInt32 ofsCBoldAll = hdr.GetOfsAppInfo() + OffsetOf0( ThisAppInfoType, prefs ) + 26;

		    if( spFile->Seek( ofsCBoldAll, e ) &&
			spFile->Read( &cBoldAll, sizeof( cBoldAll ), e ) )
		    {
			pqo->m_cNeeded = MacWord( cBoldAll );
		    }
		}
	    }
	}

	return pqo;
}

SysError
Hs2DbQuickOpen::Rename( const TCHAR* pszFile, const StrT* ps )
{
	Assert( pszFile );
#ifdef OS_WINDOWSMOBILE
	Assert( *pszFile == '\\' );
#endif
	Assert( ps->Length() < dmDBNameLength );

	StrT tmp;
	StrA s;

	// Prepend the HS2_ name prefix for consistency and Palm
	// compatibility.

	tmp.Printf( TEXT( "HS2_%s" ), ps->Text() );

	s.SetT( tmp );

	if( s.Length() >= dmDBNameLength )
	    s.SetLength( dmDBNameLength - 1 );

	// Open the file and write the name.

	SysError e = SE_OK;
	SPNew<FileIO> spFile = FileIO::Create();

	if( !spFile )
	    return SE_OUTOFMEMORY;

	spFile->Set( pszFile );
	if( !spFile->Open( FM_READWRITE, e ) )
	    return e;

	if( !spFile->Write( s.Text(), s.Length() + 1, e ) )
	    return e;

	return SE_OK;
}

/*
 * DbOpenSimpleHs2.
 */

class DbOpenSimpleHs2 : public DbOpenSimple
{
	typedef DbOpenSimple base;

    public:
				DbOpenSimpleHs2();
				~DbOpenSimpleHs2();

	static DbOpenSimple*	SimpleOpenHs2Database( LPCTSTR pszFile );

	OverrideConst( base, LPCTSTR, Name, () );
	OverrideConst( base, bool, HasPrefix, () );
	OverrideConst( base, LPCTSTR, FileName, () );
	OverrideConst( base, unsigned, Count, () );

	// Read records.

	Override( base, bool, SetLookup, ( StrT const* psLookup, bool fBroad ) );
	OverrideConst( base, LPCTSTR, GetLookup, () );
	OverrideConst( base, bool, IsLookupBroad, () );

	Override( base, bool, SetFilterReasonCompleteness, ( bool fComplete ) );
	OverrideConst( base, FilterReason, IsRecordFiltered, ( unsigned record ) );

	OverrideConst( base, Item*, UnPackRecord, ( unsigned record ) );
	OverrideConst( base, bool, GetItemQuickInfo, ( unsigned record, ItemQuickInfo* pqi ) );

	OverrideConst( base, bool, GetRecordUniqueId, ( unsigned record, DWORD* pdwUniqueId ) );
	OverrideConst( base, bool, FindRecordByUniqueId, ( DWORD uniqueId, unsigned* precord ) );

    private:
	unsigned		DmCount() const { return m_hdr.NumRecords(); }
	bool			ReadRawRecordEntry( unsigned dmRecord, RawRecordEntry* praw ) const;
	DWORD			GetRecordSize( unsigned dmRecord, DWORD ofsData ) const;
	BYTE*			ReadRecordData( unsigned dmRecord, RawRecordEntry* praw ) const;

	unsigned		ToDmRecord( unsigned record ) const;
	unsigned		FromDmRecord( unsigned dmRecord ) const;

    private:
	SPNew<FileIO>		m_spFile;
	PdbHeader		m_hdr;

#ifdef OS_WINDOWS
	DWORD const		m_codepage;
#endif

	StrT			m_sFile;
	StrT			m_sName;
	unsigned		m_cItems;
	unsigned		m_cMetadata;

	bool			m_fCompleteReasons;
	bool			m_fFilterPrivate;
	bool			m_fHasPrefix;
	StrT			m_sLookup;
	StrA			m_sLookupA;
};

inline unsigned
DbOpenSimpleHs2::ToDmRecord( unsigned record ) const
{
	if( record == noRecordSelected )
	    return noRecordSelected;

	Assert( record + m_cMetadata < dmMaxRecordIndex );

	return record + m_cMetadata;
}

inline unsigned
DbOpenSimpleHs2::FromDmRecord( unsigned dmRecord ) const
{
	if( dmRecord == noRecordSelected )
	    return noRecordSelected;

	Assert( dmRecord >= m_cMetadata );

	return dmRecord - m_cMetadata;
}

DbOpenSimple*
DbOpenSimpleHs2::SimpleOpenHs2Database( LPCTSTR pszFile )
{
	Assert( pszFile );
#ifdef OS_WINDOWSMOBILE
	Assert( *pszFile == '\\' );
#endif

	SysError e = SE_OK;
	SPNew<DbOpenSimpleHs2> spNew = new DbOpenSimpleHs2;

	if( !spNew )
	    return 0;

	spNew->m_spFile = FileIO::Create();

	if( !spNew->m_spFile )
	    return 0;

	spNew->m_spFile->Set( pszFile );
	if( !spNew->m_spFile->Open( FM_READ, e ) )
	    return 0;

	if( !spNew->m_spFile->Read( &spNew->m_hdr, sizeof( spNew->m_hdr ), e ) ||
	    spNew->m_hdr.GetCreator() != AppType ||
	    spNew->m_hdr.GetType() != DBType )
	    return 0;

	char sz[dmDBNameLength];
	const char* psz = sz;

	CchCopy( spNew->m_hdr.GetName(), RgC( sz ) );
	SkipPrefix( &psz );

	spNew->m_sFile.Set( pszFile );
	spNew->m_sName.SetA( psz );
	spNew->m_fHasPrefix = psz != sz;
	spNew->m_cItems = spNew->m_hdr.NumRecords();
	spNew->m_cMetadata = 0;

	if( spNew->m_cItems )
	{
	    BYTE bTag;
	    RawRecordEntry raw;

	    if( !spNew->ReadRawRecordEntry( 0, &raw ) ||
		!spNew->m_spFile->Seek( raw.GetOfsData(), e ) ||
		!spNew->m_spFile->Read( &bTag, sizeof( bTag ), e ) )
	    {
		PanicSz( "Error trying to access record #0 to check if there is a store list." );
		return 0;
	    }

	    if( bTag == tagMasterStoreList )
	    {
		spNew->m_cItems--;
		spNew->m_cMetadata++;

		if( spNew->m_cItems &&
		    !spNew->ReadRawRecordEntry( 1, &raw ) ||
		    !spNew->m_spFile->Seek( raw.GetOfsData(), e ) ||
		    !spNew->m_spFile->Read( &bTag, sizeof( bTag ), e ) )
		{
		    PanicSz( "Error trying to access record #1 to check if there is a category list." );
		    return 0;
		}
	    }

	    if( bTag == tagCategoryList )
	    {
		spNew->m_cItems--;
		spNew->m_cMetadata++;
	    }
	}

	return spNew.Transfer();
}

DbOpenSimple*
SimpleOpenHs2Database( LPCTSTR pszFile )
{
	return DbOpenSimpleHs2::SimpleOpenHs2Database( pszFile );
}

DbOpenSimpleHs2::DbOpenSimpleHs2()
#ifdef OS_WINDOWS
	: m_codepage( GetACP() )
#endif
{
	Assert( !m_cItems );
	Assert( !m_cMetadata );
	Assert( !m_fCompleteReasons );
	Assert( !m_fFilterPrivate );

	m_fFilterPrivate = privShowRecords != GetPrivateRecordVisualMode();
}

DbOpenSimpleHs2::~DbOpenSimpleHs2()
{
}

LPCTSTR
DbOpenSimpleHs2::Name() const
{
	return m_sName.Text();
}

bool
DbOpenSimpleHs2::HasPrefix() const
{
	return m_fHasPrefix;
}

LPCTSTR
DbOpenSimpleHs2::FileName() const
{
	return m_sFile.Text();
}

unsigned
DbOpenSimpleHs2::Count() const
{
	return m_cItems;
}

bool
DbOpenSimpleHs2::SetLookup( StrT const* psLookup, bool const fBroad )
{
	SUPPRESS_UNUSED_VAR( fBroad );

#ifdef _WIN32_WCE
	m_fFilterPrivate = privShowRecords != GetPrivateRecordVisualMode();
#endif

	if( psLookup )
	{
	    m_sLookup.Set( *psLookup );
	    m_sLookupA.SetT( *psLookup );
	}
	else
	{
	    m_sLookup.Clear();
	    m_sLookupA.Clear();
	}

	return true;
}

LPCTSTR
DbOpenSimpleHs2::GetLookup() const
{
	if( !m_sLookup.Length() )
	    return 0;

	return m_sLookup.Text();
}

bool
DbOpenSimpleHs2::IsLookupBroad() const
{
	return false;
}

bool
DbOpenSimpleHs2::SetFilterReasonCompleteness( bool fComplete )
{
	bool const fRet = m_fCompleteReasons;

	m_fCompleteReasons = fComplete;

	return fRet;
}

FilterReason
DbOpenSimpleHs2::IsRecordFiltered( unsigned const record ) const
{
	FilterReason reason = notFiltered;

	// EFIGS Palm devices essentially use 1252, so use this higher
	// performance lookup algorithm when possible.

#ifdef OS_WINDOWS
	if( 1252 == m_codepage )
	{
	    if( !m_sLookupA.Length() )
		return notFiltered;

	    RawRecordEntry raw;
	    unsigned const dmRecord = ToDmRecord( record );
	    SPMem<BYTE> const spRec = ReadRecordData( dmRecord, &raw );

	    if( spRec )
	    {
		void* p;

		if( m_fFilterPrivate )
		{
		    p = GetMemberPtr( spRec, tagFlags );

		    if( p )
		    {
			WORD flags;

			memcpy( &flags, p, sizeof( flags ) );
			flags = MacWord( flags );

			if( flags & flagPrivate )
			{
			    reason = reason | filteredPrivate;

			    if( !m_fCompleteReasons )
				return reason;
			}
		    }
		}

		p = GetMemberPtr( spRec, tagDescription );

		if( p && StrLookup( &m_sLookupA, reinterpret_cast<char const*>( p ) ) >= 0 )
		    return notFiltered;

		p = GetMemberPtr( spRec, tagNote );

		if( p && StrLookup( &m_sLookupA, reinterpret_cast<char const*>( p ) ) >= 0 )
		    return notFiltered;

		p = GetMemberPtr( spRec, tagCustomText );

		if( p && StrLookup( &m_sLookupA, reinterpret_cast<char const*>( p ) ) >= 0 )
		    return notFiltered;
	    }
	}
	else
#endif
	{
	    // This has not been tested in DbOpenSimpleHs2, but similar code
	    // has been tested in DbOpenHs2::IsRecordFiltered().

	    if( !m_sLookup.Length() )
		return notFiltered;

	    SPNew<Item> const spitem = UnPackRecord( record );

	    if( spitem )
	    {
		if( m_fFilterPrivate && spitem->GetPrivate() )
		{
		    reason = reason | filteredPrivate;

		    if( !m_fCompleteReasons )
			return reason;
		}

		if( StrLookup( &m_sLookup, spitem->GetDescription() ) >= 0 ||
		    StrLookup( &m_sLookup, spitem->GetNote() ) >= 0 ||
		    StrLookup( &m_sLookup, spitem->GetCustomText() ) >= 0 )
		    return notFiltered;
	    }
	}

	reason = reason | filteredLookup;
	return reason;
}

Item*
DbOpenSimpleHs2::UnPackRecord( unsigned const record ) const
{
	Assert( record != noRecordSelected );
	Assert( int( record ) >= 0 && record < Count() );

	if( int( record ) < 0 || record >= Count() )
	    return 0;

	RawRecordEntry raw;
	unsigned const dmRecord = ToDmRecord( record );
	SPMem<BYTE> spRec = ReadRecordData( dmRecord, &raw );

	if( !spRec )
	    return 0;

	SPNew<Item> spItem = new Item;

	if( !UnPackRecordPtr( spRec, maskALL, spItem, record ) )
	    return 0;

	return spItem.Transfer();
}

bool
DbOpenSimpleHs2::GetItemQuickInfo( unsigned const record, ItemQuickInfo* const pqi ) const
{
	Assert( pqi );

	BYTE bTag;
	WORD flags;
	RawRecordEntry raw;
	unsigned const dmRecord = ToDmRecord( record );
	SysError e = SE_OK;

	if( !ReadRawRecordEntry( dmRecord, &raw ) )
	    return false;

	if( !m_spFile->Seek( raw.GetOfsData(), e ) ||
	    !m_spFile->Read( &bTag, sizeof( bTag ), e ) )
	    return false;

	if( bTag != tagFlags )
	    return false;

	if( !m_spFile->Read( &flags, sizeof( flags ), e ) )
	    return false;

	flags = MacWord( flags );
	bool const fDeleted = !!( raw.GetAttr() & dmRecAttrDelete );

	pqi->Initialize( flags, fDeleted );

	return true;
}

bool
DbOpenSimpleHs2::GetRecordUniqueId( unsigned const record, DWORD* const pdwUniqueId ) const
{
	Assert( pdwUniqueId );

	RawRecordEntry raw;
	unsigned const dmRecord = ToDmRecord( record );

	if( !ReadRawRecordEntry( dmRecord, &raw ) )
	    return false;

	*pdwUniqueId = raw.GetUid();

	return true;
}

bool
DbOpenSimpleHs2::FindRecordByUniqueId( DWORD const uidFind, unsigned* const precord ) const
{
	Assert( precord );

	DWORD uidRecord;

	for( unsigned ii = 0; ii < Count(); ii++ )
	{
	    if( !GetRecordUniqueId( ToDmRecord( ii ), &uidRecord ) )
		break;

	    if( uidFind == uidRecord )
	    {
		*precord = ii;
		return true;
	    }
	}

	return false;
}

bool
DbOpenSimpleHs2::ReadRawRecordEntry( unsigned const dmRecord, RawRecordEntry* const praw ) const
{
	Assert( praw );
	Assert( noRecordSelected != dmRecord );
	Assert( int( dmRecord ) >= 0 && dmRecord < DmCount() );

	if( int( dmRecord ) < 0 || dmRecord >= DmCount() )
	    return false;

	SysError e = SE_OK;
	DWORD const dwOffset = sizeof( PdbHeader ) + ( dmRecord * sizeof( *praw ) );

	if( !m_spFile->Seek( dwOffset, e ) ||
	    !m_spFile->Read( praw, sizeof( *praw ), e ) )
	    return false;

	return true;
}

DWORD
DbOpenSimpleHs2::GetRecordSize( unsigned const dmRecord, DWORD const ofsData ) const
{
	Assert( dmRecord != noRecordSelected );
	Assert( int( dmRecord ) >= 0 && dmRecord < DmCount() );

#ifdef DEBUG
	RawRecordEntry rawVerify;
	Assert( ReadRawRecordEntry( dmRecord, &rawVerify ) );
	Assert( rawVerify.GetOfsData() == ofsData );
#endif

	unsigned cbRecord = 0;
	unsigned const dmRecordNext = dmRecord + 1;

	if( dmRecordNext < DmCount() )
	{
	    // For most records in the database, compute the size by comparing
	    // the offset of this record versus the offset of the next record.

	    RawRecordEntry rawNext;

	    if( !ReadRawRecordEntry( dmRecordNext, &rawNext ) )
		return 0;

	    Assert( rawNext.GetOfsData() > ofsData );

	    cbRecord = rawNext.GetOfsData() - ofsData;
	}
	else
	{
	    // For the last record in the database the only way to find its
	    // size is to compare its offset versus the end of the file.

	    Assert( dmRecordNext == DmCount() );

	    SysError e = SE_OK;
	    unsigned const cbFileSize = m_spFile->Size( e );

	    if( e || cbFileSize <= ofsData )
		return 0;

	    cbRecord = cbFileSize - ofsData;
	}

	//$ review: (chrisant) I don't think that the <64kb size is enforced
	// in HS2 (except by the Palm OS crashing) or in HS3.  That could
	// represent a problem, but at least for now I'm going to gloss over
	// that scenario since realistically it's a pretty obscure and
	// questionable scenario.

	Assert( cbRecord > 0 );
	Assert( cbRecord < 65535 );

	return cbRecord;
}

BYTE*
DbOpenSimpleHs2::ReadRecordData( unsigned const dmRecord, RawRecordEntry* const praw ) const
{
	if( !ReadRawRecordEntry( dmRecord, praw ) )
	    return 0;

	DWORD const cbRecord = GetRecordSize( dmRecord, praw->GetOfsData() );

	if( !cbRecord )
	    return 0;

	SysError e = SE_OK;
	SPMem<BYTE> spRec = static_cast<BYTE*>( MemAlloc( cbRecord ) );

	//$ todo: (chrisant) Trace on errors/corruption?

	if( !m_spFile->Seek( praw->GetOfsData(), e ) ||
	    !m_spFile->Read( spRec, cbRecord, e ) )
	    return 0;

	return spRec.Transfer();
}

