/*
 * Progress.h - Abstract interface for reporting progress.
 */

#pragma once

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

/*
 * IProgress - Progress reporting.
 */

class IProgress;

extern void CreateProgressObject( IProgress** ppProgress );

enum
{
	PROGRESS_NOMETER	= 0x0001,	// Show a progress caption, but no meter.
	PROGRESS_FORCEVISIBLE	= 0x0002,	// Show the progress window immediately.
	PROGRESS_NOWAIT		= 0x0004,	// Hide the progress window immediately when the operation completes:  normally the progress bar window will stay visible for at least 1/2 second, even if the operation completed more quickly, to avoid confusing the user with brief unreadable flickers.
	PROGRESS_PREDICTIVEWAIT	= 0x0008,	// Avoid showing the progress window unless the initial progress suggests the entire operation will complete quickly.
	PROGRESS_CANCELABLE	= 0x0010,	// Operation is cancellable:  the progress window can show a [Cancel] button and Update() can return false if the user has cancelled.
	PROGRESS_NODISPATCH	= 0x0020,	// Not safe to handle events or dispatch messages during the operation in progress.
};

enum ProgressCaption
{
	CAPTION_WORKING,			// "Working..."
	CAPTION_SCANNING,			// "Scanning..."
	CAPTION_SORTING,			// "Sorting..."
	CAPTION_OPENING,			// "Opening..."
	CAPTION_SAVING,				// "Saving..."
	CAPTION_COPYING,			// "Copying..."
	CAPTION_MOVING,				// "Moving..."
	CAPTION_VERIFYING,			// "Verifying..."
	CAPTION_ITEMCOPIED,			// "Item Copied."
};

class IProgress
{
    public:
	virtual		~IProgress() {}
	virtual void	Init( unsigned cMax, WORD flags, ProgressCaption caption ) = 0;
	virtual bool	Update( unsigned cProgress ) = 0;
	virtual void	End() = 0;
};

class ScopedProgressHolder
{
    public:
	ScopedProgressHolder( IProgress* pProgress = 0 ) : m_pProgress( pProgress ), m_fDelete( false ) {}
	~ScopedProgressHolder() { End(); if( m_fDelete ) delete m_pProgress; }

	void		Create();

	void		Init( unsigned cMax, WORD flags = 0, ProgressCaption caption = CAPTION_WORKING ) { if( m_pProgress ) m_pProgress->Init( cMax, flags, caption ); }
	bool		Update( unsigned cProgress ) { return m_pProgress ? m_pProgress->Update( cProgress ) : false; }
	void		End() { if( m_pProgress ) m_pProgress->End(); }

    private:
	IProgress*	m_pProgress;
	bool		m_fDelete;
};

inline void
ScopedProgressHolder::Create()
{
	Assert( !m_pProgress );

	if( !m_pProgress )
	{
	    CreateProgressObject( &m_pProgress );
	    m_fDelete = true;
	}
}

