/*
 * SecureGlue.cpp - Ease compilation for compilers that don't support the
 * secure CRT function signatures.
 */

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#include "precomp.h"

#if !defined( __STDC_WANT_SECURE_LIB__ )

#include "secureglue.h"

inline void
__HelperToEnsureBufferWillBeNullTerminated( void* const pv, size_t& cchMax, size_t const cbChar )
{
	// The "sn" variety of printf functions did not ensure the buffer is
	// null terminated.  To force the buffer to always be null terminated,
	// we zero out the entire buffer space, and then decrement the buffer
	// size so subsequent use will not overwrite the null at the end.

	ZeroMemory( pv, cchMax * cbChar );

	if( cchMax )
	    cchMax--;
}

int
vsprintf_s( char* pszOut, size_t cchMax, char const* pszFormat, va_list args )
{
	__HelperToEnsureBufferWillBeNullTerminated( pszOut, cchMax, sizeof( *pszOut ) );

	return _vsnprintf( pszOut, cchMax, pszFormat, args );
}

int
vswprintf_s( WCHAR* pszOut, size_t cchMax, WCHAR const* pszFormat, va_list args )
{
	__HelperToEnsureBufferWillBeNullTerminated( pszOut, cchMax, sizeof( *pszOut ) );

	return _vsnwprintf( pszOut, cchMax, pszFormat, args );
}

int
_vstprintf_s( TCHAR* pszOut, size_t cchMax, TCHAR const* pszFormat, va_list args )
{
	__HelperToEnsureBufferWillBeNullTerminated( pszOut, cchMax, sizeof( *pszOut ) );

	return _vsntprintf( pszOut, cchMax, pszFormat, args );
}

int
sprintf_s( char* pszOut, size_t cchMax, char const* pszFormat, ... )
{
	va_list args;
	va_start( args, pszFormat );

	int const cchWritten = vsprintf_s( pszOut, cchMax, pszFormat, args );

	va_end( args );

	return cchWritten;
}

int
swprintf_s( WCHAR* pszOut, size_t cchMax, WCHAR const* pszFormat, ... )
{
	va_list args;
	va_start( args, pszFormat );

	int const cchWritten = vswprintf_s( pszOut, cchMax, pszFormat, args );

	va_end( args );

	return cchWritten;
}

int
_stprintf_s( TCHAR* pszOut, size_t cchMax, TCHAR const* pszFormat, ... )
{
	va_list args;
	va_start( args, pszFormat );

	int const cchWritten = _vstprintf_s( pszOut, cchMax, pszFormat, args );

	va_end( args );

	return cchWritten;
}

#endif

