/*
 * StoreList.h - list of stores.
 */

#pragma once

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

/*
 * Store, PerStore, and AutoOrder data types.
 */

#define cchMaxAisle	( 7 )	// 6 + 1 for null terminator

enum
{
	storeMultipleStores	= 0xfc,		// Only used by the UI.
	storeEditStores		= 0xfd,		// Only used by the UI.
	storeAllStores		= 0xfe,
	storeUnfiledStore	= 0xff,
};

typedef BYTE Store;

struct PerStore
{
	Store		m_idStore;
	TCHAR		m_szAisle[cchMaxAisle];
	LONG		m_lPrice;
};
#ifdef UNICODE
CASSERT( sizeof( PerStore ) == 20 );
#else
CASSERT( sizeof( PerStore ) == 12 );
#endif

struct AutoOrder
{
	Store		m_idStore;
	DWORD		m_ulOrder;
};

inline bool
IsPhantomStore( Store const store )
{
	Assert( store != storeMultipleStores );
	Assert( store != storeEditStores );

	return ( store == storeAllStores || store == storeUnfiledStore );
}

/*
 * StoreList and PerStoreList - list of Store and PerStore data types.
 */

enum XlateResult
{
	xrFailure,
	xrSome,
	xrAll
};

struct StoreTranslationInfo;
class DbOpen;

class StoreList
{
    public:
			StoreList() : m_c( 0 ), m_cAlloc( 0 ), m_prg( 0 ) {}
			~StoreList() { Free(); }

	WORD		Count() const { return m_c; }
	Store const*	Array() const { return m_prg; }
	Store		operator[]( WORD const index ) const { Assert( index < m_c ); return m_prg[index]; }

	bool		Insert( Store id );
	bool		Remove( Store id );
	bool		Find( Store id ) const;

	bool		Copy( StoreList const& other ) { return Copy( other.m_c, other.m_prg ); }
	void		Swap( StoreList& other );
	void		Sort();

	XlateResult	Translate( StoreTranslationInfo* pxi ) const;
	bool		RemoveDuplicates();
	bool		RepairList();

	Store const*	Alloc( WORD c );
	void		Borrow( WORD c, Store* prg );
	bool		Copy( WORD c, Store const* prg );
	void		Clear();
	void		Free();

	bool		FBorrowed() const { return m_c && !m_cAlloc; }

#ifdef DEBUG
	bool		FPristine() const { return !m_c && !m_cAlloc && !m_prg; }
#endif

    private:
	void		SortWorker();

    private:
	BYTE		m_c;
	BYTE		m_cAlloc;
	Store*		m_prg;
};

class PerStoreList
{
    public:
			PerStoreList() : m_c( 0 ), m_cAlloc( 0 ), m_prg( 0 ) {}
			~PerStoreList() { Free(); }

	WORD		Count() const { return m_c; }
	PerStore const*	Array() const { return m_prg; }
	PerStore const&	operator[]( WORD ii ) const { Assert( ii < m_c ); return m_prg[ii]; }

	LONG		GetPrice( Store id ) const;
	Store		GetLowestPriceInfo( LONG* plPrice, Store idCheck = storeUnfiledStore ) const;
	TCHAR const*	GetAisle( Store id ) const;

	PerStore*	Insert( Store id );
	void		Remove( PerStore const* p );
	PerStore*	Find( Store id, WORD* pindex = 0 ) const;

	bool		Copy( PerStoreList const& other ) { return Copy( other.m_c, other.m_prg ); }
	void		Swap( PerStoreList& other );
	void		Sort();

	XlateResult	Translate( StoreTranslationInfo* pxi ) const;
	bool		RepairList();

	PerStore const*	Alloc( WORD c );
#if 0
	void		Borrow( WORD c, PerStore* prg );
#endif
	bool		Copy( WORD c, PerStore const* prg );
	void		Clear();
	void		Free();

	bool		FBorrowed() const { return m_c && !m_cAlloc; }

#ifdef DEBUG
	bool		FPristine() const { return !m_c && !m_cAlloc && !m_prg; }
#endif

    private:
	BYTE		m_c;
	BYTE		m_cAlloc;
	PerStore*	m_prg;
};

/*
 * AutoOrderList - list of AutoOrder data types.
 */

enum AutoOrderListValidity
{
	aolValid,
	aolDuplicates,
	aolInvalidStore_AllStores,
	aolInvalidStore_OutOfBounds,
	aolInvalidStore_Inactive,
	aolUnsorted,
};

class AutoOrderList
{
    public:
			AutoOrderList() : m_c( 0 ), m_cAlloc( 0 ), m_prg( 0 ) {}
			~AutoOrderList() { Free(); }

	WORD		Count() const { return m_c; }
	AutoOrder const* Array() const { return m_prg; }
	AutoOrder const& operator[]( WORD ii ) const { Assert( ii < m_c ); return m_prg[ii]; }

	DWORD		GetAutoOrder( Store id ) const;

	AutoOrder*	Insert( Store id );
	void		Remove( AutoOrder const* p );
	AutoOrder*	Find( Store id, WORD* pindex = 0 ) const;

	void		Swap( AutoOrderList& other );
	void		Sort();

	AutoOrder const* Alloc( WORD c );
	bool		AllocForXfer( WORD c );
	void		Clear();
	void		Free();

	bool		FBorrowed() const { return m_c && !m_cAlloc; }

#ifdef DEBUG
	bool		FPristine() const { return !m_c && !m_cAlloc && !m_prg; }
	AutoOrderListValidity TestValid() const;
#endif

    private:
	BYTE		m_c;
	BYTE		m_cAlloc;
	AutoOrder*	m_prg;
};

/*
 * Store translation info, when copying an item from one database to another.
 */

struct StoreTranslationInfo
{
			StoreTranslationInfo() : pdbFrom( 0 ), pdbTo( 0 ) {}

	DbOpen*		pdbFrom;
	DbOpen*		pdbTo;
	StoreList	storesXlate;	// Translated store ID list.
	PerStoreList	perstoreXlate;	// Translated PerStore info.
	StoreList	storesAdd;	// List of store IDs to add (built from storesXlate).
	StoreList	storesAddPer;	// List of store IDs to add (built from perstoreXlate).
};

/*
 * Worker functions.
 */

bool FAddStores( DbOpen* pdbFrom, DbOpen* pdbTo, StoreList const* pstoresAdd, StoreList const* pstoresAddPer );

