/*
 * stores.h
 */

#pragma once

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#pragma pack( push, 2 )

/*
 * Store List Metadata Record
 *
 * The master store list is a record in the database, at index 0.  The record
 * is a fixed size, which eliminates complicated code for resizing the record
 * and the lists it contains, making the app run faster as well as saving
 * battery power.  The record is not created until the user defines one or
 * more stores; if the user deletes all defined stores, we also delete the
 * master store list record.
 *
 * The format of the record is:
 *
 *	UInt8 tag - tagMasterStoreList, identifies this as the record
 *	containing the master store list.
 *
 *	UInt8 cMax - maximum number of stores this database supports (cMax
 *	must be an even number, to ensure UInt16 alignment for rgStore array).
 *	This is for future compatability and is otherwise unused.
 *
 *	UInt8 cStores - number of stores (not including "All Stores").
 *
 *	UInt8 reserved - ensure correct alignment.
 *
 *	UInt8 rgIndexToId[cMax] - sorted index of stores, maps sorted indices
 *	to actual store entries.
 *
 *	StoreDef rgStore[cMax] - info for each store, hashes based on first
 *	character of name.
 *
 * For a 100-store list, the total data size of the record is 3702 bytes.
 */

#define MAX_STORES		( 100 )		// Maximum number of stores.
#define cchMaxStoreName		( 32 )		// 31 chars max, plus null terminator.

struct StoreDef
{
    public:
	IMPL_WORD( m_cBold )
    private:
	UInt16			m_cBold;
    public:
	Char			m_szStore[cchMaxStoreName];
};
CASSERT( ( sizeof( StoreDef ) & 1 ) == 0 );	// Verify alignment.
CASSERT( ( MAX_STORES & 1 ) == 0 );		// MAX_STORES must be even, for alignment.

class MasterStoreList
{
    public:
				MasterStoreList();

	UInt8			IdFromIndex( UInt8 i ) const;
	UInt8			IndexFromId_Slow( UInt8 id ) const;
	StoreDef*		StoreFromIndex( UInt8 i );

    public:
	//$ todo: (chrisant) hide these behind accessor functions
	UInt8			m_tag;		// Must be tagMasterStoreList.
	UInt8			m_cMax;		// Must be MAX_STORES.
	UInt8			m_cStores;	// Number of stores.
	UInt8			m_reserved;	// Alignment.
	Store			m_rgIndexToId[MAX_STORES];
	StoreDef		m_rgStore[MAX_STORES];
};

/*
 * External types.
 */

class Item;
typedef Item ItemType;
typedef Item* ItemPtr;

class DbOpenHs2;

/*
 * Store and PerStore.
 */

typedef UInt8 Store;

#define cchMaxAisleHs2		( 7 )		// 6 + 1 for null terminator.

struct PerStore0
{
	Store			m_idStore;
	UInt8			m_aisle;
	Int32			m_lPrice;
};
CASSERT( sizeof( PerStore0 ) == 6 );

struct PerStore2
{
	Store			m_idStore;
	Char			m_szAisle[cchMaxAisleHs2];
	Int32			m_lPrice;
};
CASSERT( sizeof( PerStore2 ) == 12 );

#define noStore			( Store( -1 ) )
#define allStores		( Store( -2 ) )

/*
 * PerStoreHs2List.
 */

class PerStoreHs2List
{
    public:
				PerStoreHs2List() : m_c( 0 ), m_cAlloc( 0 ), m_prg( 0 ) {}
				PerStoreHs2List( PerStoreList const& list );
				~PerStoreHs2List() { Free(); }

	WORD			Count() const { return m_c; }
	PerStore2 const*	Array() const { return m_prg; }
	PerStore2 const&	operator[]( UInt16 ii ) const { Assert( ii < m_c ); return m_prg[ii]; }

	Int32			GetPrice( Store id ) const;
	Store			GetLowestPriceInfo( Int32* plPrice, Store idCheck /*= noStore*/ ) const;
	Char const*		GetAisle( Store id ) const;

	PerStore2*		Find( Store id, UInt16* pindex = 0 ) const;

	PerStore2 const*	Alloc( UInt16 c );
	bool			Copy( UInt16 c, PerStore0 const* prg );
	bool			Copy( UInt16 c, PerStore2 const* prg );
	void			Clear();
	void			Free();

	void			MacAdjust();

	void			UpConvert( PerStoreList* pList );

#ifdef DEBUG
	bool			FPristine() const { return !m_c && !m_cAlloc && !m_prg; }
#endif

private:
	UInt8			m_c;
	UInt8			m_cAlloc;
	PerStore2*		m_prg;
};

#pragma pack( pop )

